Package: GreedyExperimentalDesign
Type: Package
Title: Greedy Experimental Design Construction
Version: 1.6.1
Date: 2026-04-30
Author: Adam Kapelner [aut, cre] (ORCID: 0000-0001-5985-6792),
  David Azriel [aut],
  Abba Krieger [aut]
Authors@R: c(person(given = "Adam",
                      family = "Kapelner",
                      role = c("aut", "cre"),
                      email = "kapelner@qc.cuny.edu",
                      comment = "ORCID: 0000-0001-5985-6792"),
               person(given = "David",
                      family = "Azriel",
                      role = "aut"),
               person(given = "Abba",
                      family = "Krieger",
                      role = "aut"))
Maintainer: Adam Kapelner <kapelner@qc.cuny.edu>
Description: Computes experimental designs for two-arm experiments with
    covariates using multiple methods, including: (0) complete randomization
    and randomization with forced-balance; (1) greedy optimization of a balance
    objective function via pairwise switching; (2) numerical optimization via
    'gurobi'; (3) rerandomization; (4) Karp's method for one covariate; (5)
    exhaustive enumeration for small sample sizes; (6) binary pair matching
    using 'nbpMatching'; (7) binary pair matching plus method (1) to further
    optimize balance; (8) binary pair matching plus method (3) to further
    optimize balance; (9) Hadamard designs; and (10) simultaneous multiple
    kernels. For the greedy, rerandomization, and related methods, three
    objective functions are supported: Mahalanobis distance, standardized sums
    of absolute differences, and kernel distances via the 'kernlab' library.
    This package is the result of a stream of research that can be found in
    Krieger, A. M., Azriel, D. A., and Kapelner, A. (2019). "Nearly Random
    Designs with Greatly Improved Balance." Biometrika 106(3), 695-701
    <doi:10.1093/biomet/asz026>. Krieger, A. M., Azriel, D. A., and
    Kapelner, A. (2023). "Better experimental design by hybridizing binary
    matching with imbalance optimization." Canadian Journal of Statistics,
    51(1), 275-292 <doi:10.1002/cjs.11685>.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.1.0), rJava (>= 0.9-6)
SystemRequirements: Java (>= 22.0); wgpu-native >= 0.19.0 (optional,
        for GPU acceleration, see
        https://github.com/gfx-rs/wgpu-native)
LinkingTo: Rcpp
Imports: Rcpp, checkmate, nbpMatching, rlist, stringr, stringi,
        kernlab, ggplot2, graphics, grDevices, stats
Suggests: testthat (>= 3.0.0), pkgload, R6
Config/testthat/edition: 3
URL: https://github.com/kapelner/GreedyExperimentalDesign
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-30 05:54:19 UTC; kapelner
Repository: CRAN
Date/Publication: 2026-04-30 07:40:09 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-01 19:20:20 UTC; unix
Archs: GreedyExperimentalDesign.so.dSYM

Package: GreedyExperimentalDesign
Type: Package
Title: Greedy Experimental Design Construction
Version: 1.6.1
Date: 2026-04-30
Author: Adam Kapelner [aut, cre] (ORCID: 0000-0001-5985-6792),
  David Azriel [aut],
  Abba Krieger [aut]
Authors@R: c(person(given = "Adam",
                      family = "Kapelner",
                      role = c("aut", "cre"),
                      email = "kapelner@qc.cuny.edu",
                      comment = "ORCID: 0000-0001-5985-6792"),
               person(given = "David",
                      family = "Azriel",
                      role = "aut"),
               person(given = "Abba",
                      family = "Krieger",
                      role = "aut"))
Maintainer: Adam Kapelner <kapelner@qc.cuny.edu>
Description: Computes experimental designs for two-arm experiments with
    covariates using multiple methods, including: (0) complete randomization
    and randomization with forced-balance; (1) greedy optimization of a balance
    objective function via pairwise switching; (2) numerical optimization via
    'gurobi'; (3) rerandomization; (4) Karp's method for one covariate; (5)
    exhaustive enumeration for small sample sizes; (6) binary pair matching
    using 'nbpMatching'; (7) binary pair matching plus method (1) to further
    optimize balance; (8) binary pair matching plus method (3) to further
    optimize balance; (9) Hadamard designs; and (10) simultaneous multiple
    kernels. For the greedy, rerandomization, and related methods, three
    objective functions are supported: Mahalanobis distance, standardized sums
    of absolute differences, and kernel distances via the 'kernlab' library.
    This package is the result of a stream of research that can be found in
    Krieger, A. M., Azriel, D. A., and Kapelner, A. (2019). "Nearly Random
    Designs with Greatly Improved Balance." Biometrika 106(3), 695-701
    <doi:10.1093/biomet/asz026>. Krieger, A. M., Azriel, D. A., and
    Kapelner, A. (2023). "Better experimental design by hybridizing binary
    matching with imbalance optimization." Canadian Journal of Statistics,
    51(1), 275-292 <doi:10.1002/cjs.11685>.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 4.1.0), rJava (>= 0.9-6)
SystemRequirements: Java (>= 22.0); wgpu-native >= 0.19.0 (optional,
        for GPU acceleration, see
        https://github.com/gfx-rs/wgpu-native)
LinkingTo: Rcpp
Imports: Rcpp, checkmate, nbpMatching, rlist, stringr, stringi,
        kernlab, ggplot2, graphics, grDevices, stats
Suggests: testthat (>= 3.0.0), pkgload, R6
Config/testthat/edition: 3
URL: https://github.com/kapelner/GreedyExperimentalDesign
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-30 05:54:19 UTC; kapelner
Repository: CRAN
Date/Publication: 2026-04-30 07:40:09 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-30 11:05:53 UTC; unix
Archs: GreedyExperimentalDesign.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,greedy_experimental_design_search)
S3method(plot,greedy_multiple_kernel_experimental_design)
S3method(print,binary_match_structure)
S3method(print,binary_then_greedy_experimental_design)
S3method(print,binary_then_rerandomization_experimental_design)
S3method(print,greedy_experimental_design_search)
S3method(print,greedy_multiple_kernel_experimental_design)
S3method(print,karp_experimental_design_search)
S3method(print,optimal_experimental_design_search)
S3method(print,pairwise_matching_experimental_design_search)
S3method(print,rerandomization_experimental_design_search)
S3method(summary,binary_match_structure)
S3method(summary,binary_then_greedy_experimental_design)
S3method(summary,binary_then_rerandomization_experimental_design)
S3method(summary,greedy_experimental_design_search)
S3method(summary,greedy_multiple_kernel_experimental_design)
S3method(summary,karp_experimental_design_search)
S3method(summary,optimal_experimental_design_search)
S3method(summary,pairwise_matching_experimental_design_search)
S3method(summary,rerandomization_experimental_design_search)
export(all_elements_same_cpp_wrap)
export(complete_randomization)
export(complete_randomization_with_forced_balanced)
export(computeBinaryMatchStructure)
export(compute_distance_matrix_cpp_wrap)
export(compute_distance_matrix_gpu)
export(compute_gram_matrix)
export(compute_kernel_matrix_gpu)
export(compute_multiple_kernel_objective_vals_gpu)
export(compute_objective_val)
export(compute_objective_vals_gpu)
export(compute_randomization_metrics)
export(compute_randomization_metrics_gpu)
export(create_all_ys_cpp_wrap)
export(full_greedy_search_gpu)
export(ged_gpu_available)
export(ged_gpu_devices)
export(gen_pm_designs_cpp_wrap)
export(gen_var_cov_matrix_block_designs)
export(generate_block_design_cpp_wrap)
export(generate_stdzied_design_matrix)
export(greedy_orthogonalization_curation)
export(greedy_orthogonalization_curation2)
export(hadamardExperimentalDesign)
export(imbalanced_block_designs)
export(imbalanced_complete_randomization)
export(initBinaryMatchExperimentalDesignSearchObject)
export(initBinaryMatchFollowedByGreedyExperimentalDesignSearchObject)
export(initBinaryMatchFollowedByRerandomizationDesignSearchObject)
export(initGreedyExperimentalDesignObject)
export(initGreedyMultipleKernelExperimentalDesignObject)
export(initGurobiNumericalOptimizationExperimentalDesignObject)
export(initKarpExperimentalDesignObject)
export(initOptimalExperimentalDesignObject)
export(initRerandomizationExperimentalDesignObject)
export(optimize_asymmetric_treatment_assignment)
export(plot_obj_val_by_iter)
export(plot_obj_val_order_statistic)
export(resultsBinaryMatchSearch)
export(resultsBinaryMatchThenGreedySearch)
export(resultsBinaryMatchThenRerandomizationSearch)
export(resultsGreedySearch)
export(resultsGurobiNumericalOptimizeSearch)
export(resultsKarpSearch)
export(resultsMultipleKernelGreedySearch)
export(resultsOptimalSearch)
export(resultsRerandomizationSearch)
export(safe_cov_inverse)
export(searchTimeElapsed)
export(shuffle_cpp_wrap)
export(standardize_data_matrix)
export(startSearch)
export(stopSearch)
import(checkmate)
import(grDevices)
import(graphics)
import(rJava)
import(stats)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(checkmate,assertChoice)
importFrom(checkmate,assertClass)
importFrom(checkmate,assertCount)
importFrom(checkmate,assertLogical)
importFrom(checkmate,assertMatrix)
importFrom(checkmate,assertNumeric)
importFrom(checkmate,assertTRUE)
importFrom(checkmate,vname)
importFrom(nbpMatching,distancematrix)
importFrom(nbpMatching,nonbimatch)
importFrom(rlist,list.cbind)
importFrom(stringi,stri_replace_first_regex)
importFrom(stringr,str_detect)

#' Curate More Orthogonal Vectors Greedily
#' 
#' This function takes a set of allocation vectors and pares them down one-by-one
#' by eliminating the vector that can result in the largest reduction in Avg[ |r_ij| ].
#' It is recommended to begin with a set of unmirrored vectors for speed. Then add the mirrors later
#' for whichever subset you wish.
#' 
#' @param W 		A matrix in \eqn{{-1, 1}^{R x n}} which have R allocation vectors for an experiment of sample size n.
#' @param R0 		The minimum number of vectors to consider in a design. The default is the true bottom, two.
#' @param verbose 	Default is \code{FALSE} but if not, it will print out a message for each iteration.
#' @return 			A list with two elements: (1) \code{avg_abs_rij_by_R} which is a data frame with R - Rmin + 1 rows and 
#' 					columns R and average absolute r_ij and (2) \code{Wsorted} which provides the collection of vectors in
#' 					sorted by best average absolute r_ij in row order from best to worst.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' W = matrix(sample(c(-1, 1), 6 * 8, replace = TRUE), nrow = 6)
#' W2 = greedy_orthogonalization_curation2(W, R0 = 4, verbose = FALSE)
#' dim(W2)
#' }
#' @export
greedy_orthogonalization_curation2 = function(W, R0 = 100, verbose = FALSE){
	assertClass(W, "matrix")
	assertCount(R0)
	R = nrow(W)
	assertNumeric(R0, upper = R)
	assertLogical(verbose)
	
	Rij = abs(W %*% t(W)) / ncol(W)
	diag(Rij) = 0 #for speed
	res = greedy_orthogonalization_eliminate_cpp(Rij, R0, FALSE)
	if (verbose && length(res$eliminated) > 0){
		for (iter_num in seq_along(res$eliminated)){
			if (iter_num %% 100 == 0){
				cat("iter", iter_num, "\n")
			}
		}
	}
	#now return the subset of original allocations in order of elimination
	W[res$remaining, ]
}

#' Begin A Greedy Pair Multiple Kernel Switching Search
#' 
#' This method creates an object of type greedy_multiple_kernel_experimental_design and will immediately initiate
#' a search through allocation space for forced balance designs. For debugging, you can use set the \code{seed}
#' parameter and \code{num_cores = 1} to be assured of deterministic output.
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more optimal design. We will standardize this matrix by column internally.
#' @param max_designs 		The maximum number of designs to be returned. Default is 10,000. Make this large 
#' 							so you can search however long you wish as the search can be stopped at any time by
#' 							using the \code{\link{stopSearch}} method
#' @param objective			The method used to aggregate the kernel objective functions together. Default is "added_pct_reduction".
#' @param kernel_pre_num_designs	How many designs per kernel to run to explore the space of kernel objective values. Default is 2000.
#' @param kernel_names		An array with the kernels to compute with default parameters. Must have elements in the following set:
#' 							"mahalanobis", "poly_s" where the "s" is a natural number 1 or greater,
#' 							"exponential", "laplacian", "inv_mult_quad", "gaussian". Default is \code{NULL} to
#' 							indicate the kernels are specified manually using the \code{Kgrams} parameter.
#' @param Kgrams			A list of M >= 1 elements where each is a \code{n x n} matrix whose
#' 							entries are the evaluation of the kernel function between subject i and subject j. 
#' 							Default is \code{NULL} to indicate this was specified using the convenience parameter 
#' 							\code{kernel_names}.
#' @param maximum_gain_scaling	This controls how much the percentage of possible improvement on a kernel objective function
#' 								should be scaled by. The minimum is 1 which allows for designs that could potentially have >=100%
#' 								improvement over original. We recommend 1.1 which means that a design that was found to be the best
#' 								of the \code{kernel_pre_num_designs} still has 1/1.1 = 9\% room to grow making it highly unlikely that
#' 								any design could be >= 100\%.
#' @param kernel_weights	A vector with positive weights (need not be normalized) where each element represents the weight of 
#' 							each kernel. The default is \code{NULL} for uniform weighting.
#' @param wait				Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 							deafult is \code{FALSE}.
#' @param start				Should we start searching immediately (default is \code{TRUE}).
#' @param semigreedy		Should we use a fully greedy approach or the quicker semi-greedy approach? The default is
#' 							\code{FALSE} corresponding to the fully greedy approach.
#' @param max_iters			Should we impose a maximum number of greedy switches? The default is \code{Inf} which a flag 
#' 							for ``no limit.''
#' @param diagnostics		Returns diagnostic information about the iterations including (a) the initial starting
#' 							vectors, (b) the switches at every iteration and (c) information about the objective function
#' 							at every iteration (default is \code{FALSE} to decrease the algorithm's run time).
#' @param num_cores 		The number of CPU cores you wish to use during the search. The default is \code{1}.
#' @param seed				The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
#' 							the output will not be deterministic. Default is \code{NULL} for no seed set.
#' @param verbose			Should the algorithm emit progress output? Default is \code{TRUE}.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' @return					An object of type \code{greedy_experimental_design_search} which can be further operated upon
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' mk = initGreedyMultipleKernelExperimentalDesignObject(
#'   X,
#'   max_designs = 4,
#'   kernel_pre_num_designs = 4,
#'   num_cores = 1,
#'   kernel_names = c("mahalanobis", "gaussian"),
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' mk
#' }
#' @export
initGreedyMultipleKernelExperimentalDesignObject = function(
		X = NULL, 
		max_designs = 10000, 
		objective = "added_pct_reduction",
		kernel_pre_num_designs = 2000,
		kernel_names = NULL,
		Kgrams = NULL,
		maximum_gain_scaling = 1.1,
		kernel_weights = NULL,
		wait = FALSE, 
		start = TRUE,
		max_iters = Inf,
		semigreedy = FALSE, 
		diagnostics = FALSE,
		num_cores = 1,
		seed = NULL,
		verbose = TRUE,
		use_safe_inverse = FALSE){
	
	assertMatrix(X)
	Xstd = standardize_data_matrix(X)
	if ((is.null(kernel_names) & is.null(Kgrams)) | (!is.null(kernel_names) & !is.null(Kgrams))){
		stop("You must specify EITHER the kernel_names or the Kgrams argument.")
	}
	assertCount(max_designs, positive = TRUE)
	assertCount(kernel_pre_num_designs, positive = TRUE)
	if (!is.infinite(max_iters)){
		assertCount(max_iters, positive = TRUE)
	}
	assertLogical(wait)
	assertLogical(start)
	assertLogical(semigreedy)
	assertLogical(diagnostics)
	assertCount(num_cores, positive = TRUE)
	assertNumeric(seed, null.ok = TRUE)
	assertNumeric(maximum_gain_scaling, lower = 1)
	assertLogical(verbose)
	assertLogical(use_safe_inverse)
	
	n = nrow(Xstd)
	p = ncol(Xstd)
	nT = n / 2
	if (n %% 2 != 0){
		stop("Design matrix must have even rows to have equal treatments and controls")
	}
	
	if (!is.null(kernel_names)){
		Kgrams = list()
		for (i_k in 1 : length(kernel_names)){
			if (kernel_names[i_k] == "mahalanobis"){
				if (use_safe_inverse){
					SinvX = safe_cov_inverse(X)
				} else {
					SinvX = solve(var(X))
				}
				Kgrams[[i_k]] = X %*% SinvX %*% t(X)
			} else {
				Kgrams[[i_k]] = matrix(NA, n, n)
				for (i in 1 : n){
					for (j in 1 : n){
						xi = Xstd[i, , drop = FALSE]
						xj = Xstd[j, , drop = FALSE]
						
						if (str_detect(kernel_names[i_k], "^poly_")){
							s = as.integer(stri_replace_first_regex(kernel_names[i_k], "^poly_(\\d+)", "$1"))
							Kgrams[[i_k]][i, j] = (1 + xi %*% t(xj) / s)^s
						} else if (kernel_names[i_k] == "exponential"){
							Kgrams[[i_k]][i, j] = exp(xi %*% t(xj))
						} else if (kernel_names[i_k] == "laplacian"){
							Kgrams[[i_k]][i, j] = exp(-sqrt(sum((xi - xj)^2)))
						} else if (kernel_names[i_k] == "inv_mult_quad"){
							Kgrams[[i_k]][i, j] = 1 / sqrt(sum((xi - xj)^2) + 1)
						} else if (kernel_names[i_k] == "gaussian"){
							Kgrams[[i_k]][i, j] = exp(-sum((xi - xj)^2))
						} else {
							stop(paste0("Invalid kernel name: ", kernel_names[i_k]))
						}
					}
				}
			}
			
		}
	}
	m = length(Kgrams)
	
	for (i_k in 1 : m){
		assertMatrix(Kgrams[[i_k]], nrows = n, ncols = n)
	}
	
	if (!is.null(kernel_weights)){
		assertNumeric(kernel_weights, lower = 0, len = m)
		#normalize here
		kernel_weights = kernel_weights / sum(kernel_weights)
	} else {
		kernel_weights = rep(1 / m, m)
	}
	
	#we are about to construct a GreedyExperimentalDesign java object. First, let R garbage collect
	#to clean up previous GreedyExperimentalDesign objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!
	
	
	if (objective %in% c("added_pct_reduction")){
		#In this method, we explore each kernel case-by-case before doing a search using all of them
		all_starting_vecs = list()
		all_univariate_kernel_data = list()
		if (is.null(seed)){
			prelim_seed = sample.int(1e5, 1)
		} else {
			prelim_seed = seed
		}
		for (i_k in 1 : m){		
			gd = suppressWarnings(initGreedyExperimentalDesignObject(Xstd, 
							max_designs = kernel_pre_num_designs, Kgram = Kgrams[[i_k]], objective = "kernel",
							diagnostics = TRUE, wait = TRUE, num_cores = num_cores, seed = prelim_seed, verbose = verbose)) #same seed should guarantee same starting vectors
			gd_res = resultsGreedySearch(gd, max_vectors = kernel_pre_num_designs, form = "pos_one_min_one")
			objvalsi = array(NA, kernel_pre_num_designs)
			for (i in 1 : kernel_pre_num_designs){
				objvalsi[i] = gd_res$starting_indicTs[i, , drop = FALSE] %*% Kgrams[[i_k]] %*% t(gd_res$starting_indicTs[i, , drop = FALSE])
			}
			obj_val_by_iters = gd_res$obj_val_by_iters
			
			#now we want to add
			
			univariate_kernel_data = data.frame(
				kernel = i_k, 
				log10objvalsi = log10(rep(objvalsi, gd_res$num_iters + 1)), 
				log10objvalsf = log10(unlist(gd_res$obj_val_by_iters))
			)
			#we want to eliminate
			univariate_kernel_data = subset(univariate_kernel_data, abs(univariate_kernel_data$log10objvalsi - univariate_kernel_data$log10objvalsf) > 1e-9)
			univariate_kernel_data$log10_i_over_f = univariate_kernel_data$log10objvalsi - univariate_kernel_data$log10objvalsf
			univariate_kernel_data = univariate_kernel_data[]
			univariate_kernel_data$pct_red_max = 1 - univariate_kernel_data$log10_i_over_f / 
					(max(univariate_kernel_data$log10_i_over_f) * maximum_gain_scaling)
			all_univariate_kernel_data[[i_k]] = univariate_kernel_data
			all_starting_vecs[[i_k]] = gd_res$starting_indicTs
		}		
	}


	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("MultipleKernelGreedyExperimentalDesign.MultipleKernelGreedyExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setMaxDesigns", as.integer(max_designs))
	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))
	if (!is.null(seed)){
		.jcall(java_obj, "V", "setSeed", as.integer(seed))
		if (num_cores != 1){
			warning("Setting the seed with multiple cores does not guarantee deterministic output.")
		}		
	}
	.jcall(java_obj, "V", "setN", as.integer(n))
	.jcall(java_obj, "V", "setP", as.integer(p))
	.jcall(java_obj, "V", "setNumTreatments", as.integer(nT))
	.jcall(java_obj, "V", "setObjective", objective)
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}
	if (max_iters < Inf){
		.jcall(java_obj, "V", "setMaxIters", as.integer(max_iters))
	}
#	#feed in the raw data
#	for (i in 1 : n){	
#		.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), Xstd[i, , drop = FALSE]) #java indexes from 0...n-1
#	}
	#feed in the raw kernel data
	for (i_k in 1 : m){
		for (i in 1 : n){	
			.jcall(java_obj, "V", "setSpecificKgramByRow", as.integer(i_k - 1), as.integer(i - 1), Kgrams[[i_k]][i, , drop = FALSE]) #java indexes from 0...n-1
		}	
	}
	
	if (objective %in% c("added_pct_reduction")){
		#feed in data from each kernel in isolation
		for (i_k in 1 : m){
			#.jcall(java_obj, "V", "setObjValLibrary", as.integer(i_k - 1), all_univariate_kernel_data[[i_k]]$log10_i_over_f)
			.jcall(java_obj, "V", "setMaxReductionLogObjVal", as.integer(i_k - 1), max(all_univariate_kernel_data[[i_k]]$log10_i_over_f))
		}
		#set magnification factor
		.jcall(java_obj, "V", "setMaximumGainScaling", maximum_gain_scaling)
		#feed in weights
		.jcall(java_obj, "V", "setKernelWeights", kernel_weights)
	}
		
	#do we want diagnostics? Set it...
	if (diagnostics){
		.jcall(java_obj, "V", "setDiagnostics")
	}
	
	#is it semigreedy? Set it...
	if (semigreedy){
		.jcall(java_obj, "V", "setSemigreedy")
	}
	
	#now return information as an object (just a list)
	ged = list()
	ged$kernel_names = kernel_names
	ged$Kgrams = Kgrams
	ged$m = m
	ged$all_univariate_kernel_data = all_univariate_kernel_data
	ged$kernel_weights = kernel_weights
	ged$all_starting_vecs = all_starting_vecs
	ged$max_designs = max_designs
	ged$kernel_pre_num_designs = kernel_pre_num_designs
	ged$maximum_gain_scaling = maximum_gain_scaling
	ged$semigreedy = semigreedy
	ged$start = start
	ged$wait = wait
	ged$diagnostics = diagnostics
	ged$verbose = verbose
	ged$X = X
	ged$Xstd = Xstd
	ged$n = n
	ged$p = p
	ged$java_obj = java_obj
	class(ged) = "greedy_multiple_kernel_experimental_design"
	#if the user wants to run it immediately...
	if (start){
		startSearch(ged)
	}
	#return the final object
	ged
}

#' Returns the results (thus far) of the greedy design search for multiple kernels
#' 
#' @param obj 			The \code{greedy_multiple_kernel_experimental_design} object that is currently running the search
#' @param max_vectors	The number of design vectors you wish to return. \code{NULL} returns all of them. 
#' 						This is not recommended as returning over 1,000 vectors is time-intensive. The default is 9. 
#' @param form			Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' mk = initGreedyMultipleKernelExperimentalDesignObject(
#'   X,
#'   max_designs = 4,
#'   kernel_pre_num_designs = 4,
#'   num_cores = 1,
#'   kernel_names = c("mahalanobis", "gaussian"),
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsMultipleKernelGreedySearch(mk, max_vectors = 2, form = "one_zero")
#' res$obj_vals
#' }
#' @export
resultsMultipleKernelGreedySearch = function(obj, max_vectors = 9, form = "one_zero"){
	#get standard information
	greedy_multiple_kernel_experimental_design_search_results = resultsGreedySearch(obj, max_vectors, form)
	last_index = greedy_multiple_kernel_experimental_design_search_results$last_index
	ordered_indices = greedy_multiple_kernel_experimental_design_search_results$ordered_indices
	#now get information specific to this procedure
	if (obj$diagnostics){
		greedy_multiple_kernel_experimental_design_search_results$kernel_obj_vals_by_iter = 
			.jcall(obj$java_obj, "[[[D", "getObjValuesByKernel", as.integer(ordered_indices[1 : last_index] - 1), simplify = TRUE)
	}
	class(greedy_multiple_kernel_experimental_design_search_results) = "greedy_multiple_kernel_experimental_design_search_results"
	#return the final object
	greedy_multiple_kernel_experimental_design_search_results
}


#' Prints a summary of a \code{greedy_multiple_kernel_experimental_design} object
#' 
#' @param x			The \code{greedy_multiple_kernel_experimental_design} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print greedy_multiple_kernel_experimental_design
#' @export
print.greedy_multiple_kernel_experimental_design = function(x, ...){
	print.greedy_experimental_design_search(x, ...)
}

#' Prints a summary of a \code{greedy_multiple_kernel_experimental_design} object
#' 
#' @param object		The \code{greedy_multiple_kernel_experimental_design} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary greedy_multiple_kernel_experimental_design
#' @export
summary.greedy_multiple_kernel_experimental_design = function(object, ...){
	print(object, ...)
}

#' Plots a summary of a \code{greedy_multiple_kernel_experimental_design} object
#' 
#' @param x			The \code{greedy_multiple_kernel_experimental_design} object to be summarized in the plot
#' @param ...		Other parameters to pass to the default plot function
#' @return			An array of order statistics from \link{plot_obj_val_order_statistic} as a list element
#' 
#' @author 			Adam Kapelner
#' @method plot greedy_multiple_kernel_experimental_design
#' @export
plot.greedy_multiple_kernel_experimental_design = function(x, ...){
	plot.greedy_experimental_design_search(x, ...)
}

##' Curate More Orthogonal Vectors Via Numerical Optimization
##' 
##' This function takes a set of allocation vectors and pares them down
##' by eliminating the vectors that can result in the largest reduction in Avg[ |r_ij| ] via
##' Gurobi's numerical optimization for binary integer quadratic programs.
##' 
##' Make sure you setup Gurobi properly first. This means applying for a license, downloading, installing, then make sure 
##' you've registered the license on your computer using the \code{grbgetkey} command. Then you need to install the R package
##' from the local file (it is not on CRAN) via e.g. \code{install.packages("c:/gurobi902/win64/R/gurobi_9.0-2.zip", repos = NULL)}.
##' 
##' @param W 		A matrix in the set ${-1, 1}^{R x n}$ which have R allocation vectors for an experiment of sample size n.
##' @param R0 		The number of vectors in the pared-down design.
##' @param time_limit_seconds	The limit on how long in seconds the binary integer quadratic program should run (default is 2 * 60s).
##' @param num_cores				Number of cores to use during search. Default is \code{2}.
##' @param gurobi_params			A list of optional parameters to be passed to Gurobi (see their documentation online).
##' @param verbose				Should Gurobi log to console? Default is \code{TRUE}.
##' @return 			A list object which houses the results from Gurobi. Depending on the \code{gurobi_parms},
##' 					the data within will be different. The most relevant tags are \code{x} for the best found solution and \code{objval}
##' 					for the object
##' 
##' @author Adam Kapelner
#
#####EXPORT
#gurobi_numerical_optimization_orthogonalization_curation = function(W, R0, verbose = TRUE, num_cores = 2, time_limit_seconds = 2 * 60, gurobi_params = list()){
#	assertClass(W, "matrix")
#	R = nrow(W)
#	assertCount(R0, positive = TRUE)
#	assertNumeric(R0, upper = R)
#	assertCount(num_cores, positive = TRUE)
#	assertNumeric(time_limit_seconds, lower = 1)
#	assertLogical(verbose)
#	assertClass(gurobi_params, "list")
#	
#	n = ncol(W)
#	Rij = abs(W %*% t(W)) / n
#	diag(Rij) = 0
#
#	#  minimize
#	#        [s_1 s_2 ... s_R] Rij [s_1 s_2 ... s_R]^T
#	#  subject to
#	#        s_1 + s_2 + ... + s_R = R0,
#	#        s_1, s_2, ..., s_R binary
#	
#	model 		= list()
#	#quadratic component
#	model$Q     = Rij / (R0^2 - R0)
#	model$obj   = t(as.matrix(rep(0, R)))
#	#constraint component
#	model$A     = t(as.matrix(rep(1, R)))
#	model$rhs   = R0
#	model$sense = "="
#	model$vtype = "B"
#	gurobi_params$LogToConsole = as.numeric(verbose)
#	gurobi_params$Threads = num_cores
#	gurobi_params$TimeLimit = time_limit_seconds	
#	gurobi::gurobi(model, gurobi_params)
#}

.onLoad = function(libname, pkgname) {
  .jpackage(pkgname, lib.loc = libname)
  
  #need to check if proper Java is installed by special request of Prof Brian Ripley
  jv = .jcall("java/lang/System", "S", "getProperty", "java.runtime.version")
  if (substr(jv, 1L, 2L) == "1.") {
	  jvn = as.numeric(paste0(strsplit(jv, "[.]")[[1L]][1:2], collapse = "."))
	  if (jvn < 1.7){
		  warning(paste("Java v", jvn, "detected. Java 7 (at minimum) is needed for this package but is does not seem to be available. This message may be in error; apologies if it is."))
	  }		
  }
}

.onAttach = function(libname, pkgname){
  num_gigs_ram_available = .jcall(.jnew("java/lang/Runtime"), "J", "maxMemory") / 1e9
  packageStartupMessage(paste("Welcome to GreedyExperimentalDesign v", utils::packageVersion("GreedyExperimentalDesign"), ".\n", round(num_gigs_ram_available, 2), "GB memory available for the Java searcher.\n", sep = ""))
}

build_test_configs <- function() {
  n_values = c(16, 96, 192)
  p_values = c(1, 5, 10)
  B_values = c(2, 8)
  r_values = c(100, 200, 1000)
  seed_counter = 0
  configs = list()
  for (n in n_values) {
    for (B in B_values) {
      if (n %% B != 0) {
        next
      }
      n_B = n / B
      if (n_B < 4) {
        next
      }
      prop_T_values = unique(c(1 / n_B, floor(n_B / 2) / n_B))
      prop_T_values = prop_T_values[prop_T_values > 0 & prop_T_values < 1]
      for (p in p_values) {
        for (r in r_values) {
          for (prop_T in prop_T_values) {
            seed_counter = seed_counter + 1
            configs[[length(configs) + 1]] = list(
              n = n,
              p = p,
              r = r,
              B = B,
              prop_T = prop_T,
              seed = seed_counter
            )
          }
        }
      }
    }
  }
  configs
}

options(java.parameters = "-Xmx10g")

library(testthat)
library(GreedyExperimentalDesign)

is_check = nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_")) ||
  nzchar(Sys.getenv("R_CMD_CHECK")) ||
  nzchar(Sys.getenv("_R_CHECK_CRAN_INCOMING_")) ||
  nzchar(Sys.getenv("_R_CHECK_CRAN_INCOMING_REMOTE_"))
if (is_check) {
  message("Skipping testthat tests during R CMD check.")
} else {
  test_check("GreedyExperimentalDesign")
}

#' Computes Randomization Metrics (explained in paper) about a design algorithm
#' 
#' @param designs	A matrix where each column is one design.
#' 
#' @return 			A list of resulting data: the probability estimates for
#' 					each pair in the design of randomness where estmates close
#' 					to ~0.5 represent random assignment, then the entropy metric
#' 					the distance metric, the maximum eigenvalue of the allocation
#' 					var-cov matrix (operator norm) and the squared Frobenius norm 
#' 					(the sum of the squared eigenvalues)
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' designs = matrix(c(1, 0, 1, 0, 0, 1, 0, 1), nrow = 4, ncol = 2)
#' compute_randomization_metrics(designs)
#' }
#' @export
compute_randomization_metrics = function(designs){
	n = nrow(designs)
	r = ncol(designs)
	res = compute_randomization_metrics_cpp(designs)
	p_hat_ijs = res$p_hat_ijs
	rand_entropy_metric = res$rand_entropy_metric
	rand_norm_se_metric = res$rand_norm_se_metric
	
	#for the maximum eigenvalue we need to transform the allocation vector to be in {-1, 1}
	designs[designs == 0] = -1
	#now take the eigendecomposition of the variance-covariance matrix of the allocations
	cov_mat = var(t(designs))
	max_eigenval = NA_real_
	frob_norm_sqd = NA_real_
	if (all(is.finite(cov_mat))){
		e_d = eigen(cov_mat, symmetric = TRUE, only.values = TRUE)
		max_eigenval = max(e_d$values)
		frob_norm_sqd = sum(e_d$values^2)
	}
	
	list(
		p_hat_ijs = p_hat_ijs, 
		rand_entropy_metric = rand_entropy_metric, 
		rand_norm_se_metric = rand_norm_se_metric,
		max_eigenval = max_eigenval,
		frob_norm_sqd = frob_norm_sqd
	)
}

#' Begin A Greedy Pair Switching Search
#' 
#' This method creates an object of type greedy_experimental_design and will immediately initiate
#' a search through allocation space for forced balance designs. For debugging, you can use set the \code{seed}
#' parameter and \code{num_cores = 1} to be assured of deterministic output.
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more optimal design. This parameter must be specified unless you
#' 							choose objective type \code{"kernel"} in which case, the \code{Kgram} parameter must
#' 							be specified.
#' @param nT				The number of treatments to assign. Default is \code{NULL} which is for forced balance allocation
#' 							i.e. nT = nC = n / 2 where n is the number of rows in X (or Kgram if X is unspecified).
#' @param max_designs 		The maximum number of designs to be returned. Default is 10,000. Make this large 
#' 							so you can search however long you wish as the search can be stopped at any time by
#' 							using the \code{\link{stopSearch}} method 
#' @param objective			The objective function to use when searching design space. This is a string
#' 							with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".
#' @param indicies_pairs	A matrix of size $n/2$ times 2 whose rows are indicies pairs. The values of the entire matrix 
#' 							must enumerate all indicies $1, ..., n$. The default is \code{NULL} meaning to use all possible pairs.
#' @param Kgram				If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
#' 							entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.
#' @param wait				Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 							deafult is \code{FALSE}.
#' @param start				Should we start searching immediately (default is \code{TRUE}).
#' @param semigreedy		Should we use a fully greedy approach or the quicker semi-greedy approach? The default is
#' 							\code{FALSE} corresponding to the fully greedy approach.
#' @param max_iters			Should we impose a maximum number of greedy switches? The default is \code{Inf} which a flag 
#' 							for ``no limit.''
#' @param diagnostics		Returns diagnostic information about the iterations including (a) the initial starting
#' 							vectors, (b) the switches at every iteration and (c) information about the objective function
#' 							at every iteration (default is \code{FALSE} to decrease the algorithm's run time).
#' @param num_cores 		The number of CPU cores you wish to use during the search. The default is \code{1}.
#' @param seed				The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
#' 							the output will not be deterministic. Default is \code{NULL} for no seed set.
#' @param verbose			Should the algorithm emit progress output? Default is \code{TRUE}.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' @return					An object of type \code{greedy_experimental_design_search} which can be further operated upon
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' ged
#' }
#' @export
initGreedyExperimentalDesignObject = function(
		X = NULL, 
		nT = NULL,
		max_designs = 10000, 
		objective = "mahal_dist", 
		indicies_pairs = NULL,
		Kgram = NULL,
		wait = FALSE, 
		start = TRUE,
		max_iters = Inf,
		semigreedy = FALSE, 
		diagnostics = FALSE,
		num_cores = 1,
		seed = NULL,
		verbose = TRUE,
		use_safe_inverse = FALSE){
	
	if (!is.null(Kgram)){
		n = nrow(Kgram)
		p = NA
	} else {
		n = nrow(X)
		p = ncol(X)
	}
	if (is.null(nT)){
		nT = n / 2
	}
	verify_objective_function(objective, Kgram, n)
	
	checkCount(nT, positive = TRUE, null.ok = FALSE)
	assertLogical(verbose)
	assertLogical(use_safe_inverse)
	
	if (nT != n / 2 & !is.null(indicies_pairs)){
		stop("indicies_pairs cannot be specified if nT is not half of n")
	}
	
	if (!is.null(indicies_pairs)){
		if (!(all.equal(sort(c(indicies_pairs)), 1 : n))){
			stop("indicies_pairs must cover all indicies 1, 2, ..., n once each.")
		}
	}
	if (objective == "abs_sum_diff"){
		#standardize it -- much faster here
		Xstd = standardize_data_matrix(X)
	}
	if (objective == "mahal_dist"){
		if (p < n){
			if (use_safe_inverse){
				SinvX = safe_cov_inverse(X)
			} else {
				SinvX = solve(stats::var(X))
			}
		}
	}
	

	
	#we are about to construct a GreedyExperimentalDesign java object. First, let R garbage collect
	#to clean up previous GreedyExperimentalDesign objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	
	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("GreedyExperimentalDesign.GreedyExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setMaxDesigns", as.integer(max_designs))
	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))	
	if (!is.null(seed)){
		.jcall(java_obj, "V", "setSeed", as.integer(seed))
		if (num_cores != 1){
			warning("Setting the seed with multiple cores does not guarantee deterministic output.")
		}		
	}
	if (objective != "kernel"){
		p = ncol(X)
		.jcall(java_obj, "V", "setP", as.integer(p))
	}
	.jcall(java_obj, "V", "setN", as.integer(n))
	.jcall(java_obj, "V", "setNumTreatments", as.integer(nT))
	.jcall(java_obj, "V", "setObjective", objective)
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}
	if (max_iters <= 0){stop("max_iters must be positive")}
	if (max_iters < Inf){
		.jcall(java_obj, "V", "setMaxIters", as.integer(max_iters))
	}
	
	if (!is.null(indicies_pairs)){
		for (i in 1 : (n / 2)){	
			.jcall(java_obj, "V", "setLegalPair", as.integer(indicies_pairs[i, ] - 1), as.integer(i - 1)) #java indexes from 0...n-1
		}
	}
	
	#feed in the gram matrix if applicable
	if (!is.null(Kgram)){
		setGramMatrix(java_obj, Kgram)
	} else {
		#feed in the raw data
		for (i in 1 : n){	
			if (objective == "abs_sum_diff"){
				.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), Xstd[i, , drop = FALSE]) #java indexes from 0...n-1
			} else {
				.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), X[i, , drop = FALSE]) #java indexes from 0...n-1
			}
		}
		
		#feed in the inverse var-cov matrix
		if (objective == "mahal_dist"){
			if (p < n){
				for (j in 1 : p){
					.jcall(java_obj, "V", "setInvVarCovRow", as.integer(j - 1), SinvX[j, , drop = FALSE]) #java indexes from 0...n-1
				}
			}
		}
	}
	
	#do we want diagnostics? Set it...
	if (diagnostics){
		.jcall(java_obj, "V", "setDiagnostics")
	}
	
	#is it semigreedy? Set it...
	if (semigreedy){
		.jcall(java_obj, "V", "setSemigreedy")
	}
		
	#now return information as an object (just a list)
	greedy_experimental_design_search = list()
	greedy_experimental_design_search$max_designs = max_designs
	greedy_experimental_design_search$semigreedy = semigreedy
	greedy_experimental_design_search$start = start
	greedy_experimental_design_search$wait = wait
	greedy_experimental_design_search$diagnostics = diagnostics
	greedy_experimental_design_search$verbose = verbose
	greedy_experimental_design_search$X = X
	greedy_experimental_design_search$n = n
	greedy_experimental_design_search$p = p
	greedy_experimental_design_search$objective = objective
	greedy_experimental_design_search$java_obj = java_obj
	class(greedy_experimental_design_search) = "greedy_experimental_design_search"
	#if the user wants to run it immediately...
	if (start){
		startSearch(greedy_experimental_design_search)
	}
	#return the final object
	greedy_experimental_design_search
}

#' Prints a summary of a \code{greedy_experimental_design_search} object
#' 
#' @param x			The \code{greedy_experimental_design_search} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print greedy_experimental_design_search
#' @export
print.greedy_experimental_design_search = function(x, ...){
	progress = greedySearchCurrentProgress(x)
	time_elapsed = searchTimeElapsed(x)
	if (progress == 0){
		cat("No progress on the GreedyExperimentalDesign. Did you run \"startSearch?\"\n")
	} else if (progress == x$max_designs){
		cat("The search completed in", time_elapsed, "seconds.", progress, "vectors have been found.\n")
	} else {
		cat("The search has found ", progress, " vectors thus far (", round(progress / x$max_designs * 100), "%) in ", time_elapsed," seconds.\n", sep = "")
	}
}

#' Prints a summary of a \code{greedy_experimental_design_search} object
#' 
#' @param object		The \code{greedy_experimental_design_search} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary greedy_experimental_design_search
#' @export
summary.greedy_experimental_design_search = function(object, ...){
	print(object, ...)
}

#' Plots a summary of a greedy search object object
#' 
#' @param x			The greedy search object object to be summarized in the plot
#' @param ...		Other parameters to pass to the default plot function
#' @return			An array of order statistics from \link{plot_obj_val_order_statistic} as a list element
#' 
#' @author 			Adam Kapelner
#' @method plot greedy_experimental_design_search
#' @export
plot.greedy_experimental_design_search = function(x, ...){
	progress = greedySearchCurrentProgress(x)
	res = resultsGreedySearch(x, max_vectors = 2)
	breaks = max(1, round(progress / 10))
	hist_data = data.frame(obj_val = res$obj_vals_orig_order)
	hist_plot = ggplot2::ggplot(hist_data, ggplot2::aes(x = obj_val)) +
		ggplot2::geom_histogram(bins = breaks, color = "black", fill = "grey70") +
		ggplot2::labs(x = "objective value", y = NULL, title = paste("After", progress, "searches")) +
		ggplot2::theme_classic()
#	hist(res$num_iters, br = progress / 10, xlab = "# of search iterations", ylab = NULL, main = "")
	
	#now do the plot of number of searches needed
	order_plot = build_obj_val_order_statistic_plot(x)
	draw_plot_pair(hist_plot, order_plot$plot)
	invisible(list(val_order_stats = order_plot$val_order_stats))
}

#' Plots the objective value by iteration
#' 
#' @param res 		Results from a greedy search object
#' @param runs 		A vector of run indices you would like to see plotted (default is to plot the first up to 9)
#' 
#' @author 			Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   diagnostics = TRUE,
#'   objective = "abs_sum_diff",
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsGreedySearch(ged, max_vectors = 2)
#' plot_obj_val_by_iter(res)
#' }
#' @export
plot_obj_val_by_iter = function(res, runs = NULL){
	if (is.null(res$obj_val_by_iters)){
		stop("You need to set diagnostics = TRUE on the search object.")
	}
	
	if (is.null(runs)){
		runs = 1 : min(length(res$obj_val_by_iters), 9)
	}
	num_to_plot = length(runs)
	
	plot_data = data.frame(run = character(0), iteration = integer(0), obj_val = numeric(0))
	for (run in runs){
		obj_vals = res$obj_val_by_iters[[run]]
		if (length(obj_vals) == 0){
			next
		}
		plot_data = rbind(plot_data, data.frame(
			run = paste("Run #", run),
			iteration = seq_along(obj_vals),
			obj_val = obj_vals
		))
	}
	plot_data$run = factor(plot_data$run, levels = paste("Run #", runs))
	if (nrow(plot_data) == 0){
		blank_plot = build_blank_plot("iteration", "objective value", title = NULL)
		print(blank_plot)
		return(invisible(NULL))
	}
	plot_data$line_ok = ave(plot_data$iteration, plot_data$run, FUN = function(x) length(x) > 1)
	iter_plot = ggplot2::ggplot(plot_data, ggplot2::aes(x = iteration, y = obj_val)) +
		ggplot2::geom_line(data = plot_data[plot_data$line_ok, , drop = FALSE]) +
		ggplot2::geom_point() +
		ggplot2::facet_wrap(~run, ncol = ceiling(sqrt(num_to_plot))) +
		ggplot2::labs(x = "iteration", y = "objective value") +
		ggplot2::theme_classic()
	print(iter_plot)
} 

#' Plots an order statistic of the object value as a function of number of searches
#' 
#' @param obj			The greedy search object object whose search history is to be visualized
#' @param order_stat 	The order statistic that you wish to plot. The default is \code{1} for the minimum.
#' @param skip_every	Plot every nth point. This makes the plot generate much more quickly. The default is \code{5}.
#' @param type			The type parameter for plot.
#' @param ... 			Other arguments to be passed to the plot function.
#' @return 				An array of order statistics as a list element
#' 
#' @author 				Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' plot_obj_val_order_statistic(ged, order_stat = 1, skip_every = 1)
#' }
#' @export
plot_obj_val_order_statistic = function(obj, order_stat = 1, skip_every = 5, type = "o", ...){
	order_plot = build_obj_val_order_statistic_plot(obj, order_stat = order_stat, skip_every = skip_every, type = type, ...)
	print(order_plot$plot)
	invisible(list(val_order_stats = order_plot$val_order_stats))	
}

build_blank_plot = function(xlab, ylab, title = NULL, xlim = NULL, ylim = NULL){
	blank_data = data.frame(x = 0, y = 0)
	blank_plot = ggplot2::ggplot(blank_data, ggplot2::aes(x = x, y = y)) +
		ggplot2::geom_blank() +
		ggplot2::labs(x = xlab, y = ylab, title = title) +
		ggplot2::theme_classic()
	if (!is.null(xlim) || !is.null(ylim)){
		blank_plot = blank_plot + ggplot2::coord_cartesian(xlim = xlim, ylim = ylim)
	}
	blank_plot
}

draw_plot_pair = function(left_plot, right_plot){
	grid::grid.newpage()
	grid::pushViewport(grid::viewport(layout = grid::grid.layout(1, 2)))
	print(left_plot, vp = grid::viewport(layout.pos.row = 1, layout.pos.col = 1))
	print(right_plot, vp = grid::viewport(layout.pos.row = 1, layout.pos.col = 2))
}

build_obj_val_order_statistic_plot = function(obj, order_stat = 1, skip_every = 5, type = "o", ...){
	progress = greedySearchCurrentProgress(obj)
	ylab = paste("objective value (", order_stat, ")", sep = "")
	dots = list(...)
	title = NULL
	if (!is.null(dots$main)){
		title = dots$main
	}
	xlim = dots$xlim
	ylim = dots$ylim
	plot_color = NULL
	if (!is.null(dots$col)){
		plot_color = dots$col
	} else if (!is.null(dots$color)){
		plot_color = dots$color
	}
	line_size = dots$lwd
	point_shape = dots$pch
	point_size = NULL
	if (!is.null(dots$cex)){
		point_size = dots$cex * 1.5
	}
	if (progress <= 0 || progress < order_stat){
		blank_plot = build_blank_plot("Number of Searches", ylab, title = title, xlim = xlim, ylim = ylim)
		return(list(plot = blank_plot, val_order_stats = numeric(0)))
	}
	res = resultsGreedySearch(obj, max_vectors = 1)	#don't need many vectors; keep minimal for speed concerns
	vals = res$obj_vals_orig_order
	val_order_stats = array(NA, progress)
	skip_every = max(1, skip_every)
	indices = seq(order_stat, progress, by = skip_every)
	if (length(indices) == 0){
		indices = progress
	} else if (tail(indices, 1) != progress){
		indices = c(indices, progress)
	}
	for (d in indices){
		vals_d = vals[1 : d]
		vals_d = vals_d[is.finite(vals_d)]
		if (length(vals_d) == 0){
			next
		}
		if (order_stat == 1){
			val_order_stats[d] = min(vals_d)
		} else if (order_stat <= length(vals_d)){
			val_order_stats[d] = sort(vals_d)[order_stat]
		}
	}
	if (!any(is.finite(val_order_stats))){
		blank_plot = build_blank_plot("Number of Searches", ylab, title = title, xlim = xlim, ylim = ylim)
		return(list(plot = blank_plot, val_order_stats = val_order_stats))
	}
	plot_data = data.frame(searches = seq_len(progress), value = val_order_stats)
	plot_data = plot_data[is.finite(plot_data$value), , drop = FALSE]
	order_plot = ggplot2::ggplot(plot_data, ggplot2::aes(x = searches, y = value)) +
		ggplot2::labs(x = "Number of Searches", y = ylab, title = title) +
		ggplot2::theme_classic()
	type = tolower(type)
	draw_line = grepl("l", type) || grepl("o", type) || grepl("b", type) || grepl("c", type)
	draw_points = grepl("p", type) || grepl("o", type) || grepl("b", type)
	if (identical(type, "n")){
		draw_line = FALSE
		draw_points = FALSE
	}
	if (draw_line && nrow(plot_data) > 1){
		line_args = list(na.rm = TRUE)
		if (!is.null(plot_color)){
			line_args$color = plot_color
		}
		if (!is.null(line_size)){
			line_args$size = line_size
		}
		order_plot = order_plot + do.call(ggplot2::geom_line, line_args)
	}
	if (draw_points && nrow(plot_data) > 0){
		point_args = list(na.rm = TRUE)
		if (!is.null(plot_color)){
			point_args$color = plot_color
		}
		if (!is.null(point_size)){
			point_args$size = point_size
		}
		if (!is.null(point_shape)){
			point_args$shape = point_shape
		}
		order_plot = order_plot + do.call(ggplot2::geom_point, point_args)
	}
	if ((nrow(plot_data) == 0) || (!draw_line && !draw_points)){
		order_plot = order_plot + ggplot2::geom_blank()
	}
	if (!is.null(xlim) || !is.null(ylim)){
		order_plot = order_plot + ggplot2::coord_cartesian(xlim = xlim, ylim = ylim)
	}
	list(plot = order_plot, val_order_stats = val_order_stats)	
}

# Returns the number of vectors found by the greedy design search
# 
# @param obj 		The \code{greedy_experimental_design} object that is currently running the search
# 
# @author Adam Kapelner
greedySearchCurrentProgress = function(obj){
	.jcall(obj$java_obj, "I", "progress")
}


#' Returns the results (thus far) of the greedy design search
#' 
#' @param obj 			The \code{greedy_experimental_design} object that is currently running the search
#' @param max_vectors	The number of design vectors you wish to return. \code{NULL} returns all of them. 
#' 						This is not recommended as returning over 1,000 vectors is time-intensive. The default is 9. 
#' @param form			Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsGreedySearch(ged, max_vectors = 2)
#' res$obj_vals
#' }
#' @export
resultsGreedySearch = function(obj, max_vectors = 9, form = "one_zero"){
	if (!is.null(max_vectors)){
		assertCount(max_vectors, positive = TRUE)
	}
	obj_vals = .jcall(obj$java_obj, "[D", "getObjectiveVals")
	num_iters = .jcall(obj$java_obj, "[I", "getNumIters")
	#these two are in order, so let's order the indicTs by the final objective values
	ordered_indices = order(obj_vals)
	num_completed = length(obj_vals)
	if (is.null(max_vectors)){
		last_index = num_completed
	} else {
		last_index = min(max_vectors, num_completed)
	}
	ending_indicTs = NULL
	starting_indicTs = NULL
	switches = NULL
	xbarj_diffs = NULL
	obj_val_by_iters = NULL
	pct_vec_same = NULL
	ordered_java_indices = integer(0)
	ordered_java_indices_j = NULL
	if (last_index > 0){
		ordered_java_indices = as.integer(ordered_indices[1 : last_index] - 1)
		ordered_java_indices_j = .jarray(ordered_java_indices)
		ending_indicTs = .jcall(obj$java_obj, "[[I", "getEndingIndicTs", ordered_java_indices_j, simplify = TRUE)
		if (form == "pos_one_min_one"){
			if (length(ending_indicTs) > 0 && min(ending_indicTs) >= 0 && max(ending_indicTs) <= 1){
				ending_indicTs = (ending_indicTs - 0.5) * 2
			}
		} else if (form == "one_zero" && length(ending_indicTs) > 0 && min(ending_indicTs) < 0){
			ending_indicTs = (ending_indicTs + 1) / 2
		}
	}
	if (obj$diagnostics && last_index > 0){
		starting_indicTs = .jcall(obj$java_obj, "[[I", "getStartingIndicTs", ordered_java_indices_j, simplify = TRUE)
		if (form == "pos_one_min_one"){
			if (length(starting_indicTs) > 0 && min(starting_indicTs) >= 0 && max(starting_indicTs) <= 1){
				starting_indicTs = (starting_indicTs - 0.5) * 2
			}
		} else if (form == "one_zero" && length(starting_indicTs) > 0 && min(starting_indicTs) < 0){
			starting_indicTs = (starting_indicTs + 1) / 2
		}
		switches = .jcall(obj$java_obj, "[[[I", "getSwitchedPairs", ordered_java_indices_j, simplify = TRUE)
		xbarj_diffs = .jcall(obj$java_obj, "[[[D", "getXbarjDiffs", ordered_java_indices_j, simplify = TRUE)
		obj_val_by_iters = .jcall(obj$java_obj, "[[D", "getObjValByIter", ordered_java_indices_j, simplify = TRUE)
		
		pct_vec_same = colSums(starting_indicTs == ending_indicTs) / length(starting_indicTs[, 1]) * 100
	}
	greedy_experimental_design_search_results = list(
		ordered_indices = ordered_indices,
		obj_vals = obj_vals[ordered_indices], 
		obj_vals_orig_order = obj_vals,
		obj_vals_unordered = obj_vals,
		num_iters = num_iters[ordered_indices], 
		orig_order = ordered_indices, 
		ending_indicTs = ending_indicTs,
		starting_indicTs = starting_indicTs,
		obj_val_by_iters = obj_val_by_iters,
		pct_vec_same = pct_vec_same,
		switches = switches,
		xbarj_diffs = xbarj_diffs,
		last_index = last_index
	)
	class(greedy_experimental_design_search_results) = "greedy_experimental_design_search_results"
	#return the final object
	greedy_experimental_design_search_results
}

#' Compute Optimal Number of Treatments/Controls
#' 
#' Given a total budget and asymmetric treatment and control costs, calculate the
#' number of treatments and controls that optimize the variance of the estimator. 
#' The number of treatments is rounded up by default.
#' 
#' @param c_treatment		The cost of a treatment assignment. Default is \code{NULL} for symmetric costs.
#' @param c_control			The cost of a control assignment. Default is \code{NULL} for symmetric costs.
#' @param c_total_max		The total cost constraint of any allocation. Either this or \code{n} must be specified. Default is \code{NULL}.
#' @param n					The total cost constraint as specified by the total number of subjects. Either this or \code{c_total} must be 
#' 							specified. Default is \code{NULL}.
#' 
#' @return					A list with three keys: n, nT, nC plus specified arguments
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' optimize_asymmetric_treatment_assignment(n = 100)
#' optimize_asymmetric_treatment_assignment(n = 100, c_treatment = 2, c_control = 1)
#' optimize_asymmetric_treatment_assignment(c_total_max = 50, c_treatment = 2, c_control = 1)
#' }
#' @export
optimize_asymmetric_treatment_assignment = function(
		c_treatment = NULL,
		c_control = NULL,
		c_total_max = NULL,
		n = NULL
	){
	if ((is.null(c_total_max) & is.null(n)) | (!is.null(c_total_max) & !is.null(n))){
		stop("n xor c_total must be specified.")
	}
	if (!is.null(n)){
		if (is.null(c_treatment) & is.null(c_control)){
			nT = n / 2
			list(n = n, nT = nT, nC = n - nT)
		} else {
			checkNumeric(c_treatment, lower = .Machine$double.xmin, finite = TRUE)
			checkNumeric(c_control,   lower = .Machine$double.xmin, finite = TRUE)
			nT = ceiling(n * c_treatment / (c_treatment + c_control))
		 	nC = n - nT
		 	c_total = nT * c_treatment + nC * c_control
		 	list(n = n, nT = nT, nC = nC, c_treatment = c_treatment, c_control = c_control, c_total = c_total)
		}
	} else {
		checkNumeric(c_treatment, lower = .Machine$double.xmin, finite = TRUE)
		checkNumeric(c_control,   lower = .Machine$double.xmin, finite = TRUE)
		
		K_const = c_total_max / c_control
		c_const = c_treatment / c_control
		nT = K_const / (c_const + sqrt(c_const))
		nC = K_const - c_const * nT
		nT = floor(nT)
		nC = if (c_const * nT + ceiling(nC) <= K_const){
					ceiling(nC)
				} else {
					floor(nC)
				}
		c_total = nT * c_treatment + nC * c_control
		list(n = nT + nC, nT = nT, nC = nC, c_total = c_total, c_total_max = c_total_max, c_treatment = c_treatment, c_control = c_control)
	}
}

#a set of the valid kernel types for the method build_kpca_object
VALID_KERNEL_TYPES = c("vanilla", "rbf", "poly", "tanh", "bessel", "laplace", "anova", "spline")



#' Gram Matrix Computation
#' 
#' Computes the Gram Matrix for a user-specified kernel using the library \code{kernlab}. Note that 
#' this function automatically standardizes the columns of the data entered.
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more optimal design.
#' @param kernel_type 		One of the following: "vanilla", "rbf", "poly", "tanh", "bessel", "laplace", 
#' 							"anova" or "spline".
#' @param params			A vector of numeric parameters. Each \code{kernel_type} has different numbers of
#' 							parameters required. For more information see documentation for the \code{kernlab} library.
#' @return					The \code{n x n} gram matrix for the given kernel on the given data. 
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(12), nrow = 6)
#' K = compute_gram_matrix(X, kernel_type = "rbf", params = 0.5)
#' dim(K)
#' }
#' @export
compute_gram_matrix = function(X, kernel_type, params = c()){
	#ensure that the kernel_type is valid
	if (!(kernel_type %in% VALID_KERNEL_TYPES)){
		stop("The kernel type must be one of the following:", paste(VALID_KERNEL_TYPES, collapse = ", "))
	}
	
	#now build the the correct kernel based on the kernel type. This is a giant switch statement
	if (kernel_type == "vanilla"){
		if (length(params) != 0){
			stop("vanilla kernels do not take parameters")
		}
		kernel = kernlab::vanilladot()
	} else if (kernel_type == "rbf"){
		if (length(params) != 1){
			stop("gaussian radial basis function kernels require one parameter: gamma")
		}
		kernel = kernlab::rbfdot(params[1])
	} else if (kernel_type == "poly"){
		if (length(params) != 3){
			stop("polynomial kernels require three parameters: degree, scale and offset in that order")
		}
		kernel = kernlab::polydot(params[1], params[2])
	} else if (kernel_type == "tanh"){
		if (length(params) != 2){
			stop("hyperbolic tangent kernels require two parameters: scale and offset in that order")
		}
		kernel = kernlab::tanhdot(params[1], params[2])
	} else if (kernel_type == "bessel"){
		if (length(params) != 3){
			stop("bessel kernels require three parameters: gamma, order and degree in that order")
		}
		kernel = kernlab::besseldot(params[1], params[2], params[3])
	} else if (kernel_type == "laplace"){
		if (length(params) != 1){
			stop("laplace kernels require one parameter: gamma")
		}
		kernel = kernlab::laplacedot(params[1])
	} else if (kernel_type == "anova"){
		if (length(params) != 2){
			stop("anova kernels require two parameters: gamma and degree in that order")
		}
		kernel = kernlab::anovadot(params[1], params[2])
	} else if (kernel_type == "spline"){
		if (length(params) != 0){
			stop("spline kernels do not take parameters")
		}
		kernel = kernlab::splinedot()
	}
	
	#compute the Gram matrix as type matrix
	kernlab::kernelMatrix(kernel, standardize_data_matrix(X))
}


#private
setGramMatrix = function(java_obj, Kgram){
	for (i in 1 : nrow(Kgram)){
		.jcall(java_obj, "V", "setKgramRow", as.integer(i - 1), Kgram[i, , drop = FALSE]) #java indexes from 0...n-1
	}
}

#' Begin Gurobi Optimized Search
#' 
#' This method creates an object of type optimal_experimental_design and will immediately initiate
#' a search through allocation space for forced balance designs. Make sure you setup Gurobi properly first. This means
#' applying for a license, downloading, installing, registering it on your computer using the \code{grbgetkey} command
#' with the license file in the default directory. Then, in R, install the package from the file in your gurobi directory. 
#' 
#' Currently, this method does not return multiple vectors. This will be improved in a later 
#' version. If you want this functionality now, use the hacked-up method \code{gurobi_multiple_designs}.
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more optimal design.
#' @param objective			The objective function to use when searching design space. This is a string
#' 							with valid values "\code{mahal_dist}" (the default) or "\code{kernel}".
#' @param Kgram				If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
#' 							entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.
#' @param initial_time_limit_sec	The maximum amount of time the optimizer can run for in seconds. The default is \code{5 * 60}.
#' @param restart_time_limit_sec	The maximum amount of time each restart can run for in seconds. The default is \code{60}.
#' @param max_number_of_restarts		The maximum number of restarts to attempt if too few unique solutions are returned.
#' 								Default is \code{0}.
#' @param max_no_good_cuts			The maximum number of no-good cuts to attempt. Default is \code{0} (disabled).
#' @param num_cores			Number of cores to use during search. Default is \code{2}.
#' @param w_0				The initial starting location (optional).
#' @param gurobi_params		A list of optional parameters to be passed to Gurobi (see their documentation online).
#' @param verbose			Should Gurobi log to console? Default is \code{TRUE}.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' @param r					Number of solution vectors to request from the Gurobi pool.
#' @param pool_solutions 	Number of solutions to request from the Gurobi pool. Defaults to \code{10 * r}.
#' @param pool_gap 			Relative optimality gap for the pool. Default is \code{0.2}. Use \code{NULL} to skip.
#' @param pool_gap_abs 		Absolute optimality gap for the pool. Default is \code{NULL} to skip.
#' @param pool_search_mode	Solution pool search mode. Default is \code{2} for diverse solutions.
#' @param mip_gap			Relative MIP gap target (stops when \code{|best-bound - best-incumbent| / |best-incumbent| <= mip_gap}).
#' 							Lower values force deeper search. Default is \code{1e-4}.
#' @param mip_gap_abs		Absolute MIP gap target (stops when \code{|best-bound - best-incumbent| <= mip_gap_abs}).
#' 							Lower values force deeper search. Default is \code{1e-10}.
#' @param mip_focus			Search focus: \code{0} (balance), \code{1} (find feasible solutions), \code{2} (prove optimality),
#' 							\code{3} (bound improvement). Default is \code{1}.
#' @param heuristics			Heuristics effort in \code{[0,1]} where higher values spend more time on heuristics.
#' 							Default is \code{0.2}.
#' @param cuts				Cut aggressiveness: \code{-1} (automatic), \code{0} (off), \code{1} (conservative),
#' 							\code{2} (aggressive), \code{3} (very aggressive). Default is \code{2}.
#' @param presolve			Presolve aggressiveness: \code{-1} (automatic), \code{0} (off), \code{1} (conservative),
#' 							\code{2} (aggressive). Default is \code{2}.
#' @return 					A list object which houses the results from Gurobi. Depending on the \code{gurobi_parms},
#' 							the data within will be different. The most relevant tags are \code{x} for the best found solution and \code{objval}
#' 							for the object
#' 
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' if ("gurobi" %in% loadedNamespaces()) {
#'   set.seed(1)
#'   X = matrix(rnorm(12), nrow = 6)
#'   gobj = initGurobiNumericalOptimizationExperimentalDesignObject(
#'     X,
#'     r = 2,
#'     num_cores = 1,
#'     initial_time_limit_sec = 5,
#'     verbose = FALSE
#'   )
#'   gobj$n
#' }
#' }
#' @export
initGurobiNumericalOptimizationExperimentalDesignObject = function(
		X = NULL, 
		objective = "mahal_dist",
		Kgram = NULL,
		num_cores = 2, 
		w_0 = NULL,
		initial_time_limit_sec = 5 * 60,
		restart_time_limit_sec = 60,
		max_number_of_restarts = 0,
		max_no_good_cuts = 0,
		verbose = TRUE,
		gurobi_params = list(),
		use_safe_inverse = FALSE,
		r,
		pool_solutions = NULL,
		pool_gap = 0.2,
		pool_gap_abs = NULL,
		pool_search_mode = 2,
		mip_gap = 1e-4,
		mip_gap_abs = 1e-10,
		mip_focus = 1,
		heuristics = 0.2,
		cuts = 2,
		presolve = 2){
	
	gurobi_ns = get_gurobi_namespace()
	assertCount(num_cores, positive = TRUE)
	assertNumeric(initial_time_limit_sec, lower = 1)
	assertNumeric(restart_time_limit_sec, lower = 1)
	assertCount(max_number_of_restarts, positive = FALSE)
	assertNumeric(max_no_good_cuts, lower = 0)
	if (is.finite(max_no_good_cuts) && max_no_good_cuts %% 1 != 0) {
		stop("\"max_no_good_cuts\" must be an integer or Inf.")
	}
	assertLogical(verbose)
	assertClass(gurobi_params, "list")
	assertLogical(use_safe_inverse)
	assertCount(r, positive = TRUE)
	assertCount(pool_solutions, positive = TRUE, null.ok = TRUE)
	assertNumeric(pool_search_mode, lower = 0)
	assertNumeric(pool_gap, lower = 0, null.ok = TRUE)
	assertNumeric(pool_gap_abs, lower = 0, null.ok = TRUE)
	assertNumeric(mip_gap, lower = 0)
	assertNumeric(mip_gap_abs, lower = 0)
	assertNumeric(mip_focus, lower = 0, upper = 3)
	assertNumeric(heuristics, lower = 0, upper = 1)
	assertNumeric(cuts, lower = -1, upper = 3)
	assertNumeric(presolve, lower = -1, upper = 2)
	if (!is.null(w_0)){
		assertSetEqual(unique(w_0), c(1, -1))
		assertTRUE(sum(w_0) == 0)
	}
	
	gurobi_params$LogToConsole = as.numeric(verbose)
	gurobi_params$Threads = num_cores	
	gurobi_params$TimeLimit = initial_time_limit_sec
	if (is.null(gurobi_params$MIPGap)) {
		gurobi_params$MIPGap = mip_gap
	}
	if (is.null(gurobi_params$MIPGapAbs)) {
		gurobi_params$MIPGapAbs = mip_gap_abs
	}
	if (is.null(gurobi_params$MIPFocus)) {
		gurobi_params$MIPFocus = mip_focus
	}
	if (is.null(gurobi_params$Heuristics)) {
		gurobi_params$Heuristics = heuristics
	}
	if (is.null(gurobi_params$Cuts)) {
		gurobi_params$Cuts = cuts
	}
	if (is.null(gurobi_params$Presolve)) {
		gurobi_params$Presolve = presolve
	}
	if (is.null(pool_solutions)) {
		pool_solutions = 10 * r
	}
	if (!is.null(pool_solutions) && is.null(gurobi_params$PoolSolutions)) {
		gurobi_params$PoolSolutions = pool_solutions
	}
	if (!is.null(pool_solutions) && is.null(gurobi_params$PoolSearchMode)) {
		gurobi_params$PoolSearchMode = pool_search_mode
	}
	if (!is.null(pool_solutions) && !is.null(pool_gap) && is.null(gurobi_params$PoolGap)) {
		gurobi_params$PoolGap = pool_gap
	}
	if (!is.null(pool_solutions) && !is.null(pool_gap_abs) && is.null(gurobi_params$PoolGapAbs)) {
		gurobi_params$PoolGapAbs = pool_gap_abs
	}
	
	verify_objective_function(objective, Kgram, n)
	
	if (!is.null(Kgram)){
		n = nrow(Kgram)
		p = NA
	} else {
		n = nrow(X)
		p = ncol(X)
	}
	if (n %% 2 != 0){
		stop("Design matrix must have even rows to have equal treatments and controls")
	}
	
	if (objective == "abs_sum_diff"){
		stop("The \"abs_sum_diff\" objective type does not work with Gurobi.")
	}
	SinvX = NULL
	
	# Let Q = X^T SinvX X =or= K
	if (objective == "mahal_dist"){
		if (p < n){
			if (use_safe_inverse){
				SinvX = safe_cov_inverse(X)
			} else {
				SinvX = solve(stats::var(X))
			}
		}
		Q = X %*% SinvX %*% t(X)
	} else if (!is.null(Kgram)){
		Q = Kgram
	}
	one_vec = rep(1, n)
	
	#   minimize_w w^T Q w = (2b - 1)^T Q (2b - 1) = 4b^T Q b - 2 1^T Q b - 2 b^T Q 1 + 1^T Q 1 = b^T 4Q b - b^T 4Q 1 + 1^T Q 1 
	# = minimize_b b^T 4Q b - b^T 4Q 1 + 1^T Q 1 
	# minimize
	#        [b_1 b_2 ... b_n]^T Q [b_1 b_2 ... b_n] - 1^T Q [b_1 b_2 ... b_n]
	# subject to
	#        b_1 + b_2 + ... + b_n = n / 2,
	#        b_1, b_2, ..., b_n binary
	
	model 		= list()
	#quadratic component
	model$Q     = 8 * Q
	model$obj   = as.numeric(-4 * (Q %*% one_vec))
	model$ObjCon = as.numeric(t(as.matrix(one_vec)) %*% Q %*% as.matrix(one_vec)) #this will allow the objective value to be interpretable during verbose printouts
	#constraint component
	model$A     = t(as.matrix(one_vec))
	model$rhs   = n / 2
	model$sense = "="
	model$vtype = "B"
	if (!is.null(w_0)){
		model$start = w_0
	}

	#now return information as an object (just a list)
	gurobi_numerical_optimization_experimental_design_search = list()
	gurobi_numerical_optimization_experimental_design_search$X = X
	gurobi_numerical_optimization_experimental_design_search$SinvX = SinvX
	gurobi_numerical_optimization_experimental_design_search$n = n
	gurobi_numerical_optimization_experimental_design_search$p = p
	gurobi_numerical_optimization_experimental_design_search$Kgram = Kgram
	gurobi_numerical_optimization_experimental_design_search$objective = objective
	gurobi_numerical_optimization_experimental_design_search$w_0 = w_0
	gurobi_numerical_optimization_experimental_design_search$gurobi_params = gurobi_params
	gurobi_numerical_optimization_experimental_design_search$use_safe_inverse = use_safe_inverse
	gurobi_numerical_optimization_experimental_design_search$r = r
	gurobi_numerical_optimization_experimental_design_search$pool_solutions = pool_solutions
	gurobi_numerical_optimization_experimental_design_search$pool_gap = pool_gap
	gurobi_numerical_optimization_experimental_design_search$pool_search_mode = pool_search_mode
	gurobi_numerical_optimization_experimental_design_search$restart_time_limit_sec = restart_time_limit_sec
	gurobi_numerical_optimization_experimental_design_search$max_number_of_restarts = max_number_of_restarts
	gurobi_numerical_optimization_experimental_design_search$max_no_good_cuts = max_no_good_cuts
	gurobi_numerical_optimization_experimental_design_search$model = model
	class(gurobi_numerical_optimization_experimental_design_search) = "gurobi_numerical_optimization_experimental_design_search"
	#run the optimization and return the final object	
	return(c(gurobi_numerical_optimization_experimental_design_search, gurobi_call(gurobi_ns, model, gurobi_params)))

# 	#we are about to construct a GurobiNumericalOptimizeExperimentalDesign java object. First, let R garbage collect
# 	#to clean up previous objects that are no longer in use. This is important
# 	#because R's garbage collection system does not "see" the size of Java objects. Thus,
# 	#you are at risk of running out of memory without this invocation. 
# 	gc() #Delete at your own risk!	
	
# 	#now go ahead and create the Java object and set its information
# 	error_obj = NULL
# 	java_obj = .jnew("GurobiNumericalOptimizeExperimentalDesign.GurobiNumericalOptimizeExperimentalDesign")
# 	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))
# 	.jcall(java_obj, "V", "setN", as.integer(n))
# 	if (objective != "kernel"){
# 		p = ncol(X)
# 		.jcall(java_obj, "V", "setP", as.integer(p))
# 	}
# #	cat("time limit min: ", as.numeric(time_limit_min), "\n")
# 	.jcall(java_obj, "V", "setObjective", objective)
# 	.jcall(java_obj, "V", "setTimeLimitMin", as.numeric(time_limit_sec / 60))
# #	if (!is.null(node_limit)){
# #		if (node_limit <= 1){
# #			stop("Node limit must be one or more.")
# #		}
# #		.jcall(java_obj, "V", "setNodeLimit", as.numeric(round(node_limit)))
# #	}	
# #	if (!is.null(max_solutions)){
# #		if (max_solutions < 1){
# #			stop("Max solutions must be one or more.")
# #		}
# #		.jcall(java_obj, "V", "setMaxSolutions", as.integer(round(max_solutions)))
# #	}	
	
# 	if (!verbose){
# 		.jcall(java_obj, "V", "turnGurobiLogOff")
# 	}
# #	.jcall(java_obj, "V", "setLogFilename", log_file)
	
# 	#feed in the gram matrix if applicable
# 	if (!is.null(Kgram)){
# 		setGramMatrix(java_obj, Kgram)
# 	} else {
# 		#feed in the raw data
# 		for (i in 1 : n){	
# 			.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), X[i, , drop = FALSE]) #java indexes from 0...n-1
# 		}
		
# 		#feed in the inverse var-cov matrix
# 		for (j in 1 : p){
# 			.jcall(java_obj, "V", "setInvVarCovRow", as.integer(j - 1), SinvX[j, , drop = FALSE]) #java indexes from 0...n-1
# 		}
# 	}
	
# 	#now return information as an object (just a list)
# 	gurobi_numerical_optimization_experimental_design_search = list()
# 	gurobi_numerical_optimization_experimental_design_search$X = X
# 	gurobi_numerical_optimization_experimental_design_search$SinvX = SinvX
# 	gurobi_numerical_optimization_experimental_design_search$n = n
# 	gurobi_numerical_optimization_experimental_design_search$p = p
# 	gurobi_numerical_optimization_experimental_design_search$Kgram = Kgram
# 	gurobi_numerical_optimization_experimental_design_search$objective = objective
# #	gurobi_numerical_optimization_experimental_design_search$max_solutions = max_solutions
# 	gurobi_numerical_optimization_experimental_design_search$time_limit_sec = time_limit_sec
# 	gurobi_numerical_optimization_experimental_design_search$java_obj = java_obj
# 	class(gurobi_numerical_optimization_experimental_design_search) = "gurobi_numerical_optimization_experimental_design_search"
	
# 	#now start search
# 	startSearch(gurobi_numerical_optimization_experimental_design_search)
	
# 	#return the final object
# 	gurobi_numerical_optimization_experimental_design_search
}




# #' Find multiple designs using Gurobi
# #' 
# #' This method searches through allocation space using Gurobi's optimization many times.
# #' It finds many different solutions by permuting the rows of the design matrix and 
# #' rerunning the optimization.
# #' 
# #' @param X 		The design matrix with $n$ rows (one for each subject) and $p$ columns 
# #' 					(one for each measurement on the subject). This is the design matrix you wish 
# #' 					to search for a more optimal design.
# #' @param r 				The number of vectors that should be returned.
# #' @param pool_solutions 	Number of solutions to request from the Gurobi pool. Defaults to \code{r}.
# #' @param pool_gap 		Relative optimality gap for the pool. Default is \code{0.1}. Use \code{NULL} to skip.
# #' @param pool_search_mode	Solution pool search mode. Default is \code{2} for diverse solutions.
# #' @param diverse			Should the returned solutions be greedily diversified by Hamming distance? Default is \code{TRUE}.
# #' @param ... 				Additional arguments to be passed to \code{initGurobiNumericalOptimizationExperimentalDesignObject}.
# #' @return			A matrix of allocation vectors of dimension \code{r x n}.	 
# #' 
# #' @author Kapelner
# #' @examples
# #' \dontrun{
# #' if ("gurobi" %in% loadedNamespaces()) {
# #'   set.seed(1)
# #'   X = matrix(rnorm(12), nrow = 6)
# #'   indicTs = gurobi_multiple_designs(
# #'     X,
# #'     r = 2,
# #'     num_cores = 1,
# #'     initial_time_limit_sec = 5,
# #'     verbose = FALSE
# #'   )
# #'   dim(indicTs)
# #' }
# #' }
# #' @export
# gurobi_multiple_designs = function(X, r, pool_solutions = NULL, pool_gap = 0.1, pool_search_mode = 2, diverse = TRUE, ...){
# 	n = nrow(X)
# 	assertCount(r, positive = TRUE)
# 	assertLogical(diverse)
# 	if (is.null(pool_solutions)){
# 		pool_solutions = r
# 	}
# 	assertCount(pool_solutions, positive = TRUE)
# 	assertNumeric(pool_search_mode, lower = 0)
# 	assertNumeric(pool_gap, lower = 0, null.ok = TRUE)

# 	select_diverse_rows = function(mat, k){
# 		if (nrow(mat) <= k){
# 			return(seq_len(nrow(mat)))
# 		}
# 		selected = 1L
# 		remaining = setdiff(seq_len(nrow(mat)), selected)
# 		while (length(selected) < k){
# 			min_dist = vapply(remaining, function(i){
# 				min(rowSums(abs(mat[i, , drop = FALSE] - mat[selected, , drop = FALSE])))
# 			}, numeric(1))
# 			pick = remaining[which.max(min_dist)]
# 			selected = c(selected, pick)
# 			remaining = setdiff(remaining, pick)
# 		}
# 		selected
# 	}

# 	dots = list(...)
# 	gnoed = do.call(
# 		initGurobiNumericalOptimizationExperimentalDesignObject,
# 		c(
# 			list(
# 				X,
# 				r = r,
# 				pool_solutions = pool_solutions,
# 				pool_gap = pool_gap,
# 				pool_search_mode = pool_search_mode
# 			),
# 			dots
# 		)
# 	)
# 	res = resultsGurobiNumericalOptimizeSearch(gnoed)
# 	indicTs = res$indicTs
# 	if (is.null(dim(indicTs))){
# 		indicTs = matrix(indicTs, nrow = 1)
# 	}

# 	if (nrow(indicTs) >= r){
# 		if (diverse && r > 1){
# 			idx = select_diverse_rows(indicTs, r)
# 			indicTs = indicTs[idx, , drop = FALSE]
# 		} else {
# 			indicTs = indicTs[seq_len(r), , drop = FALSE]
# 		}
# 	} else {
# 		warning("Only ", nrow(indicTs), " unique Gurobi pool solutions available; returning with replacement to reach ", r, ".")
# 		if (nrow(indicTs) == 0){
# 			stop("No solutions available from Gurobi.")
# 		}
# 		indicTs = indicTs[sample(seq_len(nrow(indicTs)), r, replace = TRUE), , drop = FALSE]
# 	}

# 	indicTs
# }

# #' Find multiple designs using Gurobi and returns the minimum
# #' 
# #' This method searches through allocation space using Gurobi's optimization many times.
# #' It finds many different solutions by permuting the rows of the design matrix and 
# #' rerunning the optimization.
# #' 
# #' @param X 		The design matrix with $n$ rows (one for each subject) and $p$ columns 
# #' 					(one for each measurement on the subject). This is the design matrix you wish 
# #' 					to search for a more optimal design.
# #' @param r 		The number of vectors that should be returned
# #' @param objective The objective function to calculate. Default is \code{"mahal_dist"} for Mahalanobis distance
# #' @param ... 		Additional arguments to be passed to \code{initGurobiNumericalOptimizationExperimentalDesignObject}.
# #' @return			A list with the minimum objective value and its vector	 
# #' 
# #' @author Kapelner
# #' @examples
# #' \dontrun{
# #' if ("gurobi" %in% loadedNamespaces()) {
# #'   set.seed(1)
# #'   X = matrix(rnorm(12), nrow = 6)
# #'   res = gurobi_min_of_multiple_designs(
# #'     X,
# #'     r = 2,
# #'     num_cores = 1,
# #'     initial_time_limit_sec = 5,
# #'     verbose = FALSE
# #'   )
# #'   res$obj
# #' }
# #' }
# #' @export
# gurobi_min_of_multiple_designs = function(X, r, objective = "mahal_dist", ...){
# 	indicTs = gurobi_multiple_designs(X, r, ...)
# 	obj_min = .Machine$double.xmax
# 	i_min = NA
# 	for (i in 1 : r){
# 		obj_i = compute_objective_val(X, indicTs[i, ], objective = "mahal_dist")
# 		if (obj_i < obj_min){
# 			obj_min = obj_i
# 			i_min = i
# 		}
# 	}
# 	list(indicT = indicTs[i_min, ], obj = obj_min)
# }

#' Query the Gurobi Results
#' 
#' Returns the results (thus far) of the Gurobi numerical optimization design search
#' 
#' @param obj 				The \code{gurobi_numerical_optimization_experimental_design_search} object that is currently running the search
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' if ("gurobi" %in% loadedNamespaces()) {
#'   set.seed(1)
#'   X = matrix(rnorm(12), nrow = 6)
#'   gobj = initGurobiNumericalOptimizationExperimentalDesignObject(
#'     X,
#'     r = 2,
#'     num_cores = 1,
#'     initial_time_limit_sec = 5,
#'     verbose = FALSE
#'   )
#'   res = resultsGurobiNumericalOptimizeSearch(gobj)
#'   res$obj_vals
#' }
#' }
#' @export
resultsGurobiNumericalOptimizeSearch = function(obj){
	indicTs = NULL
	pool_vecs = list()
	extract_pool_vec = function(sol){
		if (is.null(sol)){
			return(NULL)
		}
		if (is.numeric(sol)){
			return(sol)
		}
		if (is.list(sol)){
			for (field in c("poolx", "poolnx", "x", "xn")){
				val = sol[[field]]
				if (!is.null(val)){
					return(val)
				}
			}
		}
		NULL
	}
	make_balanced_start = function(n){
		w = integer(n)
		w[sample.int(n, n / 2)] = 1L
		w
	}
	coerce_pool_vecs = function(pool_val, n, solcount = NULL){
		if (is.null(pool_val)){
			return(list())
		}
		if (is.matrix(pool_val) || is.data.frame(pool_val)){
			pool_mat = as.matrix(pool_val)
			if (nrow(pool_mat) == n){
				return(lapply(seq_len(ncol(pool_mat)), function(j) pool_mat[, j]))
			}
			if (ncol(pool_mat) == n){
				return(lapply(seq_len(nrow(pool_mat)), function(i) pool_mat[i, ]))
			}
			return(list(as.numeric(pool_mat)))
		}
		if (is.list(pool_val)){
			flat = list()
			for (sol in pool_val){
				val = extract_pool_vec(sol)
				if (is.null(val)){
					next
				}
				if (is.matrix(val) || is.data.frame(val)){
					mat = as.matrix(val)
					if (nrow(mat) == n){
						flat = c(flat, lapply(seq_len(ncol(mat)), function(j) mat[, j]))
					} else if (ncol(mat) == n){
						flat = c(flat, lapply(seq_len(nrow(mat)), function(i) mat[i, ]))
					} else {
						flat = c(flat, list(as.numeric(mat)))
					}
				} else {
					flat = c(flat, list(val))
				}
			}
			return(flat)
		}
		if (is.numeric(pool_val)){
			if (!is.null(solcount) && length(pool_val) == n * solcount){
				mat = matrix(pool_val, nrow = n)
				return(lapply(seq_len(ncol(mat)), function(j) mat[, j]))
			}
			if (length(pool_val) %% n == 0 && length(pool_val) > n){
				mat = matrix(pool_val, nrow = n)
				return(lapply(seq_len(ncol(mat)), function(j) mat[, j]))
			}
			return(list(pool_val))
		}
		list()
	}
	solcount = obj$solcount
	if (is.null(solcount) && !is.null(obj$poolobjval)) {
		solcount = length(obj$poolobjval)
	}
	for (field in c("pool", "poolx", "poolnx", "poolX", "poolnX")){
		pool_vecs = coerce_pool_vecs(obj[[field]], obj$n, solcount)
		if (length(pool_vecs) > 0){
			break
		}
	}
	if (length(pool_vecs) == 0){
		for (field in c("xn", "x")){
			pool_vecs = coerce_pool_vecs(obj[[field]], obj$n, 1)
			if (length(pool_vecs) > 0){
				break
			}
		}
	}
	if (length(pool_vecs) > 0){
		pool_vecs = Filter(function(x) !is.null(x) && length(x) > 0, pool_vecs)
		pool_vecs = lapply(pool_vecs, function(x) as.integer(as.numeric(x) > 0.5))
		indicTs = ged_helper_normalize_rows(pool_vecs, n = obj$n)
	}

	if (is.null(indicTs)){
		stop("No Gurobi solution found on object.")
	}

	indicTs = ged_helper_unique_rows(indicTs)
if (!is.null(obj$max_number_of_restarts) && obj$max_number_of_restarts > 0 &&
		!is.null(obj$model)) {
		gurobi_ns = tryCatch(get_gurobi_namespace(), error = function(e) NULL)
		if (!is.null(gurobi_ns)) {
			restarts = as.integer(obj$max_number_of_restarts)
			base_seed = if (!is.null(obj$gurobi_params$Seed)) {
				as.integer(obj$gurobi_params$Seed)
			} else {
				sample.int(1000000L, 1L)
			}
			for (attempt in seq_len(restarts)) {
				model = obj$model
				model$start = make_balanced_start(obj$n)
				jitter_scale = 1e-6 * max(1, mean(abs(model$obj)))
				model$obj = as.numeric(model$obj) + stats::runif(length(model$obj), -jitter_scale, jitter_scale)
				params = obj$gurobi_params
				if (!is.null(obj$restart_time_limit_sec)) {
					params$TimeLimit = obj$restart_time_limit_sec
				}
				params$Seed = base_seed + attempt
				sol = gurobi_call(gurobi_ns, model, params)
				solcount = sol$solcount
				if (is.null(solcount) && !is.null(sol$poolobjval)) {
					solcount = length(sol$poolobjval)
				}
				new_vecs = list()
				for (field in c("pool", "poolx", "poolnx", "poolX", "poolnX")){
					new_vecs = coerce_pool_vecs(sol[[field]], obj$n, solcount)
					if (length(new_vecs) > 0){
						break
					}
				}
				if (length(new_vecs) == 0){
					for (field in c("xn", "x")){
						new_vecs = coerce_pool_vecs(sol[[field]], obj$n, 1)
						if (length(new_vecs) > 0){
							break
						}
					}
				}
				if (length(new_vecs) > 0){
					new_vecs = lapply(new_vecs, function(x) as.integer(as.numeric(x) > 0.5))
					new_mat = ged_helper_normalize_rows(new_vecs, n = obj$n)
					indicTs = rbind(indicTs, new_mat)
				}
			}
		}
	}

	if (!is.null(obj$max_no_good_cuts) && obj$max_no_good_cuts > 0 &&
		!is.null(obj$model) && !is.null(obj$r) && nrow(indicTs) < obj$r) {
		gurobi_ns = tryCatch(get_gurobi_namespace(), error = function(e) NULL)
		if (!is.null(gurobi_ns)) {
			model = obj$model
			model$A = as.matrix(model$A)
			model$rhs = as.numeric(model$rhs)
			model$sense = as.character(model$sense)
			cut_keys = character(0)
			cuts_added = 0L
			make_key = function(w) paste(as.integer(w), collapse = "")
			add_cut = function(w){
				if (cuts_added >= obj$max_no_good_cuts) {
					return(FALSE)
				}
				w = as.integer(as.numeric(w) > 0.5)
				key = make_key(w)
				if (key %in% cut_keys) {
					return(FALSE)
				}
				coeff = 2 * w - 1
				rhs = sum(w == 1L) - 1L
				model$A <<- rbind(model$A, coeff)
				model$rhs <<- c(model$rhs, rhs)
				model$sense <<- c(model$sense, "<")
				cut_keys <<- c(cut_keys, key)
				cuts_added <<- cuts_added + 1L
				TRUE
			}
			if (nrow(indicTs) > 0) {
				for (i in seq_len(nrow(indicTs))) {
				if (cuts_added >= obj$max_no_good_cuts) {
					break
				}
					add_cut(indicTs[i, ])
				}
			}
			params = obj$gurobi_params
			if (!is.null(obj$restart_time_limit_sec)) {
				params$TimeLimit = obj$restart_time_limit_sec
			}
			while (cuts_added < obj$max_no_good_cuts && nrow(indicTs) < obj$r) {
				sol = gurobi_call(gurobi_ns, model, params)
				solcount = sol$solcount
				if (is.null(solcount) && !is.null(sol$poolobjval)) {
					solcount = length(sol$poolobjval)
				}
				new_vecs = list()
				for (field in c("pool", "poolx", "poolnx", "poolX", "poolnX")){
					new_vecs = coerce_pool_vecs(sol[[field]], obj$n, solcount)
					if (length(new_vecs) > 0){
						break
					}
				}
				if (length(new_vecs) == 0){
					for (field in c("xn", "x")){
						new_vecs = coerce_pool_vecs(sol[[field]], obj$n, 1)
						if (length(new_vecs) > 0){
							break
						}
					}
				}
				if (length(new_vecs) == 0) {
					break
				}
				new_vecs = lapply(new_vecs, function(x) as.integer(as.numeric(x) > 0.5))
				new_mat = ged_helper_normalize_rows(new_vecs, n = obj$n)
				indicTs = rbind(indicTs, new_mat)
				indicTs = ged_helper_unique_rows(indicTs)
				for (i in seq_len(nrow(new_mat))) {
					if (cuts_added >= obj$max_no_good_cuts) {
						break
					}
					add_cut(new_mat[i, ])
				}
			}
		}
	}

	indicTs = ged_helper_unique_rows(indicTs)
	if (nrow(indicTs) == 0){
		stop("No feasible Gurobi solutions found.")
	}
	message("Gurobi unique designs after restarts: ", nrow(indicTs))

	is_feasible = rowSums(indicTs) == obj$n / 2
	if (!all(is_feasible)) {
	  warning("Infeasible solutions violating n_T - n_C returned by Gurobi. Discarding.")
	  indicTs = indicTs[is_feasible, , drop = FALSE]
	}

	inv_cov_X = NULL
	if (identical(obj$objective, "mahal_dist") && !is.null(obj$SinvX)) {
		inv_cov_X = obj$SinvX
	}
	obj_vals = apply(indicTs, 1, FUN = function(wj){
		compute_objective_val(
			obj$X,
			wj,
			objective = obj$objective,
			inv_cov_X = inv_cov_X,
			use_safe_inverse = isTRUE(obj$use_safe_inverse)
		)
	})

	if (identical(obj$objective, "mahal_dist") && !is.null(obj$x)) {
		w_best = as.integer(as.numeric(obj$x) > 0.5)
		if (length(w_best) == obj$n) {
			objval_check = compute_objective_val(
				obj$X,
				w_best,
				objective = "mahal_dist",
				inv_cov_X = inv_cov_X,
				use_safe_inverse = isTRUE(obj$use_safe_inverse)
			)
			Q_check = if (!is.null(obj$Kgram)) {
				obj$Kgram
			} else {
				if (is.null(inv_cov_X)) {
					inv_cov_X = if (isTRUE(obj$use_safe_inverse)) {
						safe_cov_inverse(obj$X)
					} else {
						solve(stats::var(obj$X))
					}
				}
				obj$X %*% inv_cov_X %*% t(obj$X)
			}
			one_vec = rep(1, obj$n)
			qb = as.numeric(Q_check %*% w_best)
			q1 = as.numeric(Q_check %*% one_vec)
			obj_no_const = as.numeric(4 * crossprod(w_best, qb) - 4 * crossprod(w_best, q1))
			obj_with_const = obj_no_const + as.numeric(crossprod(one_vec, q1))
			obj_expected = (4 / (obj$n ^ 2)) * obj_with_const
			delta_expected = abs(objval_check - obj_expected)
			tol_expected = 1e-8 * max(1, abs(obj_expected))
			if (is.finite(delta_expected) && delta_expected > tol_expected) {
				warning(
					"Computed objective value differs from expected scaled Gurobi objective (",
					format(objval_check, digits = 6),
					" vs ",
					format(obj_expected, digits = 6),
					"). Check objective alignment and scaling."
				)
			}
		}
	}
	order_idx = order(obj_vals)
	indicTs = indicTs[order_idx, , drop = FALSE]
	list(indicT = indicTs[1, ], indicTs = indicTs, obj_vals = obj_vals[order_idx])	
}

ged_helper_unique_rows = function(mat){
	mat[!duplicated(as.data.frame(mat)), , drop = FALSE]
}

ged_helper_normalize_rows = function(mat, n = NULL){
	if (is.null(mat)){
		return(mat)
	}
	if (is.list(mat)){
		mat = do.call(rbind, lapply(mat, as.integer))
	} else if (!is.matrix(mat)){
		mat = matrix(mat, nrow = 1)
	}
	if (!is.null(n) && ncol(mat) != n && nrow(mat) == n){
		mat = t(mat)
	}
	if (is.null(dim(mat))){
		mat = matrix(mat, nrow = 1)
	}
	mat
}

get_gurobi_namespace = function() {
	if (!"gurobi" %in% loadedNamespaces()) {
		stop("Package \"gurobi\" is required. Call library(gurobi) before using this function.")
	}
	getNamespace("gurobi")
}

gurobi_call = function(gurobi_ns, model, params) {
	gurobi_fun = get("gurobi", envir = gurobi_ns)
	gurobi_fun(model, params)
}

testthat::test_that("gurobi multiple designs returns valid allocations", {
  skip_on_cmd_check()
  testthat::skip_if_not_installed("gurobi")

  license_ok = TRUE
  gurobi_ns = getNamespace("gurobi")
  gurobi_fun = get("gurobi", envir = gurobi_ns)
  tryCatch({
    model = list(
      obj = c(1, 1),
      modelsense = "max",
      A = matrix(c(1, 0, 0, 1), nrow = 2, byrow = TRUE),
      sense = c("<", "<"),
      rhs = c(1, 1),
      vtype = "B"
    )
    gurobi_fun(model, list(OutputFlag = 0))
  }, error = function(e) {
    if (grepl("HostID mismatch", e$message, fixed = TRUE) ||
        grepl("10009", e$message, fixed = TRUE)) {
      license_ok <<- FALSE
    } else {
      stop(e)
    }
  })
  if (!license_ok) {
    testthat::skip("Gurobi license not available on this host.")
  }

  set.seed(1)
  n = 20
  p = 3
  r = 3
  X = matrix(rnorm(n * p), nrow = n, ncol = p)

  gobj = initGurobiNumericalOptimizationExperimentalDesignObject(
    X,
    r = r,
    objective = "mahal_dist",
    initial_time_limit_sec = 5,
    num_cores = 1,
    verbose = FALSE,
    use_safe_inverse = TRUE
  )
  res = resultsGurobiNumericalOptimizeSearch(gobj)
  indicTs = res$indicTs
  if (is.null(dim(indicTs))) {
    indicTs = matrix(indicTs, nrow = 1)
  }
  if (nrow(indicTs) < r) {
    indicTs = indicTs[sample(seq_len(nrow(indicTs)), r, replace = TRUE), , drop = FALSE]
  } else if (nrow(indicTs) > r) {
    indicTs = indicTs[seq_len(r), , drop = FALSE]
  }

  testthat::expect_equal(dim(indicTs), c(r, n))
  testthat::expect_true(all(indicTs %in% c(0, 1)))
  testthat::expect_true(all(rowSums(indicTs) == n / 2))
})

#' Implements forced balanced randomization
#' 
#' For debugging, you can use \code{set.seed}
#' to be assured of deterministic output.
#' 
#' @param n 		number of observations
#' @param r 		number of randomized designs you would like
#' @param form		Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' @param seed		An integer which is the seed to be set within C++. Default is \code{NULL} which means the seed is set from the system clock.
#' @return 			a matrix where each column is one of the \code{r} designs
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' complete_randomization_with_forced_balanced(n = 6, r = 2, seed = 1)
#' }
#' @export
complete_randomization_with_forced_balanced = function(n, r, form = "one_zero", seed = NULL){
	assert_count(n, positive = TRUE)
	assert_count(r, positive = TRUE)
	assert_choice(form, c("one_zero", "pos_one_min_one"))
	assert_count(seed, positive = TRUE, null.ok = TRUE)
	seed = ifelse(is.null(seed), NA_integer_, as.integer(seed))
	assert_integer(seed)
	
	indicTs = complete_randomization_forced_balanced_cpp(n, r, seed)
	if (form == "pos_one_min_one"){
		indicTs = (indicTs - 0.5) * 2
	}
	indicTs
}


#' Implements complete randomization (without forced balance)
#' 
#' For debugging, you can use \code{set.seed}
#' to be assured of deterministic output.
#' 
#' @param n 		number of observations
#' @param r 		number of randomized designs you would like
#' @param form		Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' @return 			a matrix where each column is one of the \code{r} designs
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' complete_randomization(n = 6, r = 2)
#' }
#' @export
complete_randomization = function(n, r, form = "one_zero"){
	assert_count(n, positive = TRUE)
	assert_count(r, positive = TRUE)
	assert_choice(form, c("one_zero", "pos_one_min_one"))
	
	indicTs = complete_randomization_cpp(n, r)
	if (form == "pos_one_min_one"){
		indicTs = (indicTs - 0.5) * 2
	}
	indicTs
}

#' Implements unequally allocated complete randomization
#' 
#' For debugging, you can use \code{set.seed}
#' to be assured of deterministic output.
#' 
#' @param n 		number of observations
#' @param prop_T    the proportion of treatments needed
#' @param r 		number of randomized designs you would like
#' @param form		Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' @param seed      An integer which is the seed to be set within C++. Default is \code{NULL} which means the seed is set from the system clock.
#' @return 			a matrix where each column is one of the \code{r} designs
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' imbalanced_complete_randomization(n = 10, prop_T = 0.3, r = 2, seed = 1)
#' }
#' @export
imbalanced_complete_randomization = function(n, prop_T, r, form = "one_zero", seed = NULL){
	assert_count(n, positive = TRUE)
	assert_numeric(prop_T, lower = .Machine$double.eps, upper = 1 - .Machine$double.eps)
	n_T = n * prop_T
	assert_count(n_T, positive = TRUE)
	assert_count(r, positive = TRUE)
	assert_choice(form, c("one_zero", "pos_one_min_one"))
	seed = ifelse(is.null(seed), NA_integer_, as.integer(seed))
	assert_integer(seed)
	
	indicTs = complete_randomization_imbalanced_cpp(n, n_T, r, seed)
	if (form == "pos_one_min_one"){
		indicTs = (indicTs - 0.5) * 2
	}
	indicTs
}

#' Implements unequally allocated block designs
#' 
#' For debugging, you can use \code{set.seed}
#' to be assured of deterministic output. The following quantities
#' in this design must be integer valued or an error will be thrown: 
#'   n_B := n / B and n_B * prop_T
#' 
#' @param n 		number of observations
#' @param prop_T    the proportion of treatments allocated
#' @param B 		the number of blocks
#' @param r 		number of randomized designs you would like
#' @param form		Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' @param seed      An integer which is the seed to be set within C++. Default is \code{NULL} which means the seed is set from the system clock.
#' @return 			a matrix where each column is one of the \code{r} designs
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' imbalanced_block_designs(n = 12, prop_T = 0.5, B = 3, r = 2, seed = 1)
#' }
#' @export
imbalanced_block_designs = function(n, prop_T, B, r, form = "one_zero", seed = NULL){
	assert_count(n, positive = TRUE)
	assert_numeric(prop_T, lower = .Machine$double.eps, upper = 1 - .Machine$double.eps)
	assert_count(B, positive = TRUE) 
	assert_count(r, positive = TRUE)
	seed = ifelse(is.null(seed), NA_integer_, as.integer(seed))
	assert_integer(seed)
	
	n_B = n / B
	n_B_T = n_B * prop_T
	assert_count(n_B, positive = TRUE)
	assert_count(n_B_T, positive = TRUE)
	indicTs = imbalanced_block_designs_cpp(n_B, n_B_T, B, r, seed)
	if (form == "pos_one_min_one"){
		indicTs = (indicTs - 0.5) * 2
	}
	indicTs
}

# internal cache for block design var-cov results
.gen_var_cov_cache = new.env(parent = emptyenv())
.gen_var_cov_cache$max = 8L

#' Computes varcov matrix for block designs
#' 
#' The varcov matrix for block designs consists of a block-
#' diagonal matrix with B blocks (the number of blocks in
#' the design) with off-diagonal entries = -1 / (n/B - 1)
#' where n is the number of subjected in the study.
#'
#' @param n 		number of observations
#' @param prop_T    the proportion of treatments allocated
#' @param B 		the number of blocks
#' @param use_cache	Cache results for repeated calls with identical inputs. Default is \code{TRUE}.
#' @return 			varcov matrix for the specific block design
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' gen_var_cov_matrix_block_designs(n = 12, prop_T = 0.5, B = 3)
#' }
#' @export
gen_var_cov_matrix_block_designs = function(n, prop_T, B, use_cache = TRUE){
  assertCount(n, positive = TRUE)
  assert_numeric(prop_T, lower = .Machine$double.eps, upper = 1 - .Machine$double.eps)
  assertCount(B, positive = TRUE)
  assertLogical(use_cache)
  key = NULL
  if (use_cache){
    key = paste("cache", n, sprintf("%.17g", prop_T), B, sep = "|")
    cached = .gen_var_cov_cache[[key]]
    if (!is.null(cached)){
      return(cached)
    }
  }
  SigmaW = gen_var_cov_matrix_block_designs_cpp(n, prop_T, B)
  if (use_cache){
    .gen_var_cov_cache[[key]] = SigmaW
    order = .gen_var_cov_cache$order
    if (is.null(order)){
      order = character()
    }
    order = c(order[order != key], key)
    if (length(order) > .gen_var_cov_cache$max){
      drop_key = order[1]
      rm(list = drop_key, envir = .gen_var_cov_cache)
      order = order[-1]
    }
    .gen_var_cov_cache$order = order
  }
  SigmaW
}

#' A tool to find many types of a priori experimental designs
#'
#' @name 		GreedyExperimentalDesign
#' @title 		Greedy Experimental Design Search
#' @author 		Adam Kapelner \email{kapelner@@qc.cuny.edu}
#' @references 	Kapelner, A
#' @keywords 	design optimize
#' @import      rJava stats graphics grDevices checkmate
#' @importFrom 	nbpMatching distancematrix nonbimatch
#' @importFrom  Rcpp sourceCpp evalCpp
#' @importFrom  stringr str_detect
#' @importFrom  stringi stri_replace_first_regex
#' @importFrom  rlist list.cbind
#' @importFrom  checkmate assertClass assertCount assertNumeric assertChoice assertLogical assertTRUE assertMatrix vname
#' @useDynLib 	GreedyExperimentalDesign
##### Run "library(roxygen2); roxygenise("GreedyExperimentalDesign", clean = TRUE)" to regenerate all Rd files and NAMESPACE and DESCRIPTION file
##### but make sure you are in the root directory of the project. Make sure to add these two to namespace afterwards:
NULL

#' Create a Hadamard Design
#' 
#' This method returns unique designs according to a Hadamard matrix. For debugging, you can use \code{set.seed}
#' to be assured of deterministic output.
#' 
#' @param X				The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 						(one for each measurement on the subject). The measurements aren't used to
#' 						compute the Hadamard designs, only the number of rows.
#' @param strict		Hadamard matrices are not available for all $n$.
#' @param form			Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' 
#' @return				An matrix of dimension $R$ x $n$ where $R$ is the number of Hadamard allocations.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' W = hadamardExperimentalDesign(X, strict = TRUE, form = "one_zero")
#' dim(W)
#' }
#' @export
hadamardExperimentalDesign = function(X, strict = TRUE, form = "one_zero"){
	assertMatrixWithEvenNumRows(X)
	n = nrow(X)
	is_power_of_two = function(x){
		x > 0 && bitwAnd(x, x - 1) == 0
	}
	build_hadamard = function(m){
		if (m == 1){
			return(matrix(1, 1, 1))
		}
		if (!is_power_of_two(m)){
			stop("Hadamard order must be a power of two.")
		}
		H = build_hadamard(m / 2)
		rbind(cbind(H, H), cbind(H, -H))
	}
	if (is_power_of_two(n)){
		m = n
	} else if (strict){
		stop("A Hadamard matrix for this sample size doesn't exist. Use strict = FALSE to approximate one randomly.")
	} else {
		m = 2 ^ ceiling(log2(n))
	}
	W = (build_hadamard(m) + 1) / 2
	W = t(W)[-1, , drop = FALSE] #the first row is always just 1's, so remove it
	#also retain only those that are forced balance (I think this is all of them, but I want to just make sure
	W = W[rowMeans(W) == 0.5, , drop = FALSE]
	if (ncol(W) != n){
		if (strict){
			stop("A Hadamard matrix for this sample size doesn't exist. Use strict = FALSE to approximate one randomly.")
		}
		Wcut = matrix(NA, nrow = nrow(W), ncol = n)
		for (r in 1 : nrow(W)){
			w = W[r, ]
			idx_Ts = which(w == 1)
			idx_Cs = which(w == 0)
			idx_trimmed = sort(c(
				sample(idx_Ts, n / 2), 
				sample(idx_Cs, n / 2)
			))
			Wcut[r, ] = w[idx_trimmed]
		}
		W = Wcut #overwrite
	}
	
	if (form == "pos_one_min_one"){
		W = (W - 0.5) * 2
	}
	#now return only those that are unique
	unique(W)
}

assertMatrixWithEvenNumRows = function(x, .var.name = vname(x), add = NULL) {
	assertClass(x, "matrix")
	makeAssertion(x, (function(x){if (nrow(x) %% 2 != 0) "Matrix must have an even number of rows" else TRUE})(x), .var.name, add)
}

#' Begin a Search for the Optimal Solution
#' 
#' This method creates an object of type optimal_experimental_design and will immediately initiate
#' a search through allocation space. Since this search takes exponential time, for most machines, 
#' this method is futile beyond 28 samples. You've been warned! For debugging, you can use set 
#' \code{num_cores = 1} to be assured of deterministic output.
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more optimal design.
#' @param objective			The objective function to use when searching design space. This is a string
#' 							with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".
#' @param Kgram				If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
#' 							entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.
#' @param wait				Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 							deafult is \code{FALSE}.
#' @param start				Should we start searching immediately (default is \code{TRUE}).
#' @param num_cores 		The number of CPU cores you wish to use during the search. The default is \code{1}.
#' @param verbose			Should the algorithm emit progress output? Default is \code{TRUE}.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' @return					An object of type \code{optimal_experimental_design_search} which can be further operated upon
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(12), nrow = 6)
#' obj = initOptimalExperimentalDesignObject(
#'   X,
#'   objective = "abs_sum_diff",
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' obj
#' }
#' @export
initOptimalExperimentalDesignObject = function(
		X = NULL,
		objective = "mahal_dist", 
		Kgram = NULL,
		wait = FALSE, 
		start = TRUE,
		num_cores = 1,
		verbose = TRUE,
		use_safe_inverse = FALSE){
	
	verify_objective_function(objective, Kgram, n)
	
	if (!is.null(Kgram)){
		n = nrow(Kgram)
		p = NA
	} else {
		n = nrow(X)
		p = ncol(X)
	}
	if (n %% 2 != 0){
		stop("Design matrix must have even rows to have equal treatments and controls")
	}
	
	
	if (objective == "abs_sum_diff"){
		#standardize it -- much faster here
		Xstd = standardize_data_matrix(X)
	}
	if (objective == "mahal_dist"){
		if (p < n){
			if (use_safe_inverse){
				SinvX = safe_cov_inverse(X)
			} else {
				SinvX = solve(stats::var(X))
			}
		}
	}
	assertLogical(verbose)
	assertLogical(use_safe_inverse)
	
	#we are about to construct a OptimalExperimentalDesign java object. First, let R garbage collect
	#to clean up previous objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	
	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("OptimalExperimentalDesign.OptimalExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))
	.jcall(java_obj, "V", "setN", as.integer(n))
	if (objective != "kernel"){
		p = ncol(X)
		.jcall(java_obj, "V", "setP", as.integer(p))
	}
	.jcall(java_obj, "V", "setObjective", objective)
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}	
	
	#feed in the gram matrix if applicable
	if (!is.null(Kgram)){
		setGramMatrix(java_obj, Kgram)
	} else {
		#feed in the data
		if (objective == "abs_sum_diff"){
			.jcall(java_obj, "V", "setDataMatrix", as.double(Xstd), as.integer(n), as.integer(p))
		} else {
			.jcall(java_obj, "V", "setDataMatrix", as.double(X), as.integer(n), as.integer(p))
		}
		
		#feed in the inverse var-cov matrix
		if (objective == "mahal_dist"){
			if (p < n){
				.jcall(
					java_obj,
					"V",
					"setInvVarCovMatrix",
					.jarray(as.double(SinvX), contents.class = "double"),
					as.integer(p)
				)
			}
		}
	}
		
	#now return information as an object (just a list)
	optimal_experimental_design_search = list()
	optimal_experimental_design_search$start = start
	optimal_experimental_design_search$wait = wait
	optimal_experimental_design_search$X = X
	optimal_experimental_design_search$n = n
	optimal_experimental_design_search$p = p
	optimal_experimental_design_search$objective = objective
	optimal_experimental_design_search$java_obj = java_obj
	optimal_experimental_design_search$verbose = verbose
	class(optimal_experimental_design_search) = "optimal_experimental_design_search"
	#if the user wants to run it immediately...
	if (start){
		startSearch(optimal_experimental_design_search)
	}
	#return the final object
	optimal_experimental_design_search
}

#' Prints a summary of a \code{optimal_experimental_design_search} object
#' 
#' @param x			The \code{optimal_experimental_design_search} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print optimal_experimental_design_search
#' @export
print.optimal_experimental_design_search = function(x, ...){
	progress = .jcall(x$java_obj, "D", "progress")
	time_elapsed = searchTimeElapsed(x)
	if (progress == 0){
		cat("No progress on the OptimalExperimentalDesign. Did you run \"startOptimalSearch?\"\n")
	} else if (progress < 1){
		cat("The search is", round(progress * 100, 2), "% complete.\n")
	} else {
		cat("The search is complete.\n")
	}
}

#' Prints a summary of a \code{optimal_experimental_design_search} object
#' 
#' @param object		The \code{optimal_experimental_design_search} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary optimal_experimental_design_search
#' @export
summary.optimal_experimental_design_search = function(object, ...){
	print(object, ...)
}

#' Returns the results (thus far) of the optimal design search
#' 
#' @param obj 				The \code{optimal_experimental_design} object that is currently running the search
#' @param num_vectors		How many allocation vectors you wish to return. The default is 1 meaning the best vector. If \code{Inf},
#' 							it means all vectors.
#' @param form				Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(12), nrow = 6)
#' obj = initOptimalExperimentalDesignObject(
#'   X,
#'   objective = "abs_sum_diff",
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsOptimalSearch(obj, num_vectors = 2, form = "one_zero")
#' res$opt_obj_val
#' }
#' @export
resultsOptimalSearch = function(obj, num_vectors = 2, form = "one_zero"){
	obj_vals = .jcall(obj$java_obj, "[D", "getAllObjectiveVals", simplify = TRUE)
	ordered_indices = order(obj_vals)
	
	if (num_vectors == Inf){
		num_vectors = length(ordered_indices)
	}

	indicTs = NULL
	if (num_vectors > 0){
		ordered_java_indices = as.integer(ordered_indices[1 : num_vectors] - 1)
		ordered_java_indices_j = .jarray(ordered_java_indices)
		indicTs = .jcall(obj$java_obj, "[[I", "getIndicTs", ordered_java_indices_j, simplify = TRUE)
		if (form == "pos_one_min_one"){
			if (length(indicTs) > 0 && min(indicTs) >= 0 && max(indicTs) <= 1){
				indicTs = (indicTs - 0.5) * 2
			}
		} else if (form == "one_zero" && length(indicTs) > 0 && min(indicTs) < 0){
			indicTs = (indicTs + 1) / 2
		}
	}
	list(
		opt_obj_val = obj_vals[ordered_indices[1]],
		ordered_obj_vals = obj_vals[ordered_indices],
		obj_vals_unordered = obj_vals,
		orig_order = ordered_indices,
		indicTs = indicTs
	)
}

#' Begin a Search for Binary Matching Followed by Rerandomization
#' 
#' This method creates an object of type binary_then_rerandomization_experimental_design and will find optimal matched pairs which
#' are then rerandomized in order to further minimize a balance metric. You can then
#' use the function \code{resultsBinaryMatchThenRerandomizationSearch} to obtain the randomized allocation vectors. For one column
#' in X, the matching just sorts the values to find the pairs trivially.
#' 
#' @param X						The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 								(one for each measurement on the subject). This is the design matrix you wish 
#' 								to search for a more optimal design.
#' @param compute_dist_matrix	The function that computes the distance matrix between every two observations in \code{X}, 
#' 								its only argument. The default is \code{NULL} signifying euclidean squared distance optimized in C++.
#' @param ...					Arguments passed to \code{initGreedyExperimentalDesignObject}. It is recommended to set
#' 								\code{max_designs} otherwise it will default to 10,000.
#' @param verbose				Should the algorithm emit progress output? Default is \code{TRUE}.
#' @return						An object of type \code{binary_experimental_design} which can be further operated upon.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' obj = initBinaryMatchFollowedByRerandomizationDesignSearchObject(
#'   X,
#'   max_designs = 4,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   obj_val_cutoff_to_include = Inf,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' obj
#' }
#' @export
initBinaryMatchFollowedByRerandomizationDesignSearchObject = function(X, compute_dist_matrix = NULL, verbose = TRUE, ...){
	n = nrow(X)
	p = ncol(X)
	assertLogical(verbose)
	
	if (n %% 4 != 0){
		stop("Design matrix must have number of rows divisible by four for this type of design.")
	}
	binary_match_structure = computeBinaryMatchStructure(X, compute_dist_matrix = compute_dist_matrix)
	binary_match_structure$verbose = verbose
	#now we create a reduced matrix X by diffing the pairs
	Xdiffs = matrix(NA, nrow = nrow(X) / 2, ncol = ncol(X))
	for (i in 1 : (nrow(X) / 2)){		
		Xdiffs[i, ] = X[binary_match_structure$indicies_pairs[i, 1], ] - X[binary_match_structure$indicies_pairs[i, 2], ]
	}
	
	binary_then_rerandomization_experimental_design = list()
	binary_then_rerandomization_experimental_design$X = X
	binary_then_rerandomization_experimental_design$n = n
	binary_then_rerandomization_experimental_design$p = p
	binary_then_rerandomization_experimental_design$binary_match_structure = binary_match_structure
	binary_then_rerandomization_experimental_design$verbose = verbose
	dots = list(...)
	if (is.null(dots$verbose)){
		dots$verbose = verbose
	}
	binary_then_rerandomization_experimental_design$rerandomization_design = do.call(initRerandomizationExperimentalDesignObject, c(list(Xdiffs), dots))
	class(binary_then_rerandomization_experimental_design) = "binary_then_rerandomization_experimental_design"
	binary_then_rerandomization_experimental_design
}

#' Returns unique allocation vectors that are binary matched
#' 
#' @param obj 				The \code{binary_then_greedy_experimental_design} object where the pairs are computed.
#' @param num_vectors		How many random allocation vectors you wish to return. The default is \code{NULL} indicating you want all of them.
#' @param compute_obj_vals	Should we compute all the objective values for each allocation? Default is \code{FALSE}.
#' @param form				Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' obj = initBinaryMatchFollowedByRerandomizationDesignSearchObject(
#'   X,
#'   max_designs = 4,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   obj_val_cutoff_to_include = Inf,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsBinaryMatchThenRerandomizationSearch(obj, num_vectors = 3, form = "one_zero")
#' dim(res$indicTs)
#' }
#' @export
resultsBinaryMatchThenRerandomizationSearch = function(obj, num_vectors = NULL, compute_obj_vals = FALSE, form = "one_zero", use_safe_inverse = FALSE){
	assertClass(obj, "binary_then_rerandomization_experimental_design")
	assertCount(num_vectors, positive = TRUE, null.ok = TRUE)
	if (is.null(num_vectors)){
		num_vectors = obj$rerandomization_design$max_designs
	}
	num_vectors_completed = .jcall(obj$rerandomization_design$java_obj, "I", "progress")
	if (num_vectors > num_vectors_completed){
		warning("You requested ", num_vectors, " but only ", num_vectors_completed, " are available.")
		num_vectors = num_vectors_completed
	}
	assertLogical(compute_obj_vals)
	assertChoice(form, c("one_zero", "pos_one_min_one"))
	assertLogical(use_safe_inverse)
	
	rerand_res = resultsRerandomizationSearch(obj$rerandomization_design, include_assignments = TRUE, "one_zero")
	#the allocation vectors returned here have entries = 1 if the pair is left unswitched and = 0 if we should switch the pair
	indicTs = matrix(NA, nrow = num_vectors, ncol = obj$n)
	for (r in 1 : num_vectors){
		#first we copy the binary indices starting point
		pair_matrix_copy = obj$binary_match_structure$indicies_pairs
		#now we pull out a w_diff
		w_diff = rerand_res$ending_indicTs[r, ]
		#split the pair matrix based on the rerandomization vector
		pair_matrix_T_is_first = pair_matrix_copy[w_diff == 1, ] #"one_zero" form was forced above
		pair_matrix_C_is_first = pair_matrix_copy[w_diff == 0, ] #"one_zero" form was forced above
		#now set all the entries as T if T is first and it's in the first column and 0 if second column
		indicTs[r, pair_matrix_T_is_first[, 1]] = 1
		indicTs[r, pair_matrix_T_is_first[, 2]] = 0
		#vice versa if C is first
		indicTs[r, pair_matrix_C_is_first[, 1]] = 0
		indicTs[r, pair_matrix_C_is_first[, 2]] = 1		
	}
	
	obj_vals = NULL
	if (compute_obj_vals){
		if (obj$rerandomization_design$objective == "mahal_dist"){
			if (use_safe_inverse){
				SinvX = safe_cov_inverse(obj$X)
			} else {
				SinvX = solve(stats::var(obj$X))
			}
			obj_vals = apply(indicTs, 1, FUN = function(w){compute_objective_val(obj$X, w, objective = "mahal_dist", SinvX)})
		} else {
			obj_vals = apply(indicTs, 1, FUN = function(w){compute_objective_val(obj$X, w, objective = "abs_sum_diff")})
		}
	}

	if (form == "pos_one_min_one"){
		indicTs = (indicTs - 0.5) * 2
	}
	list(
		obj_vals = obj_vals,
		indicTs = indicTs,
		form = form
	)
}

#' Prints a summary of a \code{binary_then_rerandomization_experimental_design} object
#' 
#' @param x			The \code{binary_then_rerandomization_experimental_design} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print binary_then_rerandomization_experimental_design
#' @export
print.binary_then_rerandomization_experimental_design = function(x, ...){
	print(x$rerandomization_design, ...)
}

#' Prints a summary of a \code{binary_then_rerandomization_experimental_design} object
#' 
#' @param object		The \code{binary_then_rerandomization_experimental_design} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary binary_then_rerandomization_experimental_design
#' @export
summary.binary_then_rerandomization_experimental_design = function(object, ...){
	print(object$rerandomization_design, ...)
}

#' Curate More Orthogonal Vectors Greedily
#' 
#' This function takes a set of allocation vectors and pares them down one-by-one
#' by eliminating the vector that can result in the largest reduction in Avg[ |r_ij| ].
#' It is recommended to begin with a set of unmirrored vectors for speed. Then add the mirrors later
#' for whichever subset you wish.
#' 
#' @param W 		A matrix in in the set \eqn{{-1, 1}^{R x n}} which have R allocation vectors for an experiment of sample size n.
#' @param Rmin 		The minimum number of vectors to consider in a design. The default is the true bottom, two.
#' @param verbose 	Default is \code{FALSE} but if not, it will print out a message for each iteration.
#' @return 			A list with two elements: (1) \code{avg_abs_rij_by_R} which is a data frame with R - Rmin + 1 rows and 
#' 					columns R and average absolute r_ij and (2) \code{Wsorted} which provides the collection of vectors in
#' 					sorted by best average absolute r_ij in row order from best to worst.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' W = matrix(sample(c(-1, 1), 6 * 8, replace = TRUE), nrow = 6)
#' res = greedy_orthogonalization_curation(W, Rmin = 3, verbose = FALSE)
#' res$avg_abs_rij_by_R
#' }
#' @export
greedy_orthogonalization_curation = function(W, Rmin = 2, verbose = FALSE){
	assertClass(W, "matrix")
	assertCount(Rmin)
	R = nrow(W)
	assertNumeric(Rmin, upper = R)
	
	if (R > 1000){
		warning("If number of designs > 1000, it goes pretty slow.\n")
	}
	
	n = ncol(W)
	Rij = abs(W %*% t(W)) / n
	diag(Rij) = 0
	res = greedy_orthogonalization_eliminate_cpp(Rij, Rmin, TRUE)
	avg_abs_rijss = res$avg_abs_rijss
	eliminations = res$eliminated
	remaining = res$remaining

	if (verbose){
		cat("iter", 0, "avg abs rijs", round(avg_abs_rijss[1], 4), "\n")
		if (length(eliminations) > 0){
			for (iter_num in 1 : length(eliminations)){
				cat("iter", iter_num, "obs eliminated:", eliminations[iter_num],
					"avg abs rijs", round(avg_abs_rijss[iter_num + 1], 4), "\n")
			}
		}
	}
	#now return the original allocations in order of elimination as well as the rho reductions
	list(
		avg_abs_rij_by_R =
			data.frame(R = R - (0 : (length(avg_abs_rijss) - 1)), avg_abs_rijs = avg_abs_rijss),
		Wsorted =
			W[c(remaining, rev(eliminations)), ]
	)
}

options(java.parameters = "-Xmx10g")

library(testthat)

MAX_FAILS = 3
testthat::set_max_fails(MAX_FAILS)
options(testthat.progress.max_fails = MAX_FAILS)
Sys.setenv(TESTTHAT_MAX_FAILS = as.character(MAX_FAILS))
Sys.setenv(TESTTHAT_PARALLEL = "false")

ged_failure_state = new.env(parent = emptyenv())
ged_failure_state$count = 0L
ged_failure_state$max_fail = MAX_FAILS
ged_failure_state$aborting = FALSE

resolve_testthat_dir = function() {
  wd = getwd()
  if (dir.exists(file.path(wd, "tests", "testthat"))) {
    return(file.path(wd, "tests", "testthat"))
  }
  if (dir.exists(file.path(wd, "GreedyExperimentalDesign", "tests", "testthat"))) {
    return(file.path(wd, "GreedyExperimentalDesign", "tests", "testthat"))
  }
  if (basename(wd) == "testthat" && basename(dirname(wd)) == "tests") {
    return(wd)
  }
  wd
}

resolve_package_root = function() {
  testthat_dir = resolve_testthat_dir()
  candidate = normalizePath(file.path(testthat_dir, "..", ".."), winslash = "/", mustWork = FALSE)
  if (file.exists(file.path(candidate, "DESCRIPTION"))) {
    return(candidate)
  }
  candidate
}

package_root = resolve_package_root()
local_lib = file.path(package_root, ".Rlib")
if (dir.exists(local_lib)) {
  .libPaths(c(local_lib, .libPaths()))
}

add_java_classpath = function(root_dir) {
  jar_path = file.path(root_dir, "inst", "java", "GreedyExperimentalDesign.jar")
  if (file.exists(jar_path)) {
    rJava::.jaddClassPath(jar_path)
  }
  invisible(NULL)
}

use_pkgload = requireNamespace("pkgload", quietly = TRUE) &&
  file.exists(file.path(package_root, "DESCRIPTION")) &&
  dir.exists(file.path(package_root, "R"))
if (use_pkgload) {
  pkgload::load_all(package_root, export_all = FALSE, quiet = TRUE)
  add_java_classpath(package_root)
} else {
  library(GreedyExperimentalDesign)
}

ged_failure_state$log_path = file.path(resolve_testthat_dir(), "ged-failures.log")
dir.create(dirname(ged_failure_state$log_path), recursive = TRUE, showWarnings = FALSE)
if (file.exists(ged_failure_state$log_path)) {
  file.remove(ged_failure_state$log_path)
}
cat(
  paste0("---- ", format(Sys.time(), "%Y-%m-%d %H:%M:%S"), " ----\n"),
  "Test run started.\n\n",
  file = ged_failure_state$log_path,
  append = TRUE
)

append_failure_log = function(exp) {
  if (is.null(ged_failure_state$log_path) || !nzchar(ged_failure_state$log_path)) {
    ged_failure_state$log_path = file.path(tempdir(), "ged-failures.log")
  }
  header = paste0(
    "---- ",
    format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    " ----\n"
  )
  body = testthat:::issue_summary(exp)
  cat(
    header,
    body,
    "\n\n",
    file = ged_failure_state$log_path,
    append = TRUE
  )
}

emit_failure_console = function(exp) {
  header = paste0(
    "---- ",
    format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    " ----\n"
  )
  body = testthat:::issue_summary(exp)
  cat(header, body, "\n\n", file = stderr())
  flush.console()
}

record_immediate_failure = function(exp) {
  ged_failure_state$count = ged_failure_state$count + 1L
  append_failure_log(exp)
  emit_failure_console(exp)
  if (ged_failure_state$count >= ged_failure_state$max_fail) {
    ged_failure_state$aborting = TRUE
    stop("Too many failures; aborting remaining tests.")
  }
  invisible(NULL)
}

LoggingReporter = R6::R6Class(
  "LoggingReporter",
  inherit = testthat::Reporter,
  public = list(
    add_result = function(context, test, result) {
      if (!testthat:::expectation_success(result)) {
        record_immediate_failure(result)
      }
    }
  )
)

install_reporter_hook = function() {
  if (!exists("ged_orig_test_files_reporter", envir = .GlobalEnv, inherits = FALSE)) {
    assign("ged_orig_test_files_reporter", testthat:::test_files_reporter, envir = .GlobalEnv)
    utils::assignInNamespace(
      "test_files_reporter",
      function(reporter, mode = c("serial", "parallel"), desc = NULL, frame = rlang::caller_env()) {
        res = get("ged_orig_test_files_reporter", envir = .GlobalEnv)(
          reporter,
          mode = mode,
          desc = desc,
          frame = frame
        )
        res$multi = testthat::MultiReporter$new(reporters = c(list(LoggingReporter$new()), res$multi$reporters))
        res
      },
      ns = "testthat"
    )
  }
  invisible(TRUE)
}

patch_recover2 = function() {
  if (!exists("ged_orig_recover2", envir = .GlobalEnv, inherits = FALSE)) {
    assign("ged_orig_recover2", testthat:::recover2, envir = .GlobalEnv)
    utils::assignInNamespace(
      "recover2",
      function(...) {
        if (!interactive()) {
          return(invisible(FALSE))
        }
        get("ged_orig_recover2", envir = .GlobalEnv)(...)
      },
      ns = "testthat"
    )
  }
  invisible(TRUE)
}

abort_if_too_many_failures = function() {
  if (isTRUE(ged_failure_state$aborting)) {
    stop("Too many failures; aborting remaining tests.")
  }
  invisible(NULL)
}

with_immediate_failures = function(expr) {
  abort_if_too_many_failures()
  force(expr)
}

skip_on_cmd_check = function() {
  is_check = nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_")) ||
    nzchar(Sys.getenv("R_CMD_CHECK")) ||
    nzchar(Sys.getenv("_R_CHECK_CRAN_INCOMING_")) ||
    nzchar(Sys.getenv("_R_CHECK_CRAN_INCOMING_REMOTE_"))
  if (is_check) {
    testthat::skip("Skipped during R CMD check.")
  }
  invisible(NULL)
}

install_reporter_hook()
patch_recover2()

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

is_symmetric_cpp <- function(D, tol) {
    .Call('_GreedyExperimentalDesign_is_symmetric_cpp', PACKAGE = 'GreedyExperimentalDesign', D, tol)
}

sort_unique_pairs_cpp <- function(pairs) {
    .Call('_GreedyExperimentalDesign_sort_unique_pairs_cpp', PACKAGE = 'GreedyExperimentalDesign', pairs)
}

generate_block_design_cpp <- function(B, nR, dummy_block) {
    .Call('_GreedyExperimentalDesign_generate_block_design_cpp', PACKAGE = 'GreedyExperimentalDesign', B, nR, dummy_block)
}

compute_distance_matrix_cpp <- function(X) {
    .Call('_GreedyExperimentalDesign_compute_distance_matrix_cpp', PACKAGE = 'GreedyExperimentalDesign', X)
}

shuffle_cpp <- function(w, seed) {
    .Call('_GreedyExperimentalDesign_shuffle_cpp', PACKAGE = 'GreedyExperimentalDesign', w, seed)
}

create_all_ys_cpp <- function(pCs, pTs, W, two_n, nY) {
    .Call('_GreedyExperimentalDesign_create_all_ys_cpp', PACKAGE = 'GreedyExperimentalDesign', pCs, pTs, W, two_n, nY)
}

all_elements_same_cpp <- function(vec) {
    .Call('_GreedyExperimentalDesign_all_elements_same_cpp', PACKAGE = 'GreedyExperimentalDesign', vec)
}

gen_pm_designs_cpp <- function(indicies_pairs, n, r) {
    .Call('_GreedyExperimentalDesign_gen_pm_designs_cpp', PACKAGE = 'GreedyExperimentalDesign', indicies_pairs, n, r)
}

greedy_orthogonalization_eliminate_cpp <- function(Rij, Rmin, stop_if_worse) {
    .Call('_GreedyExperimentalDesign_greedy_orthogonalization_eliminate_cpp', PACKAGE = 'GreedyExperimentalDesign', Rij, Rmin, stop_if_worse)
}

compute_indicTs_from_pairs_cpp <- function(pairs, w_diff, n) {
    .Call('_GreedyExperimentalDesign_compute_indicTs_from_pairs_cpp', PACKAGE = 'GreedyExperimentalDesign', pairs, w_diff, n)
}

compute_indicTs_from_pairs_lgl_cpp <- function(pairs, w_diff, n) {
    .Call('_GreedyExperimentalDesign_compute_indicTs_from_pairs_lgl_cpp', PACKAGE = 'GreedyExperimentalDesign', pairs, w_diff, n)
}

compute_indicTs_from_pairs_raw_cpp <- function(pairs, w_diff, n) {
    .Call('_GreedyExperimentalDesign_compute_indicTs_from_pairs_raw_cpp', PACKAGE = 'GreedyExperimentalDesign', pairs, w_diff, n)
}

compute_kernel_matrix_cpp <- function(X, kernel_type, poly_s) {
    .Call('_GreedyExperimentalDesign_compute_kernel_matrix_cpp', PACKAGE = 'GreedyExperimentalDesign', X, kernel_type, poly_s)
}

compute_kernel_quadratic_forms_cpp <- function(W, K) {
    .Call('_GreedyExperimentalDesign_compute_kernel_quadratic_forms_cpp', PACKAGE = 'GreedyExperimentalDesign', W, K)
}

compute_avg_abs_rijss_left <- function(Rij) {
    .Call('_GreedyExperimentalDesign_compute_avg_abs_rijss_left', PACKAGE = 'GreedyExperimentalDesign', Rij)
}

compute_mahalanobis_distance_matrix_cpp <- function(X, Sinv) {
    .Call('_GreedyExperimentalDesign_compute_mahalanobis_distance_matrix_cpp', PACKAGE = 'GreedyExperimentalDesign', X, Sinv)
}

standardize_data_matrix_cpp <- function(X) {
    .Call('_GreedyExperimentalDesign_standardize_data_matrix_cpp', PACKAGE = 'GreedyExperimentalDesign', X)
}

compute_objective_val_cpp <- function(X, indic_T, objective, inv_cov_X = NULL) {
    .Call('_GreedyExperimentalDesign_compute_objective_val_cpp', PACKAGE = 'GreedyExperimentalDesign', X, indic_T, objective, inv_cov_X)
}

compute_objective_vals_cpp <- function(X, indicTs, objective, inv_cov_X = NULL) {
    .Call('_GreedyExperimentalDesign_compute_objective_vals_cpp', PACKAGE = 'GreedyExperimentalDesign', X, indicTs, objective, inv_cov_X)
}

compute_pair_differences_cpp <- function(X, pairs) {
    .Call('_GreedyExperimentalDesign_compute_pair_differences_cpp', PACKAGE = 'GreedyExperimentalDesign', X, pairs)
}

complete_randomization_forced_balanced_cpp <- function(n, r, seed) {
    .Call('_GreedyExperimentalDesign_complete_randomization_forced_balanced_cpp', PACKAGE = 'GreedyExperimentalDesign', n, r, seed)
}

complete_randomization_imbalanced_cpp <- function(n, nT, r, seed) {
    .Call('_GreedyExperimentalDesign_complete_randomization_imbalanced_cpp', PACKAGE = 'GreedyExperimentalDesign', n, nT, r, seed)
}

complete_randomization_cpp <- function(n, r) {
    .Call('_GreedyExperimentalDesign_complete_randomization_cpp', PACKAGE = 'GreedyExperimentalDesign', n, r)
}

imbalanced_block_designs_cpp <- function(n_B, n_B_T, B, r, seed) {
    .Call('_GreedyExperimentalDesign_imbalanced_block_designs_cpp', PACKAGE = 'GreedyExperimentalDesign', n_B, n_B_T, B, r, seed)
}

compute_randomization_metrics_cpp <- function(designs) {
    .Call('_GreedyExperimentalDesign_compute_randomization_metrics_cpp', PACKAGE = 'GreedyExperimentalDesign', designs)
}

gen_var_cov_matrix_block_designs_cpp <- function(n, prop_T, B) {
    .Call('_GreedyExperimentalDesign_gen_var_cov_matrix_block_designs_cpp', PACKAGE = 'GreedyExperimentalDesign', n, prop_T, B)
}

#' Begin a Rerandomization Search
#' 
#' This method creates an object of type rerandomization_experimental_design and will immediately initiate
#' a search through allocation space for forced-balance designs. For debugging, you can use set the \code{seed}
#' parameter and \code{num_cores = 1} to be assured of deterministic output.
#' 
#' @param X							The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 									(one for each measurement on the subject). This is the design matrix you wish 
#' 									to search for a more optimal design.
#' @param obj_val_cutoff_to_include	Only allocation vectors with objective values lower than this threshold will be returned.
#' 									If the cutoff is infinity, you are doing BCRD and you should use the \code{complete_randomization_with_forced_balanced}
#' 									function instead.
#' @param max_designs 				The maximum number of designs to be returned. Default is 10,000. Make this large 
#' 									so you can search however long you wish as the search can be stopped at any time by
#' 									using the \code{\link{stopSearch}} method 
#' @param objective					The objective function to use when searching design space. This is a string
#' 									with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".
#' @param Kgram						If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
#' 									entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.
#' @param wait						Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 									default is \code{FALSE}.
#' @param start						Should we start searching immediately (default is \code{TRUE}).
#' @param num_cores 				The number of CPU cores you wish to use during the search. The default is \code{1}.
#' @param seed						The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
#' 									the output will not be deterministic. Default is \code{NULL} for no seed set.
#' @param verbose					Should the algorithm emit progress output? Default is \code{TRUE}.
#' @param use_safe_inverse			Should a regularized inverse be used for the Mahalanobis objective?
#' 									Default is \code{FALSE}.
#' @return							An object of type \code{rerandomization_experimental_design_search} which can be further operated upon.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' obj = initRerandomizationExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   obj_val_cutoff_to_include = Inf,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' obj
#' }
#' @export
initRerandomizationExperimentalDesignObject = function(
		X = NULL, 
		obj_val_cutoff_to_include,
		max_designs = 1000,
		objective = "mahal_dist", 
		Kgram = NULL,
		wait = FALSE, 
		start = TRUE,
		num_cores = 1,
		seed = NULL,
		verbose = TRUE,
		use_safe_inverse = FALSE){
	
	verify_objective_function(objective, Kgram, n)
	
	if (!is.null(Kgram)){
		n = nrow(Kgram)
		p = NA
	} else {
		n = nrow(X)
		p = ncol(X)
	}
	if (n %% 2 != 0){
		stop("Design matrix must have even rows to have equal treatments and controls")
	}
	
	if (objective == "abs_sum_diff"){
		#standardize it -- much faster here
		Xstd = standardize_data_matrix(X)
	}
	if (objective == "mahal_dist"){
		if (p < n){
			if (use_safe_inverse){
				SinvX = safe_cov_inverse(X)
			} else {
				SinvX = solve(stats::var(X))
			}
		}
	}
	assertLogical(verbose)
	assertLogical(use_safe_inverse)
	
	#we are about to construct a RerandomizationExperimentalDesign java object. First, let R garbage collect
	#to clean up previous RerandomizationExperimentalDesign objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	
	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("RerandomizationExperimentalDesign.RerandomizationExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setMaxDesigns", as.integer(max_designs))
	if (!is.null(obj_val_cutoff_to_include)){
		.jcall(java_obj, "V", "setObjValCutoffToInclude", as.numeric(obj_val_cutoff_to_include))
	}
	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))	
	if (!is.null(seed)){
		.jcall(java_obj, "V", "setSeed", as.integer(seed))
		if (num_cores != 1){
			warning("Setting the seed with multiple cores does not guarantee deterministic output.")
		}		
	}
	.jcall(java_obj, "V", "setN", as.integer(n))
	if (objective != "kernel"){
		p = ncol(X)
		.jcall(java_obj, "V", "setP", as.integer(p))
	}
	.jcall(java_obj, "V", "setObjective", objective)
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}
	
	#feed in the gram matrix if applicable
	if (!is.null(Kgram)){
		setGramMatrix(java_obj, Kgram)
	} else {
		#feed in the raw data
		for (i in 1 : n){
			if (objective == "abs_sum_diff"){
				.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), Xstd[i, , drop = FALSE]) #java indexes from 0...n-1
			} else {
				.jcall(java_obj, "V", "setDataRow", as.integer(i - 1), X[i, , drop = FALSE]) #java indexes from 0...n-1
			}
		}	
		
		#feed in the inverse var-cov matrix
		if (objective == "mahal_dist"){
			if (p < n){
				for (j in 1 : p){
					.jcall(java_obj, "V", "setInvVarCovRow", as.integer(j - 1), SinvX[j, , drop = FALSE]) #java indexes from 0...n-1
				}
			}
		}
	}
	
	#now return information as an object (just a list)
	rerandomization_experimental_design_search = list()
	rerandomization_experimental_design_search$max_designs = max_designs
	rerandomization_experimental_design_search$obj_val_cutoff_to_include = obj_val_cutoff_to_include
	rerandomization_experimental_design_search$start = start
	rerandomization_experimental_design_search$wait = wait
	rerandomization_experimental_design_search$X = X
	rerandomization_experimental_design_search$n = n
	rerandomization_experimental_design_search$p = p
	rerandomization_experimental_design_search$objective = objective
	rerandomization_experimental_design_search$java_obj = java_obj
	rerandomization_experimental_design_search$verbose = verbose
	class(rerandomization_experimental_design_search) = "rerandomization_experimental_design_search"
	#if the user wants to run it immediately...
	if (start){
		startSearch(rerandomization_experimental_design_search)
	}
	#return the final object
	rerandomization_experimental_design_search
}

#' Prints a summary of a \code{rerandomization_experimental_design_search} object
#' 
#' @param x			The \code{rerandomization_experimental_design_search} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print rerandomization_experimental_design_search
#' @export
print.rerandomization_experimental_design_search = function(x, ...){
	progress = .jcall(x$java_obj, "I", "progress")
	time_elapsed = searchTimeElapsed(x)
	if (progress == 0){
		cat("No progress on the RerandomizationExperimentalDesign. Did you run \"startSearch?\"\n")
	} else if (progress == x$max_designs){
		cat("The search completed in", time_elapsed, "seconds.", progress, "vectors have been found.\n")
	} else {
		cat("The search has found ", progress, " vectors thus far (", round(progress / x$max_designs * 100), "%) in ", time_elapsed," seconds.\n", sep = "")
	}
}

#' Prints a summary of a \code{rerandomization_experimental_design_search} object
#' 
#' @param object		The \code{rerandomization_experimental_design_search} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary rerandomization_experimental_design_search
#' @export
summary.rerandomization_experimental_design_search = function(object, ...){
	print(object, ...)
}


#' Returns the results (thus far) of the rerandomization design search
#' 
#' @param obj 					The \code{rerandomization_experimental_design} object that is currently running the search
#' @param include_assignments	Do we include the assignments (takes time) and default is \code{FALSE}.
#' @param form					Which form should the assignments be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' obj = initRerandomizationExperimentalDesignObject(
#'   X,
#'   max_designs = 5,
#'   num_cores = 1,
#'   objective = "abs_sum_diff",
#'   obj_val_cutoff_to_include = Inf,
#'   start = TRUE,
#'   wait = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsRerandomizationSearch(obj, include_assignments = TRUE, form = "one_zero")
#' dim(res$ending_indicTs)
#' }
#' @export
resultsRerandomizationSearch = function(obj, include_assignments = FALSE, form = "one_zero"){
	obj_vals = .jcall(obj$java_obj, "[D", "getObjectiveVals")
	
	ending_indicTs = NULL
	if (include_assignments){
		num_completed = length(obj_vals)
		if (num_completed > 0){
			ordered_java_indices = as.integer(seq_len(num_completed) - 1)
			ordered_java_indices_j = .jarray(ordered_java_indices)
			ending_indicTs = .jcall(obj$java_obj, "[[I", "getEndingIndicTs", ordered_java_indices_j, simplify = TRUE)
			if (form == "pos_one_min_one"){
				if (length(ending_indicTs) > 0 && min(ending_indicTs) >= 0 && max(ending_indicTs) <= 1){
					ending_indicTs = (ending_indicTs - 0.5) * 2
				}
			} else if (length(ending_indicTs) > 0 && min(ending_indicTs) < 0){
				ending_indicTs = (ending_indicTs + 1) / 2
			}
		} else {
			ending_indicTs = matrix(integer(0), nrow = 0, ncol = obj$n)
		}
	}	
	
	rerandomization_experimental_design_search_results = list(
		obj_vals = obj_vals, 
		ending_indicTs = ending_indicTs
	)
	class(rerandomization_experimental_design_search_results) = "rerandomization_experimental_design_search_results"
	#return the final object
	rerandomization_experimental_design_search_results
}
#}

#' Compute Binary Matching Strcuture
#' 
#' This method creates an object of type binary_match_structure and will compute pairs. You can then
#' use the functions \code{initBinaryMatchExperimentalDesignSearchObject} and \code{resultsBinaryMatchSearch} 
#' to create randomized allocation vectors. For one column in X, we just sort to find the pairs trivially.
#' 
#' @param X						The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 								(one for each measurement on the subject). This is the design matrix you wish 
#' 								to search for a more optimal design.
#' @param compute_dist_matrix	The function that computes the distance matrix between every two observations in \code{X}, 
#' 								its only argument. The default is \code{NULL} signifying euclidean squared distance optimized in C++.
#' @param mahal_match			Match using Mahalanobis distance. Default is \code{FALSE}.
#' @param D						A distance matrix precomputed. The default is \code{NULL} indicating the distance matrix should be computed.
#' @param symmetry_tol			Tolerance for symmetry check on \code{D}. Default is \code{1e-12}.
#' @param use_safe_inverse		Should a regularized inverse be used for the Mahalanobis objective?
#' 								Default is \code{FALSE}.
#' @return						An object of type \code{binary_experimental_design} which can be further operated upon.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' bms = computeBinaryMatchStructure(X)
#' bms$indicies_pairs
#' }
#' @export
computeBinaryMatchStructure = function(X, mahal_match = FALSE, compute_dist_matrix = NULL, D = NULL, symmetry_tol = 1e-12, use_safe_inverse = FALSE){
	assertClass(X, "matrix")
	assertClass(compute_dist_matrix, "function", null.ok = TRUE)
	n = nrow(X)
	assertTRUE(n > 1)
	p = ncol(X)
	assertNumeric(symmetry_tol, lower = 0)
	assertLogical(use_safe_inverse)
	if (!is.null(D)){
		assertClass(D, "matrix")
		assertTRUE(nrow(D) == n)
		assertTRUE(ncol(D) == n)
		for (i in 1 : (n - 1)){ #ensure symmetric
			for (j in (i + 1) : n){
				assertTRUE(abs(D[i, j] - D[j, i]) <= symmetry_tol)
			}
		}
	}
	if (n %% 2 != 0){
		stop("Design matrix must have even rows to have equal treatments and controls")
	}
	
	if (is.null(compute_dist_matrix) & is.null(D) & p == 1){
		#we don't need to do anything except order them up
		indicies_pairs = matrix(order(X[, 1]), ncol = 2, byrow = TRUE)
	} else {
		if (is.null(compute_dist_matrix) & is.null(D)) {	
			if (mahal_match){
				if (use_safe_inverse){
					S_X_inv = safe_cov_inverse(X)
				} else {
					S_X_inv = solve(stats::var(X))
				}
				D = matrix(NA, nrow = n, ncol = n)
				for (i in 1 : (n - 1)){ #ensure symmetric
					for (j in (i + 1) : n){
						xdiff = X[i, ] - X[j, ]
						D[i, j] = D[j, i] = xdiff %*% (S_X_inv %*% xdiff)
					}
				}
			} else {
				#default is C++-optimized sqd euclidean distance function		
				D = compute_distance_matrix_cpp(X)				
			}
		} else if (is.null(D)){
			D = compute_dist_matrix(X)
		}
		#ensure diagonal is infinity in the R language
		diag(D) = .Machine$double.xmax
		
		#get the matching solution using the heuristic
		indicies_pairs = as.matrix(nbpMatching::nonbimatch(nbpMatching::distancematrix(D))$matches[, c("Group1.Row", "Group2.Row")])
		
		for (i in 1 : n){
			indicies_pairs[i, ] = sort(indicies_pairs[i, ])
		}	
		indicies_pairs = unique(indicies_pairs)
	}

	#now return information as an object (just a list)
	binary_match_structure = list()
	binary_match_structure$X = X
	binary_match_structure$n = n
	binary_match_structure$p = p
	binary_match_structure$compute_dist_matrix = compute_dist_matrix
	binary_match_structure$D = D
	binary_match_structure$indicies_pairs = indicies_pairs
	binary_match_structure$indices_pairs = indicies_pairs
	class(binary_match_structure) = "binary_match_structure"
	binary_match_structure
}

#' Begin a Binary Match Search
#' 
#' This method creates an object of type pairwise_matching_experimental_design_search and will immediately initiate
#' a search through allocation space for pairwise match designs based on the structure computed in the function \code{computeBinaryMatchStructure}. 
#' For debugging, you can use set the \code{seed} parameter and \code{num_cores = 1} to be assured of deterministic output.
#' 
#' @param binary_match_structure 	The \code{binary_experimental_design} object where the pairs are computed.
#' @param max_designs				How many random allocation vectors you wish to return. The default is 1000.
#' @param wait						Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 									default is \code{FALSE}.
#' @param start						Should we start searching immediately (default is \code{TRUE}).
#' @param num_cores					The number of CPU cores you wish to use during the search. The default is \code{1}.
#' @param seed						The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
#' 									the output will not be deterministic. Default is \code{NULL} for no seed set. 
#' @param prop_flips				Proportion of flips. Default is all. Lower for more correlated assignments (useful for research only).
#' @param verbose					Should the algorithm emit progress output? Default is \code{TRUE}.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' bms = computeBinaryMatchStructure(X)
#' bm = initBinaryMatchExperimentalDesignSearchObject(
#'   bms,
#'   max_designs = 4,
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = TRUE,
#'   seed = 1,
#'   verbose = FALSE
#' )
#' bm
#' }
#' @export
initBinaryMatchExperimentalDesignSearchObject = function(binary_match_structure, 
		max_designs = 1000, 
		wait = FALSE, 
		start = TRUE,
		num_cores = 1,
		seed = NULL,
		prop_flips = 1,
		verbose = TRUE){
	assertClass(binary_match_structure, "binary_match_structure")
	assertCount(max_designs, positive = TRUE)
	assertNumeric(prop_flips, lower = 0, upper = 1)
	assertLogical(verbose)
	
	if (prop_flips < 1){
		warning("prop_flips feature is not implemented yet")
	}
	binary_match_structure$verbose = verbose
		
	n = binary_match_structure$n
	if (2^(n / 2) < max_designs){
		stop(paste("The total number of unique pairwise matching vectors is only", 2^(n / 2), "which is less than the", max_designs, "you requested."))
	}
	
	if (max_designs > 2^(n / 2) / 2){
		warning(paste("The total number of unique pairwise matching vectors is only", 2^(n / 2), "which is less than double the", max_designs, "you requested.\nThis can be very slow as the space of designs gets crowded quickly and it is difficult to find new ones."))
	}
		
	#we are about to construct a PairwiseMatchingExperimentalDesign java object. First, let R garbage collect
	#to clean up previous java design objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	
	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("PairwiseMatchingExperimentalDesign.PairwiseMatchingExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setMaxDesigns", as.integer(max_designs))
	.jcall(java_obj, "V", "setNumCores", as.integer(num_cores))	
	if (!is.null(seed)){
		.jcall(java_obj, "V", "setSeed", as.integer(seed))
		if (num_cores != 1){
			warning("Setting the seed with multiple cores does not guarantee deterministic output.")
		}		
	}
	.jcall(java_obj, "V", "setN", as.integer(n))
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}
	
	#now upload the structure to the java object
	for (i in 1 : (n / 2)){
		.jcall(java_obj, "V", "setMatchPairIndicies", 
			as.integer(i - 1), 
			as.integer(binary_match_structure$indicies_pairs[i, ] - 1))		
	}
	
	#now return information as an object (just a list)
	pairwise_matching_experimental_design_search = list()
	pairwise_matching_experimental_design_search$binary_match_structure = binary_match_structure
	pairwise_matching_experimental_design_search$max_designs = max_designs
	pairwise_matching_experimental_design_search$start = start
	pairwise_matching_experimental_design_search$wait = wait
	pairwise_matching_experimental_design_search$num_cores = num_cores
	pairwise_matching_experimental_design_search$java_obj = java_obj
	pairwise_matching_experimental_design_search$verbose = verbose
	class(pairwise_matching_experimental_design_search) = "pairwise_matching_experimental_design_search"
	#if the user wants to run it immediately...
	if (start){
		startSearch(pairwise_matching_experimental_design_search)
	}
	#return the final object
	pairwise_matching_experimental_design_search
}

#' Binary Pair Match Search
#' 
#' Returns the results (thus far) of the binary pair match design search
#' 
#' @param obj 					The \code{pairwise_matching_experimental_design_search} object that is currently running the search
#' @param form					Which form should the assignments be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's. 
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(16), nrow = 8)
#' bms = computeBinaryMatchStructure(X)
#' bm = initBinaryMatchExperimentalDesignSearchObject(
#'   bms,
#'   max_designs = 4,
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = TRUE,
#'   seed = 1,
#'   verbose = FALSE
#' )
#' res = resultsBinaryMatchSearch(bm, form = "one_zero")
#' dim(res)
#' }
#' @export
resultsBinaryMatchSearch = function(obj, form = "one_zero"){
	assertClass(obj, "pairwise_matching_experimental_design_search")
	assertChoice(form, c("one_zero", "pos_one_min_one"))

	ending_indicTs = .jcall(obj$java_obj, "[[I", "getEndingIndicTs", simplify = TRUE)
	if (form == "pos_one_min_one"){
		if (length(ending_indicTs) > 0 && min(ending_indicTs) >= 0 && max(ending_indicTs) <= 1){
			ending_indicTs = (ending_indicTs - 0.5) * 2
		}
	} else if (length(ending_indicTs) > 0 && min(ending_indicTs) < 0){
		ending_indicTs = (ending_indicTs + 1) / 2
	}	
	ending_indicTs
}

#' Prints a summary of a \code{pairwise_matching_experimental_design_search} object
#' 
#' @param x			The \code{pairwise_matching_experimental_design_search} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print 	pairwise_matching_experimental_design_search
#' @export
print.pairwise_matching_experimental_design_search = function(x, ...){
	progress = .jcall(x$java_obj, "I", "progress")
	time_elapsed = searchTimeElapsed(x)
	if (progress == 0){
		cat("No progress on the PairwiseMatchingExperimentalDesign. Did you run \"startSearch?\"\n")
	} else if (progress == x$max_designs){
		cat("The search completed in", time_elapsed, "seconds.", progress, "vectors have been found.\n")
	} else {
		cat("The search has found ", progress, " vectors thus far (", round(progress / x$max_designs * 100), "%) in ", time_elapsed," seconds.\n", sep = "")
	}
}

#' Prints a summary of a \code{pairwise_matching_experimental_design_search} object
#' 
#' @param object		The \code{pairwise_matching_experimental_design_search} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary pairwise_matching_experimental_design_search
#' @export
summary.pairwise_matching_experimental_design_search = function(object, ...){
	print(object, ...)
}

#' Prints a summary of a \code{binary_match_structure} object
#' 
#' @param x			The \code{binary_match_structure} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print binary_match_structure
#' @export
print.binary_match_structure = function(x, ...){
	cat("The pairs have been computed. Now use the initBinaryMatchExperimentalDesignSearchObject function to create allocations.\n")
	invisible(x)
}

#' Prints a summary of a \code{binary_match_structure} object
#' 
#' @param object		The \code{binary_match_structure} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary binary_match_structure
#' @export
summary.binary_match_structure = function(object, ...){
	print(object, ...)
}

testthat::test_that("objective values rank by search method", {
  skip_on_cmd_check()
  with_immediate_failures({
    configs_raw = build_test_configs()
    config_keys = vapply(configs_raw, function(cfg) {
      paste(cfg$n, cfg$p, cfg$r, sep = "|")
    }, character(1))
    configs = configs_raw[!duplicated(config_keys)]
    configs = Filter(function(cfg) cfg$r == 200 && cfg$n == 96, configs)
    testthat::expect_true(length(configs) > 0)
    B_values = sort(unique(vapply(configs_raw, function(cfg) cfg$B, numeric(1))))
    objectives = c("abs_sum_diff", "mahal_dist")
    for (objective in objectives) {
      for (cfg in configs) {
        if (objective == "mahal_dist" && cfg$p == 10) {
          next
        }
        cat(
          "objective=", objective,
          " n=", cfg$n,
          " p=", cfg$p,
          " r=", cfg$r,
          " seed=", cfg$seed,
          "\n",
          sep = ""
        )

        X = make_X(cfg$n, cfg$p, seed = cfg$seed)
        compare_r = as.integer(cfg$r)
        if (objective == "mahal_dist") {
          X_eval = X
          inv_cov = safe_cov_inverse(X)
        } else {
          X_eval = standardize_data_matrix(X)
          inv_cov = NULL
        }

        complete_indicTs = complete_randomization(n = cfg$n, r = compare_r)
        complete_row_totals = rowSums(complete_indicTs)
        keep_complete = complete_row_totals > 0 & complete_row_totals < cfg$n
        if (!all(keep_complete)) {
          complete_indicTs = complete_indicTs[keep_complete, , drop = FALSE]
        }
        complete_obj = GreedyExperimentalDesign:::compute_objective_vals_cpp(
          X_eval,
          complete_indicTs,
          objective,
          inv_cov
        )
        complete_stat = median(complete_obj)
        cutoff = complete_stat

        ged = initGreedyExperimentalDesignObject(
          X,
          max_designs = compare_r,
          objective = objective,
          diagnostics = FALSE,
          wait = TRUE,
          num_cores = 1,
          seed = cfg$seed,
          verbose = FALSE,
          use_safe_inverse = TRUE
        )
        res_greedy = resultsGreedySearch(ged, max_vectors = compare_r, form = "one_zero")
        greedy_indicTs = res_greedy$ending_indicTs
        if (is.null(dim(greedy_indicTs))) {
          greedy_indicTs = matrix(greedy_indicTs, nrow = 1)
        }
        greedy_obj = GreedyExperimentalDesign:::compute_objective_vals_cpp(
          X_eval,
          greedy_indicTs,
          objective,
          inv_cov
        )
        greedy_stat = median(greedy_obj)

        bms = computeBinaryMatchStructure(X, use_safe_inverse = TRUE)
        bms$verbose = FALSE
        bm_unique = 2^(cfg$n / 2)
        bm_max_designs = min(compare_r, max(1L, floor(bm_unique / 2)))
        bm = initBinaryMatchExperimentalDesignSearchObject(
          bms,
          max_designs = bm_max_designs,
          wait = TRUE,
          num_cores = 1,
          seed = cfg$seed,
          verbose = FALSE
        )
        res_bm = resultsBinaryMatchSearch(bm, form = "one_zero")
        bm_indicTs = res_bm
        if (is.null(dim(bm_indicTs))) {
          bm_indicTs = matrix(bm_indicTs, nrow = 1)
        }
        bm_obj = GreedyExperimentalDesign:::compute_objective_vals_cpp(
          X_eval,
          bm_indicTs,
          objective,
          inv_cov
        )
        bm_stat = median(bm_obj)

        rerand = initRerandomizationExperimentalDesignObject(
          X,
          obj_val_cutoff_to_include = cutoff,
          max_designs = compare_r * 10L,
          objective = objective,
          wait = TRUE,
          num_cores = 1,
          seed = cfg$seed,
          verbose = FALSE,
          use_safe_inverse = TRUE
        )
        res_rerand = resultsRerandomizationSearch(rerand, include_assignments = FALSE)
        rerand_obj = res_rerand$obj_vals
        rerand_stat = median(rerand_obj)

        n = cfg$n
        r = compare_r
        block_obj_stats = numeric(0)
        for (B in B_values) {
          if (n %% B != 0) {
            next
          }
          n_B = n / B
          prop_T = floor(n_B / 2) / n_B
          if (prop_T <= 0 || prop_T >= 1) {
            next
          }
          block_designs = imbalanced_block_designs(
            n = n,
            prop_T = prop_T,
            B = B,
            r = r,
            seed = 123
          )
          block_obj = GreedyExperimentalDesign:::compute_objective_vals_cpp(
            X_eval,
            block_designs,
            objective,
            inv_cov
          )
          block_obj_stats = c(block_obj_stats, median(block_obj))
        }

        opt_obj = NULL
        if (cfg$n < 30 && cfg$n %% 2 == 0) {
          opt = initOptimalExperimentalDesignObject(
            X,
            objective = objective,
            wait = TRUE,
            num_cores = 1,
            verbose = FALSE,
            use_safe_inverse = TRUE
          )
          opt_res = resultsOptimalSearch(opt, num_vectors = min(2L, compare_r), form = "one_zero")
          opt_indicTs = opt_res$indicTs
          if (is.null(dim(opt_indicTs))) {
            opt_indicTs = matrix(opt_indicTs, nrow = 1)
          }
          opt_obj = GreedyExperimentalDesign:::compute_objective_vals_cpp(
            X_eval,
            opt_indicTs,
            objective,
            inv_cov
          )
          opt_stat = median(opt_obj)
        }

        for (b_idx in seq_along(block_obj_stats)) {
          #testthat::expect_true(block_obj_stats[b_idx] < complete_stat) #this seems to be hit or miss
          testthat::expect_true(greedy_stat <= block_obj_stats[b_idx])
          testthat::expect_true(bm_stat <= block_obj_stats[b_idx])
        }
        testthat::expect_true(greedy_stat < bm_stat)
        testthat::expect_true(bm_stat < rerand_stat)
        testthat::expect_true(rerand_stat < complete_stat)
        if (!is.null(opt_obj)) {
          tol = 1e-12
          testthat::expect_true(opt_stat <= greedy_stat + tol)
          testthat::expect_true(opt_stat <= bm_stat + tol)
          testthat::expect_true(opt_stat <= rerand_stat + tol)
          testthat::expect_true(opt_stat <= complete_stat + tol)
        }
      }
        }
      })
    })

#' Starts the parallelized greedy design search. 
#' 
#' Once begun, this function cannot be run again.
#' 
#' @param obj 		The \code{experimental_design} object that will be running the search
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 3,
#'   num_cores = 1,
#'   start = FALSE,
#'   wait = FALSE,
#'   objective = "abs_sum_diff",
#'   verbose = FALSE
#' )
#' startSearch(ged)
#' stopSearch(ged)
#' }
#' @export
startSearch = function(obj){
	if (.jcall(obj$java_obj, "Z", "began")){
		stop("Search Already begun.")
	}
	.jcall(obj$java_obj, "V", "beginSearch")
}

#' Stops the parallelized greedy design search. 
#' 
#' Once stopped, it cannot be restarted.
#' 
#' @param obj 		The \code{experimental_design} object that is currently running the search
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 3,
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = FALSE,
#'   objective = "abs_sum_diff",
#'   verbose = FALSE
#' )
#' stopSearch(ged)
#' }
#' @export
stopSearch = function(obj){
	.jcall(obj$java_obj, "V", "stopSearch")
}

#' Generates a design matrix with standardized predictors. 
#' 
#' This function is useful for debugging.
#' 
#' @param n					Number of rows in the design matrix 
#' @param p 				Number of columns in the design matrix
#' @param covariate_gen		The function to use to draw the covariate realizations (assumed to be iid).
#' 							This defaults to \code{rnorm} for $N(0,1)$ draws.
#' @param ...				Optional arguments to be passed to the \code{covariate_dist} function.
#' @return 					THe design matrix
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' X = generate_stdzied_design_matrix(n = 6, p = 2)
#' colMeans(X)
#' }
#' @export
generate_stdzied_design_matrix = function(n = 50, p = 1, covariate_gen = rnorm, ...){
	X = matrix(covariate_gen(n * p, ...), nrow = n, ncol = p)
	#now standardize the matrix to make things easier later
	standardize_data_matrix(X)
}


#' Returns the amount of time elapsed
#' 
#' @param obj 		The \code{experimental_design} object that is currently running the search
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(20), nrow = 10)
#' ged = initGreedyExperimentalDesignObject(
#'   X,
#'   max_designs = 1,
#'   num_cores = 1,
#'   start = TRUE,
#'   wait = TRUE,
#'   objective = "abs_sum_diff",
#'   verbose = FALSE
#' )
#' searchTimeElapsed(ged)
#' }
#' @export
searchTimeElapsed = function(obj){
	.jcall(obj$java_obj, "I", "timeElapsedInSeconds")
}

#' Computes a numerically stable inverse of a covariance matrix
#'
#' @param X			The n x p design matrix
#' @param ridge		Initial ridge penalty added to the diagonal
#' @param max_ridge_steps	Maximum number of ridge escalation attempts
#' @return			The inverse covariance matrix
#'
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' X = matrix(rnorm(20), nrow = 10)
#' Sinv = safe_cov_inverse(X)
#' dim(Sinv)
#' }
#' @export
safe_cov_inverse = function(X, ridge = 1e-8, max_ridge_steps = 6){
	S = stats::var(X)
	if (!all(is.finite(S))){
		stop("Covariance matrix contains non-finite values.")
	}
	Sinv = tryCatch(solve(S), error = function(e) NULL)
	if (!is.null(Sinv)){
		return(Sinv)
	}
	ridge_val = ridge
	for (i in seq_len(max_ridge_steps)){
		Sinv = tryCatch(solve(S + diag(ridge_val, ncol(S))), error = function(e) NULL)
		if (!is.null(Sinv)){
			return(Sinv)
		}
		ridge_val = ridge_val * 10
	}
	sv = svd(S)
	if (length(sv$d) == 0){
		stop("Covariance matrix is empty.")
	}
	tol = max(dim(S)) * max(sv$d) * .Machine$double.eps
	d_inv = ifelse(sv$d > tol, 1 / sv$d, 0)
	sv$v %*% (d_inv * t(sv$u))
}

#' Computes Objective Value From Allocation Vector
#' 
#' Returns the objective value given a design vector as well an an objective function.
#' This is sometimes duplicated in Java. However, within Java, tricks are played to make
#' optimization go faster so Java's objective values may not always be the same as the true
#' objective function (e.g. logs or constants dropped).
#' 
#' @param X 		 	The n x p design matrix
#' @param indic_T		The n-length binary allocation vector
#' @param objective		The objective function to use. Default is \code{abs_sum_diff} and the other option is 
#' 						\code{mahal_dist}.
#' @param inv_cov_X		Optional: the inverse sample variance covariance matrix. Use this
#' 						argument if you will be doing many calculations since passing this
#' 						in will cache this data.
#' @param use_safe_inverse	Should a regularized inverse be used for the Mahalanobis objective?
#' 							Default is \code{FALSE}.
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' X = matrix(rnorm(12), nrow = 6)
#' indic_T = c(1, 0, 1, 0, 1, 0)
#' compute_objective_val(X, indic_T, objective = "abs_sum_diff")
#' }
#' @export
compute_objective_val = function(X, indic_T, objective = "abs_sum_diff", inv_cov_X = NULL, use_safe_inverse = FALSE){
	assertLogical(use_safe_inverse)
	if (!isTRUE(all.equal(sort(unique(indic_T)), c(0, 1)))){
		stop("indic_T must be binary")
	}
	X_T = X[indic_T == 1, , drop = FALSE] #coerce as matrix in order to matrix multiply later
	X_C = X[indic_T == 0, , drop = FALSE] #coerce as matrix in order to matrix multiply later
	X_T_bar = colMeans(X_T)
	X_C_bar = colMeans(X_C)	
	
	if (objective == "abs_sum_diff"){
		s_j_s = apply(X, 2, sd)
		sum(abs((X_T_bar - X_C_bar) / s_j_s))
	} else if (objective == "mahal_dist"){
		#saves computation to pass it in if you're doing a lot of them in a row
		if (is.null(inv_cov_X)){
			if (use_safe_inverse){
				inv_cov_X = safe_cov_inverse(X)
			} else {
				inv_cov_X = solve(stats::var(X))
			}
		}	
		X_T_bar_minus_X_C_bar = as.matrix(X_T_bar - X_C_bar) #need to matricize for next computation
		as.numeric(t(X_T_bar_minus_X_C_bar) %*% inv_cov_X %*% X_T_bar_minus_X_C_bar)
	} else {
		stop("objective invalid.")
	}
}

#' Standardizes the columns of a data matrix.
#' 
#' @param X 		 	The n x p design matrix
#' @return				The n x p design matrix with columns standardized
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' X = matrix(rnorm(12), nrow = 6)
#' Xstd = standardize_data_matrix(X)
#' colMeans(Xstd)
#' }
#' @export
standardize_data_matrix = function(X){
	apply(X, 2, function(xj){(xj - mean(xj)) / sd(xj)})
}

#private
verify_objective_function = function(objective, Kgram = NULL, n = NULL){
	if (objective != "mahal_dist" && objective != "abs_sum_diff" && objective != "kernel"){
		stop("Objective function must be one of the following:\n  mahal_dist\n  abs_sum_diff\n  kernel\n\n")
	}
	if (objective == "kernel"){
		if (is.null(Kgram) || is.null(n)){
			stop("You must specify a gram matrix \"Kgram\" and \"n\".\n")
		}
		if (!inherits(Kgram, "kernelMatrix") && !inherits(Kgram, "matrix")){
			stop("The gram matrix must be type kernelMatrix or type matrix.\n")
		}
		if (!all.equal(dim(Kgram), c(n, n))){
			stop("The gram matrix must have dimension n x n.\n")
		}
	}
	if (!is.null(Kgram) && objective != "kernel"){
		stop("If you specify a gram matrix, you must specify the \"kernel\" objective.\n")
	}
}




####################C++ functions

#' Generates homogeneous block design allocations rapidly
#' 
#' @param B 		 	The number of blocks in the design
#' @param nR			The number of allocation vectors
#' @param dummy_block   The subvector of allocations in each block that will be permuted
#' @return				A matrix with rows being the nR random block allocation of sample size B x length(dummy_block).
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' generate_block_design_cpp_wrap(B = 2, nR = 3, dummy_block = c(1, 0))
#' }
#' @export
generate_block_design_cpp_wrap <- function(B, nR, dummy_block) {
    do.call(cbind, .Call('_GreedyExperimentalDesign_generate_block_design_cpp', PACKAGE = 'GreedyExperimentalDesign', B, nR, dummy_block))
}


#' Computes a Euclidean-squared distance matrix rapidly
#' 
#' @param X 		 	A numeric matrix with n rows representing each subject and p columns which are measurements on each subject
#' @return				The n x n Euclidean distances squared
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' X = matrix(c(0, 1, 2, 3), nrow = 2)
#' compute_distance_matrix_cpp_wrap(X)
#' }
#' @export
compute_distance_matrix_cpp_wrap <- function(X) {
    .Call('_GreedyExperimentalDesign_compute_distance_matrix_cpp', PACKAGE = 'GreedyExperimentalDesign', X)
}

#' Shuffles a vector rapidly
#' 
#' @param w 		 	The vector to be shuffled
#' @param seed		Optional integer seed; use NA to draw from the system clock
#' @return				The vector with elements shuffled
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' shuffle_cpp_wrap(1:5, seed = 1)
#' }
#' @export
shuffle_cpp_wrap <- function(w, seed = NA_integer_) {
    .Call('_GreedyExperimentalDesign_shuffle_cpp', PACKAGE = 'GreedyExperimentalDesign', w, seed)
}

#' Tests if a vector has all elements the same
#' 
#' @param w 		 	The vector to be queried
#' @return				A boolean if it has all same elements
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' all_elements_same_cpp_wrap(c(1, 1, 1))
#' all_elements_same_cpp_wrap(c(1, 2, 1))
#' }
#' @export
all_elements_same_cpp_wrap <- function(w) {
    .Call('_GreedyExperimentalDesign_all_elements_same_cpp', PACKAGE = 'GreedyExperimentalDesign', w)
}

#' Create all binary Y's convenience function using a randomized design
#' 
#' @param pCs			Control-group success probabilities (length \code{two_n})
#' @param pTs			Treatment-group success probabilities (length \code{two_n})
#' @param W				Assignment matrix with \code{nY} rows and \code{two_n} columns
#' @param two_n			Total number of units
#' @param nY				Number of Y vectors to generate
#' @return				A matrix of boolean Y's
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' pCs = rep(0.2, 4)
#' pTs = rep(0.8, 4)
#' W = matrix(c(1, 0, 1, 0, 0, 1, 0, 1), nrow = 2, byrow = TRUE)
#' create_all_ys_cpp_wrap(pCs, pTs, W, two_n = 4, nY = 2)
#' }
#' @export
create_all_ys_cpp_wrap <- function(pCs, pTs, W, two_n, nY) {
    .Call('_GreedyExperimentalDesign_create_all_ys_cpp', PACKAGE = 'GreedyExperimentalDesign', pCs, pTs, W, two_n, nY)
}

#' Create PM designs
#' 
#' @param indicies_pairs	A matrix of n x 2 indicies where each row is a pair of subjects' indicies
#' @param n					Half the number of subjects i.e. the number of pairs
#' @param r					The number of assignments to generate
#' @return					A matrix of r x 2n PM designs of +1/-1 assignments
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' indicies_pairs = matrix(c(1, 2, 3, 4), ncol = 2, byrow = TRUE)
#' gen_pm_designs_cpp_wrap(indicies_pairs, n = 2, r = 3)
#' }
#' @export
gen_pm_designs_cpp_wrap <- function(indicies_pairs, n, r) {
    .Call('_GreedyExperimentalDesign_gen_pm_designs_cpp', PACKAGE = 'GreedyExperimentalDesign', indicies_pairs, n, r)
}

# internal helper to avoid hard failure when older Java jars omit setVerbose
set_verbose_if_available <- function(java_obj, verbose) {
	tryCatch(
		.jcall(java_obj, "V", "setVerbose", verbose),
		error = function(e) {
			if (grepl("setVerbose", conditionMessage(e), fixed = TRUE)) {
				return(invisible(FALSE))
			}
			stop(e)
		}
	)
	invisible(TRUE)
}

# internal helper to avoid unnecessary coercion for w_diff
compute_indicTs_from_pairs <- function(pairs, w_diff, n) {
	if (is.logical(w_diff)) {
		compute_indicTs_from_pairs_lgl_cpp(pairs, w_diff, n)
	} else if (is.raw(w_diff)) {
		compute_indicTs_from_pairs_raw_cpp(pairs, w_diff, n)
	} else {
		if (!anyNA(w_diff) && all(w_diff == 0L | w_diff == 1L)) {
			return(compute_indicTs_from_pairs_lgl_cpp(pairs, w_diff == 1L, n))
		}
		compute_indicTs_from_pairs_cpp(pairs, w_diff, n)
	}
}

source("helper-package.R")
test_that("core functions work across sizes", {
  skip_on_cmd_check()
  with_immediate_failures({
    configs_raw = build_test_configs()
    config_keys = vapply(configs_raw, function(cfg) {
      paste(cfg$n, cfg$p, cfg$r, cfg$prop_T, sep = "|")
    }, character(1))
    configs = configs_raw[!duplicated(config_keys)]
    B_values = sort(unique(vapply(configs_raw, function(cfg) cfg$B, numeric(1))))
    objectives = c("abs_sum_diff", "mahal_dist")

    for (objective in objectives) {
      for (cfg in configs) {
        if (objective == "mahal_dist" && cfg$p >= 10) { #too slow
          next
        }
        X = make_X(cfg$n, cfg$p, seed = cfg$seed)
        X1 = X[, 1, drop = FALSE]
        max_vectors = max(1L, min(as.integer(cfg$r), 10L))
        cat(
          "objective=", objective,
          " n=", cfg$n,
          " p=", cfg$p,
          " r=", cfg$r,
          " prop_T=", cfg$prop_T,
          " seed=", cfg$seed,
          "\n",
          sep = ""
        )

    Xstd = standardize_data_matrix(X)
    expect_equal(dim(Xstd), dim(X))
    expect_true(all(abs(colMeans(Xstd)) < 1e-10))
    expect_true(all(abs(apply(Xstd, 2, sd) - 1) < 1e-10))

    Xgen = generate_stdzied_design_matrix(n = cfg$n, p = cfg$p, covariate_gen = rnorm)
    expect_equal(dim(Xgen), dim(X))
    expect_true(all(abs(colMeans(Xgen)) < 1e-10))

    w = 1:cfg$n
    w1 = shuffle_cpp_wrap(w, seed = cfg$seed)
    w2 = shuffle_cpp_wrap(w, seed = cfg$seed)
    expect_equal(w1, w2)
    expect_equal(sort(w1), w)
    expect_true(all_elements_same_cpp_wrap(rep(2, 5)))
    expect_false(all_elements_same_cpp_wrap(c(1, 2, 1)))

    Xsmall = matrix(c(0, 0, 1, 1, 2, 2), ncol = 2, byrow = TRUE)
    D = compute_distance_matrix_cpp_wrap(Xsmall)
    expect_equal(dim(D), c(3, 3))
    expect_true(is.na(diag(D)[1]))
    expect_equal(D[1, 2], D[2, 1])
    K = compute_gram_matrix(Xsmall, kernel_type = "vanilla")
    expect_equal(dim(K), c(3, 3))
    expect_equal(K, t(K))

    indic = c(rep(1, cfg$n / 2), rep(0, cfg$n / 2))
    obj_val = compute_objective_val(
      X,
      indic,
      objective = objective,
      use_safe_inverse = TRUE
    )
    expect_true(is.numeric(obj_val))

    designs = complete_randomization(n = cfg$n, r = cfg$r)
    metrics = compute_randomization_metrics(t(designs))
    expect_true(is.list(metrics))
    expect_equal(dim(metrics$p_hat_ijs), c(cfg$n, cfg$n))
    expect_true(is.numeric(metrics$rand_entropy_metric))

    mat = complete_randomization(n = cfg$n, r = cfg$r)
    expect_binary_matrix(mat, nrow_expected = cfg$r, ncol_expected = cfg$n)

    forced = complete_randomization_with_forced_balanced(n = cfg$n, r = cfg$r, seed = cfg$seed)
    expect_binary_matrix(forced, nrow_expected = cfg$r, ncol_expected = cfg$n)
    expect_equal(rowSums(forced), rep(cfg$n / 2, cfg$r))

    imbal = imbalanced_complete_randomization(n = cfg$n, prop_T = cfg$prop_T, r = cfg$r, seed = cfg$seed)
    expect_binary_matrix(imbal, nrow_expected = cfg$r, ncol_expected = cfg$n)
    n_T = as.integer(round(cfg$n * cfg$prop_T))
    expect_equal(rowSums(imbal), rep(n_T, cfg$r))

    for (B in B_values) {
      if (cfg$n %% B != 0) {
        next
      }
      n_B = cfg$n / B
      if (abs(n_B * cfg$prop_T - round(n_B * cfg$prop_T)) > 1e-8) {
        next
      }
      blocks = imbalanced_block_designs(
        n = cfg$n,
        prop_T = cfg$prop_T,
        B = B,
        r = cfg$r,
        seed = cfg$seed
      )
      expect_binary_matrix(blocks, nrow_expected = cfg$r, ncol_expected = cfg$n)
      expect_equal(rowSums(blocks), rep(n_T, cfg$r))

      dummy_block = rep(c(1, 0), length.out = cfg$n / B)
      block_designs = generate_block_design_cpp_wrap(B = B, nR = cfg$r, dummy_block = dummy_block)
      expect_equal(dim(block_designs), c(cfg$r, cfg$n))
      expect_equal(rowSums(block_designs), rep(sum(dummy_block) * B, cfg$r))

      SigmaW = gen_var_cov_matrix_block_designs(n = cfg$n, prop_T = cfg$prop_T, B = B)
      expect_equal(dim(SigmaW), c(cfg$n, cfg$n))
      expect_equal(SigmaW, t(SigmaW))
    }

    pairs = matrix(c(1, 2, 3, 4), ncol = 2, byrow = TRUE)
    Wpm = gen_pm_designs_cpp_wrap(pairs, n = 2, r = cfg$r)
    expect_equal(dim(Wpm), c(cfg$r, 4))
    expect_true(all(Wpm %in% c(-1, 1)))

    pCs = c(0.1, 0.2, 0.3, 0.4)
    pTs = c(0.9, 0.8, 0.7, 0.6)
    Wbin = matrix(c(1, 0, 1, 0, 0, 1, 0, 1), nrow = 2, byrow = TRUE)
    Y = create_all_ys_cpp_wrap(pCs, pTs, Wbin, two_n = 4, nY = 2)
    expect_equal(dim(Y), c(2, 4))
    expect_true(is.logical(Y))

    res1 = optimize_asymmetric_treatment_assignment(n = cfg$n)
    expect_equal(res1$nT + res1$nC, cfg$n)
    res2 = optimize_asymmetric_treatment_assignment(n = cfg$n, c_treatment = 2, c_control = 1)
    expect_equal(res2$nT + res2$nC, cfg$n)
    expect_true(res2$nT > res2$nC)

    if (cfg$n >= 4) {
      Xhad = X[1:4, , drop = FALSE]
      W = hadamardExperimentalDesign(Xhad, strict = TRUE, form = "one_zero")
      expect_true(is.matrix(W))
      expect_true(all(W %in% c(0, 1)))
      expect_true(all(rowSums(W) == nrow(Xhad) / 2))
    }

    r_cur = max(2, cfg$r)
    Wcur = matrix(sample(c(-1, 1), r_cur * cfg$n, replace = TRUE), nrow = r_cur, ncol = cfg$n)
    res_cur = greedy_orthogonalization_curation(Wcur, Rmin = 2, verbose = FALSE)
    expect_true(is.list(res_cur))
    expect_equal(ncol(res_cur$Wsorted), cfg$n)
    R0 = min(max(2, floor(r_cur / 2)), r_cur)
    Wcur2 = greedy_orthogonalization_curation2(Wcur, R0 = R0, verbose = FALSE)
    expect_true(is.matrix(Wcur2))
    expect_equal(ncol(Wcur2), cfg$n)

    ged = initGreedyExperimentalDesignObject(
      X,
      max_designs = cfg$r,
      objective = objective,
      diagnostics = TRUE,
      wait = TRUE,
      num_cores = 1,
      seed = cfg$seed,
      verbose = FALSE,
      use_safe_inverse = TRUE
    )
    expect_output(print(ged))
    expect_output(summary(ged))
    res_greedy = resultsGreedySearch(ged, max_vectors = max_vectors)
    expect_binary_matrix(res_greedy$ending_indicTs, nrow_expected = max_vectors, ncol_expected = cfg$n)
    expect_equal(rowSums(res_greedy$ending_indicTs), rep(cfg$n / 2, max_vectors))
    with_plot_device(plot_obj_val_by_iter(res_greedy))
    with_plot_device(plot_obj_val_order_statistic(ged, order_stat = 1, skip_every = 1))
    with_plot_device(plot(ged))

    mk = suppressWarnings(initGreedyMultipleKernelExperimentalDesignObject(
      X,
      max_designs = cfg$r,
      kernel_pre_num_designs = cfg$r,
      kernel_names = c("mahalanobis", "gaussian"),
      wait = TRUE,
      num_cores = 1,
      seed = cfg$seed,
      diagnostics = TRUE,
      verbose = FALSE,
      use_safe_inverse = TRUE
    ))
    expect_output(print(mk))
    expect_output(summary(mk))
    res_mk = resultsMultipleKernelGreedySearch(mk, max_vectors = max_vectors, form = "one_zero")
    expect_binary_matrix(res_mk$ending_indicTs, nrow_expected = max_vectors, ncol_expected = cfg$n)
    expect_equal(rowSums(res_mk$ending_indicTs), rep(cfg$n / 2, max_vectors))
    with_plot_device(plot(mk))

    bms = computeBinaryMatchStructure(X, use_safe_inverse = TRUE)
    bms$verbose = FALSE
    bm_unique = 2^(cfg$n / 2)
    bm_max_designs = min(cfg$r, max(1L, floor(bm_unique / 2)))
    bm = initBinaryMatchExperimentalDesignSearchObject(
      bms,
      max_designs = bm_max_designs,
      wait = TRUE,
      num_cores = 1,
      seed = cfg$seed,
      verbose = FALSE
    )
    expect_output(print(bm))
    expect_output(summary(bm))
    res_bm = resultsBinaryMatchSearch(bm, form = "one_zero")
    expect_binary_matrix(res_bm, nrow_expected = bm_max_designs, ncol_expected = cfg$n)
    expect_equal(rowSums(res_bm), rep(cfg$n / 2, bm_max_designs))

    if (cfg$n %% 4 == 0) {
      bm_greedy = initBinaryMatchFollowedByGreedyExperimentalDesignSearchObject(
        X,
        diff_method = TRUE,
        max_designs = cfg$r,
        wait = TRUE,
        num_cores = 1,
        seed = cfg$seed,
        objective = objective,
        diagnostics = TRUE,
        verbose = FALSE
      )
      expect_output(print(bm_greedy))
      expect_output(summary(bm_greedy))
      res_bm_greedy = resultsBinaryMatchThenGreedySearch(
        bm_greedy,
        num_vectors = max_vectors,
        compute_obj_vals = TRUE,
        use_safe_inverse = TRUE
      )
      expect_binary_matrix(res_bm_greedy$indicTs, nrow_expected = max_vectors, ncol_expected = cfg$n)
      expect_equal(rowSums(res_bm_greedy$indicTs), rep(cfg$n / 2, max_vectors))

      bm_rerand = initBinaryMatchFollowedByRerandomizationDesignSearchObject(
        X,
        max_designs = cfg$r,
        wait = TRUE,
        num_cores = 1,
        seed = cfg$seed,
        objective = objective,
        obj_val_cutoff_to_include = Inf,
        verbose = FALSE
      )
      expect_output(print(bm_rerand))
      expect_output(summary(bm_rerand))
      res_bm_rerand = resultsBinaryMatchThenRerandomizationSearch(
        bm_rerand,
        num_vectors = max_vectors,
        compute_obj_vals = TRUE,
        use_safe_inverse = TRUE
      )
      expect_binary_matrix(res_bm_rerand$indicTs, ncol_expected = cfg$n)
      expect_true(nrow(res_bm_rerand$indicTs) >= max_vectors)
      expect_equal(rowSums(res_bm_rerand$indicTs), rep(cfg$n / 2, nrow(res_bm_rerand$indicTs)))
    } else {
      expect_error(
        initBinaryMatchFollowedByGreedyExperimentalDesignSearchObject(
          X,
          diff_method = TRUE,
          max_designs = cfg$r,
          wait = TRUE,
          num_cores = 1,
          seed = cfg$seed,
          objective = objective,
          diagnostics = TRUE,
          verbose = FALSE
        ),
        "divisible by four"
      )
      expect_error(
        initBinaryMatchFollowedByRerandomizationDesignSearchObject(
          X,
          max_designs = cfg$r,
          wait = TRUE,
          num_cores = 1,
          seed = cfg$seed,
          objective = objective,
          obj_val_cutoff_to_include = Inf,
          verbose = FALSE
        ),
        "divisible by four"
      )
    }

    rd = initRerandomizationExperimentalDesignObject(
      X,
      obj_val_cutoff_to_include = Inf,
      max_designs = cfg$r,
      objective = objective,
      wait = TRUE,
      num_cores = 1,
      seed = cfg$seed,
      verbose = FALSE,
      use_safe_inverse = TRUE
    )
    expect_output(print(rd))
    expect_output(summary(rd))
    res_rd = resultsRerandomizationSearch(rd, include_assignments = TRUE, form = "one_zero")
    expect_equal(dim(res_rd$ending_indicTs), c(cfg$r, cfg$n))
    expect_equal(rowSums(res_rd$ending_indicTs), rep(cfg$n / 2, cfg$r))

    Xopt = X[1:4, , drop = FALSE]
    if (objective == "abs_sum_diff" || cfg$p < nrow(Xopt)) {
      od = initOptimalExperimentalDesignObject(
        Xopt,
        objective = objective,
        wait = TRUE,
        num_cores = 1,
        verbose = FALSE,
        use_safe_inverse = TRUE
      )
      expect_output(print(od))
      expect_output(summary(od))
      res_opt = resultsOptimalSearch(od, num_vectors = 2, form = "one_zero")
      expect_equal(dim(res_opt$indicTs), c(2, 4))
    }

    if (objective == "abs_sum_diff" && cfg$p == 1) {
      kd = initKarpExperimentalDesignObject(X1, wait = TRUE, balanced = TRUE, verbose = FALSE)
      expect_output(print(kd))
      expect_output(summary(kd))
      res_kd = resultsKarpSearch(kd)
      expect_equal(length(res_kd$indicT), cfg$n)
    }
      }
    }
  })
})

#' Begin Karp Search
#' 
#' This method creates an object of type karp_experimental_design and will immediately initiate
#' a search through allocation  space. Note that the Karp search only works 
#' for one covariate (i.e. $p=1$) and the objective "abs_sum_diff".
#' 
#' @param X					The design matrix with $n$ rows (one for each subject) and $p$ columns 
#' 							(one for each measurement on the subject). This is the design matrix you wish 
#' 							to search for a more karp design.
#' @param balanced			Should the final vector be balanced? Default and recommended is \code{TRUE}.
#' @param wait				Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
#' 							deafult is \code{FALSE}.
#' @param start				Should we start searching immediately (default is \code{TRUE}).
#' @param verbose			Should the algorithm emit progress output? Default is \code{TRUE}.
#' @return					An object of type \code{karp_experimental_design_search} which can be further operated upon
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(10), nrow = 10)
#' kobj = initKarpExperimentalDesignObject(
#'   X,
#'   start = TRUE,
#'   wait = TRUE,
#'   balanced = TRUE,
#'   verbose = FALSE
#' )
#' kobj
#' }
#' @export
initKarpExperimentalDesignObject = function(X,
		wait = FALSE, 
		balanced = TRUE,
		start = TRUE,
		verbose = TRUE){
	n = nrow(X)
	if (n %% 2 != 0 && balanced){
		stop("Design matrix must have even rows to have equal treatments and controls if you are requiring balance.")
	}
	p = ncol(X)
	
	if (p > 1){
		stop("Karp search only works for p = 1.")
	}
	assertLogical(verbose)
	
	#we are about to construct a KarpExperimentalDesign java object. First, let R garbage collect
	#to clean up previous objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	
	
	#now go ahead and create the Java object and set its information
	java_obj = .jnew("KarpExperimentalDesign.KarpExperimentalDesign")
	set_verbose_if_available(java_obj, verbose)
	.jcall(java_obj, "V", "setN", as.integer(n))
	.jcall(java_obj, "V", "setP", as.integer(p))
	if (wait){
		.jcall(java_obj, "V", "setWait")
	}
	if (balanced){
		.jcall(java_obj, "V", "setBalanced")
	}
	
	#feed in the data
	.jcall(java_obj, "V", "setDataMatrix", as.double(X), as.integer(n), as.integer(p))
	
	#now return information as an object (just a list)
	karp_experimental_design_search = list()
	karp_experimental_design_search$start = start
	karp_experimental_design_search$wait = wait
	karp_experimental_design_search$balanced = balanced
	karp_experimental_design_search$X = X
	karp_experimental_design_search$n = n
	karp_experimental_design_search$java_obj = java_obj
	karp_experimental_design_search$verbose = verbose
	class(karp_experimental_design_search) = "karp_experimental_design_search"
	#if the user wants to run it immediately...
	if (start){
		startSearch(karp_experimental_design_search)
	}
	#return the final object
	karp_experimental_design_search
}

#' Prints a summary of a \code{karp_experimental_design_search} object
#' 
#' @param x			The \code{karp_experimental_design_search} object to be summarized in the console
#' @param ...		Other parameters to pass to the default print function
#' 
#' @author 			Adam Kapelner
#' @method print karp_experimental_design_search
#' @export
print.karp_experimental_design_search = function(x, ...){
	progress = .jcall(x$java_obj, "D", "progress")
	time_elapsed = searchTimeElapsed(x)
	if (progress == 0){
		cat("No progress on the KarpExperimentalDesign. Did you run \"startKarpSearch?\"\n")
	} else if (progress < 1){
		cat("The search is", round(progress * 100, 2), "% complete.\n")
	} else {
		cat("The search is complete.\n")
	}
}

#' Prints a summary of a \code{karp_experimental_design_search} object
#' 
#' @param object		The \code{karp_experimental_design_search} object to be summarized in the console
#' @param ...			Other parameters to pass to the default summary function
#' 
#' @author 				Adam Kapelner
#' @method summary karp_experimental_design_search
#' @export
summary.karp_experimental_design_search = function(object, ...){
	print(object, ...)
}

#' Returns the results (thus far) of the karp design search
#' 
#' @param obj 			The \code{karp_experimental_design} object that is currently running the search
#' 
#' @author Adam Kapelner
#' @examples
#' \dontrun{
#' set.seed(1)
#' X = matrix(rnorm(10), nrow = 10)
#' kobj = initKarpExperimentalDesignObject(
#'   X,
#'   start = TRUE,
#'   wait = TRUE,
#'   balanced = TRUE,
#'   verbose = FALSE
#' )
#' res = resultsKarpSearch(kobj)
#' res$obj_val
#' }
#' @export
resultsKarpSearch = function(obj){
	obj_val = .jcall(obj$java_obj, "D", "getKarpObjectiveVal")
	indicT = .jcall(obj$java_obj, "[I", "getKarpIndicT", simplify = TRUE)
	list(
		obj_val = obj_val,
		indicT = indicT
	)
}

make_X <- function(n, p, seed = 1) {
  set.seed(seed)
  matrix(rnorm(n * p), nrow = n, ncol = p)
}

expect_binary_matrix <- function(mat, nrow_expected = NULL, ncol_expected = NULL) {
  if (!is.null(nrow_expected)) {
    testthat::expect_equal(nrow(mat), nrow_expected)
  }
  if (!is.null(ncol_expected)) {
    testthat::expect_equal(ncol(mat), ncol_expected)
  }
  testthat::expect_true(all(mat %in% c(0, 1)))
}

with_plot_device <- function(expr) {
  file = tempfile(fileext = ".png")
  png(file)
  on.exit(dev.off(), add = TRUE)
  force(expr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gurobi_numerical_optimization_search.R
\name{initGurobiNumericalOptimizationExperimentalDesignObject}
\alias{initGurobiNumericalOptimizationExperimentalDesignObject}
\title{Begin Gurobi Optimized Search}
\usage{
initGurobiNumericalOptimizationExperimentalDesignObject(
  X = NULL,
  objective = "mahal_dist",
  Kgram = NULL,
  num_cores = 2,
  w_0 = NULL,
  initial_time_limit_sec = 5 * 60,
  restart_time_limit_sec = 60,
  max_number_of_restarts = 0,
  max_no_good_cuts = 0,
  verbose = TRUE,
  gurobi_params = list(),
  use_safe_inverse = FALSE,
  r,
  pool_solutions = NULL,
  pool_gap = 0.2,
  pool_gap_abs = NULL,
  pool_search_mode = 2,
  mip_gap = 1e-04,
  mip_gap_abs = 1e-10,
  mip_focus = 1,
  heuristics = 0.2,
  cuts = 2,
  presolve = 2
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{objective}{The objective function to use when searching design space. This is a string
with valid values "\code{mahal_dist}" (the default) or "\code{kernel}".}

\item{Kgram}{If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.}

\item{num_cores}{Number of cores to use during search. Default is \code{2}.}

\item{w_0}{The initial starting location (optional).}

\item{initial_time_limit_sec}{The maximum amount of time the optimizer can run for in seconds. The default is \code{5 * 60}.}

\item{restart_time_limit_sec}{The maximum amount of time each restart can run for in seconds. The default is \code{60}.}

\item{max_number_of_restarts}{The maximum number of restarts to attempt if too few unique solutions are returned.
Default is \code{0}.}

\item{max_no_good_cuts}{The maximum number of no-good cuts to attempt. Default is \code{0} (disabled).}

\item{verbose}{Should Gurobi log to console? Default is \code{TRUE}.}

\item{gurobi_params}{A list of optional parameters to be passed to Gurobi (see their documentation online).}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}

\item{r}{Number of solution vectors to request from the Gurobi pool.}

\item{pool_solutions}{Number of solutions to request from the Gurobi pool. Defaults to \code{10 * r}.}

\item{pool_gap}{Relative optimality gap for the pool. Default is \code{0.2}. Use \code{NULL} to skip.}

\item{pool_gap_abs}{Absolute optimality gap for the pool. Default is \code{NULL} to skip.}

\item{pool_search_mode}{Solution pool search mode. Default is \code{2} for diverse solutions.}

\item{mip_gap}{Relative MIP gap target (stops when \code{|best-bound - best-incumbent| / |best-incumbent| <= mip_gap}).
Lower values force deeper search. Default is \code{1e-4}.}

\item{mip_gap_abs}{Absolute MIP gap target (stops when \code{|best-bound - best-incumbent| <= mip_gap_abs}).
Lower values force deeper search. Default is \code{1e-10}.}

\item{mip_focus}{Search focus: \code{0} (balance), \code{1} (find feasible solutions), \code{2} (prove optimality),
\code{3} (bound improvement). Default is \code{1}.}

\item{heuristics}{Heuristics effort in \code{[0,1]} where higher values spend more time on heuristics.
Default is \code{0.2}.}

\item{cuts}{Cut aggressiveness: \code{-1} (automatic), \code{0} (off), \code{1} (conservative),
\code{2} (aggressive), \code{3} (very aggressive). Default is \code{2}.}

\item{presolve}{Presolve aggressiveness: \code{-1} (automatic), \code{0} (off), \code{1} (conservative),
\code{2} (aggressive). Default is \code{2}.}
}
\value{
A list object which houses the results from Gurobi. Depending on the \code{gurobi_parms},
							the data within will be different. The most relevant tags are \code{x} for the best found solution and \code{objval}
							for the object
}
\description{
This method creates an object of type optimal_experimental_design and will immediately initiate
a search through allocation space for forced balance designs. Make sure you setup Gurobi properly first. This means
applying for a license, downloading, installing, registering it on your computer using the \code{grbgetkey} command
with the license file in the default directory. Then, in R, install the package from the file in your gurobi directory.
}
\details{
Currently, this method does not return multiple vectors. This will be improved in a later 
version. If you want this functionality now, use the hacked-up method \code{gurobi_multiple_designs}.
}
\examples{
\dontrun{
if ("gurobi" \%in\% loadedNamespaces()) {
  set.seed(1)
  X = matrix(rnorm(12), nrow = 6)
  gobj = initGurobiNumericalOptimizationExperimentalDesignObject(
    X,
    r = 2,
    num_cores = 1,
    initial_time_limit_sec = 5,
    verbose = FALSE
  )
  gobj$n
}
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{stopSearch}
\alias{stopSearch}
\title{Stops the parallelized greedy design search.}
\usage{
stopSearch(obj)
}
\arguments{
\item{obj}{The \code{experimental_design} object that is currently running the search}
}
\description{
Once stopped, it cannot be restarted.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
ged = initGreedyExperimentalDesignObject(
  X,
  max_designs = 3,
  num_cores = 1,
  start = TRUE,
  wait = FALSE,
  objective = "abs_sum_diff",
  verbose = FALSE
)
stopSearch(ged)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{plot_obj_val_by_iter}
\alias{plot_obj_val_by_iter}
\title{Plots the objective value by iteration}
\usage{
plot_obj_val_by_iter(res, runs = NULL)
}
\arguments{
\item{res}{Results from a greedy search object}

\item{runs}{A vector of run indices you would like to see plotted (default is to plot the first up to 9)}
}
\description{
Plots the objective value by iteration
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
ged = initGreedyExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  diagnostics = TRUE,
  objective = "abs_sum_diff",
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
res = resultsGreedySearch(ged, max_vectors = 2)
plot_obj_val_by_iter(res)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GreedyExperimentalDesign.R
\name{GreedyExperimentalDesign}
\alias{GreedyExperimentalDesign}
\title{Greedy Experimental Design Search}
\description{
A tool to find many types of a priori experimental designs
}
\references{
Kapelner, A
}
\author{
Adam Kapelner \email{kapelner@qc.cuny.edu}
}
\keyword{design}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{resultsGreedySearch}
\alias{resultsGreedySearch}
\title{Returns the results (thus far) of the greedy design search}
\usage{
resultsGreedySearch(obj, max_vectors = 9, form = "one_zero")
}
\arguments{
\item{obj}{The \code{greedy_experimental_design} object that is currently running the search}

\item{max_vectors}{The number of design vectors you wish to return. \code{NULL} returns all of them. 
This is not recommended as returning over 1,000 vectors is time-intensive. The default is 9.}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\description{
Returns the results (thus far) of the greedy design search
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
ged = initGreedyExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  objective = "abs_sum_diff",
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
res = resultsGreedySearch(ged, max_vectors = 2)
res$obj_vals
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete_randomization_and_blocking.R
\name{complete_randomization}
\alias{complete_randomization}
\title{Implements complete randomization (without forced balance)}
\usage{
complete_randomization(n, r, form = "one_zero")
}
\arguments{
\item{n}{number of observations}

\item{r}{number of randomized designs you would like}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\value{
a matrix where each column is one of the \code{r} designs
}
\description{
For debugging, you can use \code{set.seed}
to be assured of deterministic output.
}
\examples{
\dontrun{
complete_randomization(n = 6, r = 2)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_multiple_kernel_search.R
\name{initGreedyMultipleKernelExperimentalDesignObject}
\alias{initGreedyMultipleKernelExperimentalDesignObject}
\title{Begin A Greedy Pair Multiple Kernel Switching Search}
\usage{
initGreedyMultipleKernelExperimentalDesignObject(
  X = NULL,
  max_designs = 10000,
  objective = "added_pct_reduction",
  kernel_pre_num_designs = 2000,
  kernel_names = NULL,
  Kgrams = NULL,
  maximum_gain_scaling = 1.1,
  kernel_weights = NULL,
  wait = FALSE,
  start = TRUE,
  max_iters = Inf,
  semigreedy = FALSE,
  diagnostics = FALSE,
  num_cores = 1,
  seed = NULL,
  verbose = TRUE,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design. We will standardize this matrix by column internally.}

\item{max_designs}{The maximum number of designs to be returned. Default is 10,000. Make this large 
so you can search however long you wish as the search can be stopped at any time by
using the \code{\link{stopSearch}} method}

\item{objective}{The method used to aggregate the kernel objective functions together. Default is "added_pct_reduction".}

\item{kernel_pre_num_designs}{How many designs per kernel to run to explore the space of kernel objective values. Default is 2000.}

\item{kernel_names}{An array with the kernels to compute with default parameters. Must have elements in the following set:
"mahalanobis", "poly_s" where the "s" is a natural number 1 or greater,
"exponential", "laplacian", "inv_mult_quad", "gaussian". Default is \code{NULL} to
indicate the kernels are specified manually using the \code{Kgrams} parameter.}

\item{Kgrams}{A list of M >= 1 elements where each is a \code{n x n} matrix whose
entries are the evaluation of the kernel function between subject i and subject j. 
Default is \code{NULL} to indicate this was specified using the convenience parameter 
\code{kernel_names}.}

\item{maximum_gain_scaling}{This controls how much the percentage of possible improvement on a kernel objective function
should be scaled by. The minimum is 1 which allows for designs that could potentially have >=100%
improvement over original. We recommend 1.1 which means that a design that was found to be the best
of the \code{kernel_pre_num_designs} still has 1/1.1 = 9\% room to grow making it highly unlikely that
any design could be >= 100\%.}

\item{kernel_weights}{A vector with positive weights (need not be normalized) where each element represents the weight of 
each kernel. The default is \code{NULL} for uniform weighting.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
deafult is \code{FALSE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{max_iters}{Should we impose a maximum number of greedy switches? The default is \code{Inf} which a flag 
for ``no limit.''}

\item{semigreedy}{Should we use a fully greedy approach or the quicker semi-greedy approach? The default is
\code{FALSE} corresponding to the fully greedy approach.}

\item{diagnostics}{Returns diagnostic information about the iterations including (a) the initial starting
vectors, (b) the switches at every iteration and (c) information about the objective function
at every iteration (default is \code{FALSE} to decrease the algorithm's run time).}

\item{num_cores}{The number of CPU cores you wish to use during the search. The default is \code{1}.}

\item{seed}{The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
the output will not be deterministic. Default is \code{NULL} for no seed set.}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\value{
An object of type \code{greedy_experimental_design_search} which can be further operated upon
}
\description{
This method creates an object of type greedy_multiple_kernel_experimental_design and will immediately initiate
a search through allocation space for forced balance designs. For debugging, you can use set the \code{seed}
parameter and \code{num_cores = 1} to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
mk = initGreedyMultipleKernelExperimentalDesignObject(
  X,
  max_designs = 4,
  kernel_pre_num_designs = 4,
  num_cores = 1,
  kernel_names = c("mahalanobis", "gaussian"),
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
mk
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in
%   R/binary_match_followed_by_rerandomization_search.R
\name{initBinaryMatchFollowedByRerandomizationDesignSearchObject}
\alias{initBinaryMatchFollowedByRerandomizationDesignSearchObject}
\title{Begin a Search for Binary Matching Followed by Rerandomization}
\usage{
initBinaryMatchFollowedByRerandomizationDesignSearchObject(
  X,
  compute_dist_matrix = NULL,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{compute_dist_matrix}{The function that computes the distance matrix between every two observations in \code{X}, 
its only argument. The default is \code{NULL} signifying euclidean squared distance optimized in C++.}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}

\item{...}{Arguments passed to \code{initGreedyExperimentalDesignObject}. It is recommended to set
\code{max_designs} otherwise it will default to 10,000.}
}
\value{
An object of type \code{binary_experimental_design} which can be further operated upon.
}
\description{
This method creates an object of type binary_then_rerandomization_experimental_design and will find optimal matched pairs which
are then rerandomized in order to further minimize a balance metric. You can then
use the function \code{resultsBinaryMatchThenRerandomizationSearch} to obtain the randomized allocation vectors. For one column
in X, the matching just sorts the values to find the pairs trivially.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
obj = initBinaryMatchFollowedByRerandomizationDesignSearchObject(
  X,
  max_designs = 4,
  num_cores = 1,
  objective = "abs_sum_diff",
  obj_val_cutoff_to_include = Inf,
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
obj
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/randomization_metrics.R
\name{compute_randomization_metrics}
\alias{compute_randomization_metrics}
\title{Computes Randomization Metrics (explained in paper) about a design algorithm}
\usage{
compute_randomization_metrics(designs)
}
\arguments{
\item{designs}{A matrix where each column is one design.}
}
\value{
A list of resulting data: the probability estimates for
					each pair in the design of randomness where estmates close
					to ~0.5 represent random assignment, then the entropy metric
					the distance metric, the maximum eigenvalue of the allocation
					var-cov matrix (operator norm) and the squared Frobenius norm 
					(the sum of the squared eigenvalues)
}
\description{
Computes Randomization Metrics (explained in paper) about a design algorithm
}
\examples{
\dontrun{
designs = matrix(c(1, 0, 1, 0, 0, 1, 0, 1), nrow = 4, ncol = 2)
compute_randomization_metrics(designs)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_functions.R
\name{compute_gram_matrix}
\alias{compute_gram_matrix}
\title{Gram Matrix Computation}
\usage{
compute_gram_matrix(X, kernel_type, params = c())
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{kernel_type}{One of the following: "vanilla", "rbf", "poly", "tanh", "bessel", "laplace", 
"anova" or "spline".}

\item{params}{A vector of numeric parameters. Each \code{kernel_type} has different numbers of
parameters required. For more information see documentation for the \code{kernlab} library.}
}
\value{
The \code{n x n} gram matrix for the given kernel on the given data.
}
\description{
Computes the Gram Matrix for a user-specified kernel using the library \code{kernlab}. Note that 
this function automatically standardizes the columns of the data entered.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(12), nrow = 6)
K = compute_gram_matrix(X, kernel_type = "rbf", params = 0.5)
dim(K)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_orthogonalization_curation2.R
\name{greedy_orthogonalization_curation2}
\alias{greedy_orthogonalization_curation2}
\title{Curate More Orthogonal Vectors Greedily}
\usage{
greedy_orthogonalization_curation2(W, R0 = 100, verbose = FALSE)
}
\arguments{
\item{W}{A matrix in \eqn{{-1, 1}^{R x n}} which have R allocation vectors for an experiment of sample size n.}

\item{R0}{The minimum number of vectors to consider in a design. The default is the true bottom, two.}

\item{verbose}{Default is \code{FALSE} but if not, it will print out a message for each iteration.}
}
\value{
A list with two elements: (1) \code{avg_abs_rij_by_R} which is a data frame with R - Rmin + 1 rows and 
					columns R and average absolute r_ij and (2) \code{Wsorted} which provides the collection of vectors in
					sorted by best average absolute r_ij in row order from best to worst.
}
\description{
This function takes a set of allocation vectors and pares them down one-by-one
by eliminating the vector that can result in the largest reduction in Avg[ |r_ij| ].
It is recommended to begin with a set of unmirrored vectors for speed. Then add the mirrors later
for whichever subset you wish.
}
\examples{
\dontrun{
set.seed(1)
W = matrix(sample(c(-1, 1), 6 * 8, replace = TRUE), nrow = 6)
W2 = greedy_orthogonalization_curation2(W, R0 = 4, verbose = FALSE)
dim(W2)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimal_search.R
\name{initOptimalExperimentalDesignObject}
\alias{initOptimalExperimentalDesignObject}
\title{Begin a Search for the Optimal Solution}
\usage{
initOptimalExperimentalDesignObject(
  X = NULL,
  objective = "mahal_dist",
  Kgram = NULL,
  wait = FALSE,
  start = TRUE,
  num_cores = 1,
  verbose = TRUE,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{objective}{The objective function to use when searching design space. This is a string
with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".}

\item{Kgram}{If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
deafult is \code{FALSE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{num_cores}{The number of CPU cores you wish to use during the search. The default is \code{1}.}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\value{
An object of type \code{optimal_experimental_design_search} which can be further operated upon
}
\description{
This method creates an object of type optimal_experimental_design and will immediately initiate
a search through allocation space. Since this search takes exponential time, for most machines, 
this method is futile beyond 28 samples. You've been warned! For debugging, you can use set 
\code{num_cores = 1} to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(12), nrow = 6)
obj = initOptimalExperimentalDesignObject(
  X,
  objective = "abs_sum_diff",
  num_cores = 1,
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
obj
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerandomization_search.R
\name{resultsRerandomizationSearch}
\alias{resultsRerandomizationSearch}
\title{Returns the results (thus far) of the rerandomization design search}
\usage{
resultsRerandomizationSearch(
  obj,
  include_assignments = FALSE,
  form = "one_zero"
)
}
\arguments{
\item{obj}{The \code{rerandomization_experimental_design} object that is currently running the search}

\item{include_assignments}{Do we include the assignments (takes time) and default is \code{FALSE}.}

\item{form}{Which form should the assignments be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\description{
Returns the results (thus far) of the rerandomization design search
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
obj = initRerandomizationExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  objective = "abs_sum_diff",
  obj_val_cutoff_to_include = Inf,
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
res = resultsRerandomizationSearch(obj, include_assignments = TRUE, form = "one_zero")
dim(res$ending_indicTs)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{all_elements_same_cpp_wrap}
\alias{all_elements_same_cpp_wrap}
\title{Tests if a vector has all elements the same}
\usage{
all_elements_same_cpp_wrap(w)
}
\arguments{
\item{w}{The vector to be queried}
}
\value{
A boolean if it has all same elements
}
\description{
Tests if a vector has all elements the same
}
\examples{
\dontrun{
all_elements_same_cpp_wrap(c(1, 1, 1))
all_elements_same_cpp_wrap(c(1, 2, 1))
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{create_all_ys_cpp_wrap}
\alias{create_all_ys_cpp_wrap}
\title{Create all binary Y's convenience function using a randomized design}
\usage{
create_all_ys_cpp_wrap(pCs, pTs, W, two_n, nY)
}
\arguments{
\item{pCs}{Control-group success probabilities (length \code{two_n})}

\item{pTs}{Treatment-group success probabilities (length \code{two_n})}

\item{W}{Assignment matrix with \code{nY} rows and \code{two_n} columns}

\item{two_n}{Total number of units}

\item{nY}{Number of Y vectors to generate}
}
\value{
A matrix of boolean Y's
}
\description{
Create all binary Y's convenience function using a randomized design
}
\examples{
\dontrun{
pCs = rep(0.2, 4)
pTs = rep(0.8, 4)
W = matrix(c(1, 0, 1, 0, 0, 1, 0, 1), nrow = 2, byrow = TRUE)
create_all_ys_cpp_wrap(pCs, pTs, W, two_n = 4, nY = 2)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimal_search.R
\name{resultsOptimalSearch}
\alias{resultsOptimalSearch}
\title{Returns the results (thus far) of the optimal design search}
\usage{
resultsOptimalSearch(obj, num_vectors = 2, form = "one_zero")
}
\arguments{
\item{obj}{The \code{optimal_experimental_design} object that is currently running the search}

\item{num_vectors}{How many allocation vectors you wish to return. The default is 1 meaning the best vector. If \code{Inf},
it means all vectors.}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\description{
Returns the results (thus far) of the optimal design search
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(12), nrow = 6)
obj = initOptimalExperimentalDesignObject(
  X,
  objective = "abs_sum_diff",
  num_cores = 1,
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
res = resultsOptimalSearch(obj, num_vectors = 2, form = "one_zero")
res$opt_obj_val
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/karp_search.R
\name{print.karp_experimental_design_search}
\alias{print.karp_experimental_design_search}
\title{Prints a summary of a \code{karp_experimental_design_search} object}
\usage{
\method{print}{karp_experimental_design_search}(x, ...)
}
\arguments{
\item{x}{The \code{karp_experimental_design_search} object to be summarized in the console}

\item{...}{Other parameters to pass to the default print function}
}
\description{
Prints a summary of a \code{karp_experimental_design_search} object
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{shuffle_cpp_wrap}
\alias{shuffle_cpp_wrap}
\title{Shuffles a vector rapidly}
\usage{
shuffle_cpp_wrap(w, seed = NA_integer_)
}
\arguments{
\item{w}{The vector to be shuffled}

\item{seed}{Optional integer seed; use NA to draw from the system clock}
}
\value{
The vector with elements shuffled
}
\description{
Shuffles a vector rapidly
}
\examples{
\dontrun{
shuffle_cpp_wrap(1:5, seed = 1)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{optimize_asymmetric_treatment_assignment}
\alias{optimize_asymmetric_treatment_assignment}
\title{Compute Optimal Number of Treatments/Controls}
\usage{
optimize_asymmetric_treatment_assignment(
  c_treatment = NULL,
  c_control = NULL,
  c_total_max = NULL,
  n = NULL
)
}
\arguments{
\item{c_treatment}{The cost of a treatment assignment. Default is \code{NULL} for symmetric costs.}

\item{c_control}{The cost of a control assignment. Default is \code{NULL} for symmetric costs.}

\item{c_total_max}{The total cost constraint of any allocation. Either this or \code{n} must be specified. Default is \code{NULL}.}

\item{n}{The total cost constraint as specified by the total number of subjects. Either this or \code{c_total} must be 
specified. Default is \code{NULL}.}
}
\value{
A list with three keys: n, nT, nC plus specified arguments
}
\description{
Given a total budget and asymmetric treatment and control costs, calculate the
number of treatments and controls that optimize the variance of the estimator. 
The number of treatments is rounded up by default.
}
\examples{
\dontrun{
optimize_asymmetric_treatment_assignment(n = 100)
optimize_asymmetric_treatment_assignment(n = 100, c_treatment = 2, c_control = 1)
optimize_asymmetric_treatment_assignment(c_total_max = 50, c_treatment = 2, c_control = 1)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binary_match_search.R
\name{resultsBinaryMatchSearch}
\alias{resultsBinaryMatchSearch}
\title{Binary Pair Match Search}
\usage{
resultsBinaryMatchSearch(obj, form = "one_zero")
}
\arguments{
\item{obj}{The \code{pairwise_matching_experimental_design_search} object that is currently running the search}

\item{form}{Which form should the assignments be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\description{
Returns the results (thus far) of the binary pair match design search
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
bms = computeBinaryMatchStructure(X)
bm = initBinaryMatchExperimentalDesignSearchObject(
  bms,
  max_designs = 4,
  num_cores = 1,
  start = TRUE,
  wait = TRUE,
  seed = 1,
  verbose = FALSE
)
res = resultsBinaryMatchSearch(bm, form = "one_zero")
dim(res)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{safe_cov_inverse}
\alias{safe_cov_inverse}
\title{Computes a numerically stable inverse of a covariance matrix}
\usage{
safe_cov_inverse(X, ridge = 1e-08, max_ridge_steps = 6)
}
\arguments{
\item{X}{The n x p design matrix}

\item{ridge}{Initial ridge penalty added to the diagonal}

\item{max_ridge_steps}{Maximum number of ridge escalation attempts}
}
\value{
The inverse covariance matrix
}
\description{
Computes a numerically stable inverse of a covariance matrix
}
\examples{
\dontrun{
X = matrix(rnorm(20), nrow = 10)
Sinv = safe_cov_inverse(X)
dim(Sinv)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete_randomization_and_blocking.R
\name{gen_var_cov_matrix_block_designs}
\alias{gen_var_cov_matrix_block_designs}
\title{Computes varcov matrix for block designs}
\usage{
gen_var_cov_matrix_block_designs(n, prop_T, B, use_cache = TRUE)
}
\arguments{
\item{n}{number of observations}

\item{prop_T}{the proportion of treatments allocated}

\item{B}{the number of blocks}

\item{use_cache}{Cache results for repeated calls with identical inputs. Default is \code{TRUE}.}
}
\value{
varcov matrix for the specific block design
}
\description{
The varcov matrix for block designs consists of a block-
diagonal matrix with B blocks (the number of blocks in
the design) with off-diagonal entries = -1 / (n/B - 1)
where n is the number of subjected in the study.
}
\examples{
\dontrun{
gen_var_cov_matrix_block_designs(n = 12, prop_T = 0.5, B = 3)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_multiple_kernel_search.R
\name{plot.greedy_multiple_kernel_experimental_design}
\alias{plot.greedy_multiple_kernel_experimental_design}
\title{Plots a summary of a \code{greedy_multiple_kernel_experimental_design} object}
\usage{
\method{plot}{greedy_multiple_kernel_experimental_design}(x, ...)
}
\arguments{
\item{x}{The \code{greedy_multiple_kernel_experimental_design} object to be summarized in the plot}

\item{...}{Other parameters to pass to the default plot function}
}
\value{
An array of order statistics from \link{plot_obj_val_order_statistic} as a list element
}
\description{
Plots a summary of a \code{greedy_multiple_kernel_experimental_design} object
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{standardize_data_matrix}
\alias{standardize_data_matrix}
\title{Standardizes the columns of a data matrix.}
\usage{
standardize_data_matrix(X)
}
\arguments{
\item{X}{The n x p design matrix}
}
\value{
The n x p design matrix with columns standardized
}
\description{
Standardizes the columns of a data matrix.
}
\examples{
\dontrun{
X = matrix(rnorm(12), nrow = 6)
Xstd = standardize_data_matrix(X)
colMeans(Xstd)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete_randomization_and_blocking.R
\name{imbalanced_complete_randomization}
\alias{imbalanced_complete_randomization}
\title{Implements unequally allocated complete randomization}
\usage{
imbalanced_complete_randomization(n, prop_T, r, form = "one_zero", seed = NULL)
}
\arguments{
\item{n}{number of observations}

\item{prop_T}{the proportion of treatments needed}

\item{r}{number of randomized designs you would like}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}

\item{seed}{An integer which is the seed to be set within C++. Default is \code{NULL} which means the seed is set from the system clock.}
}
\value{
a matrix where each column is one of the \code{r} designs
}
\description{
For debugging, you can use \code{set.seed}
to be assured of deterministic output.
}
\examples{
\dontrun{
imbalanced_complete_randomization(n = 10, prop_T = 0.3, r = 2, seed = 1)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{initGreedyExperimentalDesignObject}
\alias{initGreedyExperimentalDesignObject}
\title{Begin A Greedy Pair Switching Search}
\usage{
initGreedyExperimentalDesignObject(
  X = NULL,
  nT = NULL,
  max_designs = 10000,
  objective = "mahal_dist",
  indicies_pairs = NULL,
  Kgram = NULL,
  wait = FALSE,
  start = TRUE,
  max_iters = Inf,
  semigreedy = FALSE,
  diagnostics = FALSE,
  num_cores = 1,
  seed = NULL,
  verbose = TRUE,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design. This parameter must be specified unless you
choose objective type \code{"kernel"} in which case, the \code{Kgram} parameter must
be specified.}

\item{nT}{The number of treatments to assign. Default is \code{NULL} which is for forced balance allocation
i.e. nT = nC = n / 2 where n is the number of rows in X (or Kgram if X is unspecified).}

\item{max_designs}{The maximum number of designs to be returned. Default is 10,000. Make this large 
so you can search however long you wish as the search can be stopped at any time by
using the \code{\link{stopSearch}} method}

\item{objective}{The objective function to use when searching design space. This is a string
with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".}

\item{indicies_pairs}{A matrix of size $n/2$ times 2 whose rows are indicies pairs. The values of the entire matrix 
must enumerate all indicies $1, ..., n$. The default is \code{NULL} meaning to use all possible pairs.}

\item{Kgram}{If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
deafult is \code{FALSE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{max_iters}{Should we impose a maximum number of greedy switches? The default is \code{Inf} which a flag 
for ``no limit.''}

\item{semigreedy}{Should we use a fully greedy approach or the quicker semi-greedy approach? The default is
\code{FALSE} corresponding to the fully greedy approach.}

\item{diagnostics}{Returns diagnostic information about the iterations including (a) the initial starting
vectors, (b) the switches at every iteration and (c) information about the objective function
at every iteration (default is \code{FALSE} to decrease the algorithm's run time).}

\item{num_cores}{The number of CPU cores you wish to use during the search. The default is \code{1}.}

\item{seed}{The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
the output will not be deterministic. Default is \code{NULL} for no seed set.}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\value{
An object of type \code{greedy_experimental_design_search} which can be further operated upon
}
\description{
This method creates an object of type greedy_experimental_design and will immediately initiate
a search through allocation space for forced balance designs. For debugging, you can use set the \code{seed}
parameter and \code{num_cores = 1} to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
ged = initGreedyExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  objective = "abs_sum_diff",
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
ged
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerandomization_search.R
\name{initRerandomizationExperimentalDesignObject}
\alias{initRerandomizationExperimentalDesignObject}
\title{Begin a Rerandomization Search}
\usage{
initRerandomizationExperimentalDesignObject(
  X = NULL,
  obj_val_cutoff_to_include,
  max_designs = 1000,
  objective = "mahal_dist",
  Kgram = NULL,
  wait = FALSE,
  start = TRUE,
  num_cores = 1,
  seed = NULL,
  verbose = TRUE,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{obj_val_cutoff_to_include}{Only allocation vectors with objective values lower than this threshold will be returned.
If the cutoff is infinity, you are doing BCRD and you should use the \code{complete_randomization_with_forced_balanced}
function instead.}

\item{max_designs}{The maximum number of designs to be returned. Default is 10,000. Make this large 
so you can search however long you wish as the search can be stopped at any time by
using the \code{\link{stopSearch}} method}

\item{objective}{The objective function to use when searching design space. This is a string
with valid values "\code{mahal_dist}" (the default), "\code{abs_sum_diff}" or "\code{kernel}".}

\item{Kgram}{If the \code{objective = kernel}, this argument is required to be an \code{n x n} matrix whose
entries are the evaluation of the kernel function between subject i and subject j. Default is \code{NULL}.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
default is \code{FALSE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{num_cores}{The number of CPU cores you wish to use during the search. The default is \code{1}.}

\item{seed}{The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
the output will not be deterministic. Default is \code{NULL} for no seed set.}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\value{
An object of type \code{rerandomization_experimental_design_search} which can be further operated upon.
}
\description{
This method creates an object of type rerandomization_experimental_design and will immediately initiate
a search through allocation space for forced-balance designs. For debugging, you can use set the \code{seed}
parameter and \code{num_cores = 1} to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
obj = initRerandomizationExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  objective = "abs_sum_diff",
  obj_val_cutoff_to_include = Inf,
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
obj
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binary_match_search.R
\name{computeBinaryMatchStructure}
\alias{computeBinaryMatchStructure}
\title{Compute Binary Matching Strcuture}
\usage{
computeBinaryMatchStructure(
  X,
  mahal_match = FALSE,
  compute_dist_matrix = NULL,
  D = NULL,
  symmetry_tol = 1e-12,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more optimal design.}

\item{mahal_match}{Match using Mahalanobis distance. Default is \code{FALSE}.}

\item{compute_dist_matrix}{The function that computes the distance matrix between every two observations in \code{X}, 
its only argument. The default is \code{NULL} signifying euclidean squared distance optimized in C++.}

\item{D}{A distance matrix precomputed. The default is \code{NULL} indicating the distance matrix should be computed.}

\item{symmetry_tol}{Tolerance for symmetry check on \code{D}. Default is \code{1e-12}.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\value{
An object of type \code{binary_experimental_design} which can be further operated upon.
}
\description{
This method creates an object of type binary_match_structure and will compute pairs. You can then
use the functions \code{initBinaryMatchExperimentalDesignSearchObject} and \code{resultsBinaryMatchSearch} 
to create randomized allocation vectors. For one column in X, we just sort to find the pairs trivially.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
bms = computeBinaryMatchStructure(X)
bms$indicies_pairs
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{generate_stdzied_design_matrix}
\alias{generate_stdzied_design_matrix}
\title{Generates a design matrix with standardized predictors.}
\usage{
generate_stdzied_design_matrix(n = 50, p = 1, covariate_gen = rnorm, ...)
}
\arguments{
\item{n}{Number of rows in the design matrix}

\item{p}{Number of columns in the design matrix}

\item{covariate_gen}{The function to use to draw the covariate realizations (assumed to be iid).
This defaults to \code{rnorm} for $N(0,1)$ draws.}

\item{...}{Optional arguments to be passed to the \code{covariate_dist} function.}
}
\value{
THe design matrix
}
\description{
This function is useful for debugging.
}
\examples{
\dontrun{
X = generate_stdzied_design_matrix(n = 6, p = 2)
colMeans(X)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{plot_obj_val_order_statistic}
\alias{plot_obj_val_order_statistic}
\title{Plots an order statistic of the object value as a function of number of searches}
\usage{
plot_obj_val_order_statistic(
  obj,
  order_stat = 1,
  skip_every = 5,
  type = "o",
  ...
)
}
\arguments{
\item{obj}{The greedy search object object whose search history is to be visualized}

\item{order_stat}{The order statistic that you wish to plot. The default is \code{1} for the minimum.}

\item{skip_every}{Plot every nth point. This makes the plot generate much more quickly. The default is \code{5}.}

\item{type}{The type parameter for plot.}

\item{...}{Other arguments to be passed to the plot function.}
}
\value{
An array of order statistics as a list element
}
\description{
Plots an order statistic of the object value as a function of number of searches
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(20), nrow = 10)
ged = initGreedyExperimentalDesignObject(
  X,
  max_designs = 5,
  num_cores = 1,
  objective = "abs_sum_diff",
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
plot_obj_val_order_statistic(ged, order_stat = 1, skip_every = 1)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binary_match_followed_by_greedy_search.R
\name{resultsBinaryMatchThenGreedySearch}
\alias{resultsBinaryMatchThenGreedySearch}
\title{Returns unique allocation vectors that are binary matched}
\usage{
resultsBinaryMatchThenGreedySearch(
  obj,
  num_vectors = NULL,
  compute_obj_vals = FALSE,
  form = "one_zero",
  use_safe_inverse = FALSE
)
}
\arguments{
\item{obj}{The \code{binary_then_greedy_experimental_design} object where the pairs are computed.}

\item{num_vectors}{How many random allocation vectors you wish to return. The default is \code{NULL} indicating you want all of them.}

\item{compute_obj_vals}{Should we compute all the objective values for each allocation? Default is \code{FALSE}.}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\description{
Returns unique allocation vectors that are binary matched
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
obj = initBinaryMatchFollowedByGreedyExperimentalDesignSearchObject(
  X,
  max_designs = 4,
  num_cores = 1,
  objective = "abs_sum_diff",
  start = TRUE,
  wait = TRUE,
  verbose = FALSE
)
res = resultsBinaryMatchThenGreedySearch(obj, num_vectors = 3, form = "one_zero")
dim(res$indicTs)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gurobi_numerical_optimization_search.R
\name{resultsGurobiNumericalOptimizeSearch}
\alias{resultsGurobiNumericalOptimizeSearch}
\title{Query the Gurobi Results}
\usage{
resultsGurobiNumericalOptimizeSearch(obj)
}
\arguments{
\item{obj}{The \code{gurobi_numerical_optimization_experimental_design_search} object that is currently running the search}
}
\description{
Returns the results (thus far) of the Gurobi numerical optimization design search
}
\examples{
\dontrun{
if ("gurobi" \%in\% loadedNamespaces()) {
  set.seed(1)
  X = matrix(rnorm(12), nrow = 6)
  gobj = initGurobiNumericalOptimizationExperimentalDesignObject(
    X,
    r = 2,
    num_cores = 1,
    initial_time_limit_sec = 5,
    verbose = FALSE
  )
  res = resultsGurobiNumericalOptimizeSearch(gobj)
  res$obj_vals
}
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binary_match_search.R
\name{initBinaryMatchExperimentalDesignSearchObject}
\alias{initBinaryMatchExperimentalDesignSearchObject}
\title{Begin a Binary Match Search}
\usage{
initBinaryMatchExperimentalDesignSearchObject(
  binary_match_structure,
  max_designs = 1000,
  wait = FALSE,
  start = TRUE,
  num_cores = 1,
  seed = NULL,
  prop_flips = 1,
  verbose = TRUE
)
}
\arguments{
\item{binary_match_structure}{The \code{binary_experimental_design} object where the pairs are computed.}

\item{max_designs}{How many random allocation vectors you wish to return. The default is 1000.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
default is \code{FALSE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{num_cores}{The number of CPU cores you wish to use during the search. The default is \code{1}.}

\item{seed}{The set to set for deterministic output. This should only be set if \code{num_cores = 1} otherwise
the output will not be deterministic. Default is \code{NULL} for no seed set.}

\item{prop_flips}{Proportion of flips. Default is all. Lower for more correlated assignments (useful for research only).}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}
}
\description{
This method creates an object of type pairwise_matching_experimental_design_search and will immediately initiate
a search through allocation space for pairwise match designs based on the structure computed in the function \code{computeBinaryMatchStructure}. 
For debugging, you can use set the \code{seed} parameter and \code{num_cores = 1} to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
bms = computeBinaryMatchStructure(X)
bm = initBinaryMatchExperimentalDesignSearchObject(
  bms,
  max_designs = 4,
  num_cores = 1,
  start = TRUE,
  wait = TRUE,
  seed = 1,
  verbose = FALSE
)
bm
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/greedy_search.R
\name{plot.greedy_experimental_design_search}
\alias{plot.greedy_experimental_design_search}
\title{Plots a summary of a greedy search object object}
\usage{
\method{plot}{greedy_experimental_design_search}(x, ...)
}
\arguments{
\item{x}{The greedy search object object to be summarized in the plot}

\item{...}{Other parameters to pass to the default plot function}
}
\value{
An array of order statistics from \link{plot_obj_val_order_statistic} as a list element
}
\description{
Plots a summary of a greedy search object object
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{gen_pm_designs_cpp_wrap}
\alias{gen_pm_designs_cpp_wrap}
\title{Create PM designs}
\usage{
gen_pm_designs_cpp_wrap(indicies_pairs, n, r)
}
\arguments{
\item{indicies_pairs}{A matrix of n x 2 indicies where each row is a pair of subjects' indicies}

\item{n}{Half the number of subjects i.e. the number of pairs}

\item{r}{The number of assignments to generate}
}
\value{
A matrix of r x 2n PM designs of +1/-1 assignments
}
\description{
Create PM designs
}
\examples{
\dontrun{
indicies_pairs = matrix(c(1, 2, 3, 4), ncol = 2, byrow = TRUE)
gen_pm_designs_cpp_wrap(indicies_pairs, n = 2, r = 3)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{compute_objective_val}
\alias{compute_objective_val}
\title{Computes Objective Value From Allocation Vector}
\usage{
compute_objective_val(
  X,
  indic_T,
  objective = "abs_sum_diff",
  inv_cov_X = NULL,
  use_safe_inverse = FALSE
)
}
\arguments{
\item{X}{The n x p design matrix}

\item{indic_T}{The n-length binary allocation vector}

\item{objective}{The objective function to use. Default is \code{abs_sum_diff} and the other option is 
\code{mahal_dist}.}

\item{inv_cov_X}{Optional: the inverse sample variance covariance matrix. Use this
argument if you will be doing many calculations since passing this
in will cache this data.}

\item{use_safe_inverse}{Should a regularized inverse be used for the Mahalanobis objective?
Default is \code{FALSE}.}
}
\description{
Returns the objective value given a design vector as well an an objective function.
This is sometimes duplicated in Java. However, within Java, tricks are played to make
optimization go faster so Java's objective values may not always be the same as the true
objective function (e.g. logs or constants dropped).
}
\examples{
\dontrun{
X = matrix(rnorm(12), nrow = 6)
indic_T = c(1, 0, 1, 0, 1, 0)
compute_objective_val(X, indic_T, objective = "abs_sum_diff")
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hadamard_search.R
\name{hadamardExperimentalDesign}
\alias{hadamardExperimentalDesign}
\title{Create a Hadamard Design}
\usage{
hadamardExperimentalDesign(X, strict = TRUE, form = "one_zero")
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). The measurements aren't used to
compute the Hadamard designs, only the number of rows.}

\item{strict}{Hadamard matrices are not available for all $n$.}

\item{form}{Which form should it be in? The default is \code{one_zero} for 1/0's or \code{pos_one_min_one} for +1/-1's.}
}
\value{
An matrix of dimension $R$ x $n$ where $R$ is the number of Hadamard allocations.
}
\description{
This method returns unique designs according to a Hadamard matrix. For debugging, you can use \code{set.seed}
to be assured of deterministic output.
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(16), nrow = 8)
W = hadamardExperimentalDesign(X, strict = TRUE, form = "one_zero")
dim(W)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerandomization_search.R
\name{summary.rerandomization_experimental_design_search}
\alias{summary.rerandomization_experimental_design_search}
\title{Prints a summary of a \code{rerandomization_experimental_design_search} object}
\usage{
\method{summary}{rerandomization_experimental_design_search}(object, ...)
}
\arguments{
\item{object}{The \code{rerandomization_experimental_design_search} object to be summarized in the console}

\item{...}{Other parameters to pass to the default summary function}
}
\description{
Prints a summary of a \code{rerandomization_experimental_design_search} object
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/karp_search.R
\name{resultsKarpSearch}
\alias{resultsKarpSearch}
\title{Returns the results (thus far) of the karp design search}
\usage{
resultsKarpSearch(obj)
}
\arguments{
\item{obj}{The \code{karp_experimental_design} object that is currently running the search}
}
\description{
Returns the results (thus far) of the karp design search
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(10), nrow = 10)
kobj = initKarpExperimentalDesignObject(
  X,
  start = TRUE,
  wait = TRUE,
  balanced = TRUE,
  verbose = FALSE
)
res = resultsKarpSearch(kobj)
res$obj_val
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{compute_distance_matrix_cpp_wrap}
\alias{compute_distance_matrix_cpp_wrap}
\title{Computes a Euclidean-squared distance matrix rapidly}
\usage{
compute_distance_matrix_cpp_wrap(X)
}
\arguments{
\item{X}{A numeric matrix with n rows representing each subject and p columns which are measurements on each subject}
}
\value{
The n x n Euclidean distances squared
}
\description{
Computes a Euclidean-squared distance matrix rapidly
}
\examples{
\dontrun{
X = matrix(c(0, 1, 2, 3), nrow = 2)
compute_distance_matrix_cpp_wrap(X)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/karp_search.R
\name{initKarpExperimentalDesignObject}
\alias{initKarpExperimentalDesignObject}
\title{Begin Karp Search}
\usage{
initKarpExperimentalDesignObject(
  X,
  wait = FALSE,
  balanced = TRUE,
  start = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{X}{The design matrix with $n$ rows (one for each subject) and $p$ columns 
(one for each measurement on the subject). This is the design matrix you wish 
to search for a more karp design.}

\item{wait}{Should the \code{R} terminal hang until all \code{max_designs} vectors are found? The 
deafult is \code{FALSE}.}

\item{balanced}{Should the final vector be balanced? Default and recommended is \code{TRUE}.}

\item{start}{Should we start searching immediately (default is \code{TRUE}).}

\item{verbose}{Should the algorithm emit progress output? Default is \code{TRUE}.}
}
\value{
An object of type \code{karp_experimental_design_search} which can be further operated upon
}
\description{
This method creates an object of type karp_experimental_design and will immediately initiate
a search through allocation  space. Note that the Karp search only works 
for one covariate (i.e. $p=1$) and the objective "abs_sum_diff".
}
\examples{
\dontrun{
set.seed(1)
X = matrix(rnorm(10), nrow = 10)
kobj = initKarpExperimentalDesignObject(
  X,
  start = TRUE,
  wait = TRUE,
  balanced = TRUE,
  verbose = FALSE
)
kobj
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared.R
\name{generate_block_design_cpp_wrap}
\alias{generate_block_design_cpp_wrap}
\title{Generates homogeneous block design allocations rapidly}
\usage{
generate_block_design_cpp_wrap(B, nR, dummy_block)
}
\arguments{
\item{B}{The number of blocks in the design}

\item{nR}{The number of allocation vectors}

\item{dummy_block}{The subvector of allocations in each block that will be permuted}
}
\value{
A matrix with rows being the nR random block allocation of sample size B x length(dummy_block).
}
\description{
Generates homogeneous block design allocations rapidly
}
\examples{
\dontrun{
generate_block_design_cpp_wrap(B = 2, nR = 3, dummy_block = c(1, 0))
}
}
\author{
Adam Kapelner
}
